/*
 * FTGL - OpenGL font library
 *
 * Copyright (c) 2001-2004 Henry Maddocks <ftgl@opengl.geek.nz>
 * Copyright (c) 2008 Sam Hocevar <sam@zoy.org>
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include "config.h"

#include <cassert>
#include <string> // For memset

#include "FTGL/ftgl.h"

#include "FTInternals.h"

#include "../FTGlyph/FTTextureGlyphImpl.h"
#include "./FTTextureFontImpl.h"


//
//  FTTextureFont
//


FTTextureFont::FTTextureFont(char const *fontFilePath) :
    FTFont(new FTTextureFontImpl(this, fontFilePath))
{}


FTTextureFont::FTTextureFont(const unsigned char *pBufferBytes,
                             size_t bufferSizeInBytes) :
    FTFont(new FTTextureFontImpl(this, pBufferBytes, bufferSizeInBytes))
{}


FTTextureFont::~FTTextureFont()
{}

void FTTextureFont::pushGlyph(unsigned int textureID,
                              const FTGL_FLOAT quadVertices[2 * 4],
                              const FTGL_FLOAT quadTexCoords[2 * 4])
{
    FTTextureFontImpl* myimpl = dynamic_cast<FTTextureFontImpl*>(impl);
    if(!myimpl)
    {
        return;
    }
    myimpl->pushGlyph(textureID, quadVertices, quadTexCoords);
}

void FTTextureFont::reset()
{
    FTTextureFontImpl* myimpl = dynamic_cast<FTTextureFontImpl*>(impl);
    if(!myimpl)
    {
        return;
    }
    myimpl->reset();
}

size_t FTTextureFont::getTexturesCount()
{
    FTTextureFontImpl* myimpl = dynamic_cast<FTTextureFontImpl*>(impl);
    if(!myimpl)
    {
        return 0;
    }
    return myimpl->getTexturesCount();
}

unsigned int FTTextureFont::getTextureId(size_t index)
{
    FTTextureFontImpl* myimpl = dynamic_cast<FTTextureFontImpl*>(impl);
    if(!myimpl)
    {
        return 0;
    }
    return myimpl->getTextureId(index);
}

const FTGL_FLOAT* FTTextureFont::getVertices(size_t index, size_t* outElemSize, size_t* outElemCount)
{
    FTTextureFontImpl* myimpl = dynamic_cast<FTTextureFontImpl*>(impl);
    if(!myimpl)
    {
        *outElemSize = 0;
        *outElemCount = 0;
        return NULL;
    }
    return myimpl->getVertices(index, outElemSize, outElemCount);
}

const FTGL_FLOAT* FTTextureFont::getTexCoords(size_t index, size_t* outElemSize, size_t* outElemCount)
{
    FTTextureFontImpl* myimpl = dynamic_cast<FTTextureFontImpl*>(impl);
    if(!myimpl)
    {
        *outElemSize = 0;
        *outElemCount = 0;
        return NULL;
    }
    return myimpl->getTexCoords(index, outElemSize, outElemCount);
}

FTGlyph* FTTextureFont::MakeGlyph(FT_GlyphSlot ftGlyph)
{
    FTTextureFontImpl *myimpl = dynamic_cast<FTTextureFontImpl *>(impl);
    if(!myimpl)
    {
        return NULL;
    }

    return myimpl->MakeGlyphImpl(ftGlyph);
}


//
//  FTTextureFontImpl
//


static inline unsigned int NextPowerOf2(unsigned int in)
{
     in -= 1;

     in |= in >> 16;
     in |= in >> 8;
     in |= in >> 4;
     in |= in >> 2;
     in |= in >> 1;

     return in + 1;
}


FTTextureFontImpl::FTTextureFontImpl(FTFont *ftFont, const char* fontFilePath)
:   FTFontImpl(ftFont, fontFilePath),
    maximumGLTextureSize(0),
    textureWidth(0),
    textureHeight(0),
    glyphHeight(0),
    glyphWidth(0),
    padding(3),
    xOffset(0),
    yOffset(0),
    faceSize(0),
    faceRes(0),
    textVertices(),
    textTextCoords()
{
    load_flags = FT_LOAD_NO_HINTING | FT_LOAD_NO_BITMAP;
    remGlyphs = numGlyphs = face.GlyphCount();
}

FTTextureFontImpl::FTTextureFontImpl(FTFont *ftFont,
                                     const unsigned char *pBufferBytes,
                                     size_t bufferSizeInBytes)
:   FTFontImpl(ftFont, pBufferBytes, bufferSizeInBytes),
    maximumGLTextureSize(0),
    textureWidth(0),
    textureHeight(0),
    glyphHeight(0),
    glyphWidth(0),
    padding(3),
    xOffset(0),
    yOffset(0)
{
    load_flags = FT_LOAD_NO_HINTING | FT_LOAD_NO_BITMAP;
    remGlyphs = numGlyphs = face.GlyphCount();
}


FTTextureFontImpl::~FTTextureFontImpl()
{
    if(textureIDList.size())
    {
        glDeleteTextures((GLsizei)textureIDList.size(),
                         (const GLuint*)&textureIDList[0]);
    }
}


void FTTextureFontImpl::pushGlyph(unsigned int textureID,
                                  const FTGL_FLOAT quadVertices[2 * 4],
                                  const FTGL_FLOAT quadTexCoords[2 * 4])
{
    size_t listId = 0;
    for(listId = 0; listId < textureIDList.size(); ++listId)
    {
        if(textureIDList[listId] == textureID)
        {
            break;
        }
    }
    if(listId >= textureIDList.size())
    {
        return;
    }

    // first triangle
    textVertices[listId].push_back(quadVertices[0]); // bottom left
    textVertices[listId].push_back(quadVertices[1]);
    textVertices[listId].push_back(0.0f);
    textVertices[listId].push_back(1.0f);
    textTextCoords[listId].push_back(quadTexCoords[0]);
    textTextCoords[listId].push_back(quadTexCoords[1]);
    textVertices[listId].push_back(quadVertices[2]); // top left
    textVertices[listId].push_back(quadVertices[3]);
    textVertices[listId].push_back(0.0f);
    textVertices[listId].push_back(1.0f);
    textTextCoords[listId].push_back(quadTexCoords[2]);
    textTextCoords[listId].push_back(quadTexCoords[3]);
    textVertices[listId].push_back(quadVertices[4]); // top right
    textVertices[listId].push_back(quadVertices[5]);
    textVertices[listId].push_back(0.0f);
    textVertices[listId].push_back(1.0f);
    textTextCoords[listId].push_back(quadTexCoords[4]);
    textTextCoords[listId].push_back(quadTexCoords[5]);

    // second triangle
    textVertices[listId].push_back(quadVertices[0]); // bottom left
    textVertices[listId].push_back(quadVertices[1]);
    textVertices[listId].push_back(0.0f);
    textVertices[listId].push_back(1.0f);
    textTextCoords[listId].push_back(quadTexCoords[0]);
    textTextCoords[listId].push_back(quadTexCoords[1]);
    textVertices[listId].push_back(quadVertices[4]); // top right
    textVertices[listId].push_back(quadVertices[5]);
    textVertices[listId].push_back(0.0f);
    textVertices[listId].push_back(1.0f);
    textTextCoords[listId].push_back(quadTexCoords[4]);
    textTextCoords[listId].push_back(quadTexCoords[5]);
    textVertices[listId].push_back(quadVertices[6]); // bottom right
    textVertices[listId].push_back(quadVertices[7]);
    textVertices[listId].push_back(0.0f);
    textVertices[listId].push_back(1.0f);
    textTextCoords[listId].push_back(quadTexCoords[6]);
    textTextCoords[listId].push_back(quadTexCoords[7]);
}


void FTTextureFontImpl::reset()
{
    for(size_t index = 0; index < textVertices.size(); ++index)
    {
        textVertices[index].clear();
    }
    for(size_t index = 0; index < textTextCoords.size(); ++index)
    {
        textTextCoords[index].clear();
    }
}


size_t FTTextureFontImpl::getTexturesCount()
{
    return textureIDList.size();
}


unsigned int FTTextureFontImpl::getTextureId(size_t index)
{
    if(index < textureIDList.size())
    {
        return textureIDList[index];
    }
    return 0;
}


const FTGL_FLOAT* FTTextureFontImpl::getVertices(size_t index, size_t* outElemSize, size_t* outElemCount)
{
    if(index < textVertices.size())
    {
        *outElemSize = 4;
        *outElemCount = textVertices[index].size() / 4;
        return &textVertices[index][0];
    }
    *outElemSize = 0;
    *outElemCount = 0;
    return NULL;
}


const FTGL_FLOAT* FTTextureFontImpl::getTexCoords(size_t index, size_t* outElemSize, size_t* outElemCount)
{
    if(index < textTextCoords.size())
    {
        *outElemSize = 2;
        *outElemCount = textTextCoords[index].size() / 2;
        return &textTextCoords[index][0];
    }
    *outElemSize = 0;
    *outElemCount = 0;
    return NULL;
}


FTGlyph* FTTextureFontImpl::MakeGlyphImpl(FT_GlyphSlot ftGlyph)
{
    glyphHeight = static_cast<int>(charSize.Height() + 0.5);
    glyphWidth = static_cast<int>(charSize.Width() + 0.5);

    if(glyphHeight < 1) glyphHeight = 1;
    if(glyphWidth < 1) glyphWidth = 1;

    if(textureIDList.empty())
    {
        textureIDList.push_back(CreateTexture());
        textVertices.push_back(FTVector<GLfloat>());
        textTextCoords.push_back(FTVector<GLfloat>());
        xOffset = yOffset = padding;
    }

    if(xOffset > (textureWidth - glyphWidth))
    {
        xOffset = padding;
        yOffset += glyphHeight;

        if(yOffset > (textureHeight - glyphHeight))
        {
            textureIDList.push_back(CreateTexture());
            textVertices.push_back(FTVector<GLfloat>());
            textTextCoords.push_back(FTVector<GLfloat>());
            yOffset = padding;
        }
    }

    FTTextureGlyph* tempGlyph = new FTTextureGlyph(ftGlyph,
                                                   this,
                                                   textureIDList[textureIDList.size() - 1],
                                                   xOffset, yOffset, textureWidth, textureHeight);
    xOffset += static_cast<int>(tempGlyph->BBox().Upper().X() - tempGlyph->BBox().Lower().X() + padding + 0.5);

    --remGlyphs;

    return tempGlyph;
}


void FTTextureFontImpl::CalculateTextureSize()
{
    if(!maximumGLTextureSize)
    {
        maximumGLTextureSize = 1024;
        glGetIntegerv(GL_MAX_TEXTURE_SIZE, (GLint*)&maximumGLTextureSize);
        assert(maximumGLTextureSize); // If you hit this then you have an invalid OpenGL context.
    }

    textureWidth = NextPowerOf2((remGlyphs * glyphWidth) + (padding * 2));
    textureWidth = textureWidth > maximumGLTextureSize ? maximumGLTextureSize : textureWidth;

    int h = static_cast<int>((textureWidth - (padding * 2)) / glyphWidth + 0.5);

    textureHeight = NextPowerOf2(((numGlyphs / h) + 1) * glyphHeight);
    textureHeight = textureHeight > maximumGLTextureSize ? maximumGLTextureSize : textureHeight;
}


unsigned int FTTextureFontImpl::CreateTexture()
{
    CalculateTextureSize();

    int totalMemory = textureWidth * textureHeight;
    unsigned char* textureMemory = new unsigned char[totalMemory];
    memset(textureMemory, 0, totalMemory);

    GLuint textID;
    glGenTextures(1, (GLuint*)&textID);

    glBindTexture(GL_TEXTURE_2D, textID);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

    glTexImage2D(GL_TEXTURE_2D, 0, GL_ALPHA, textureWidth, textureHeight,
                 0, GL_ALPHA, GL_UNSIGNED_BYTE, textureMemory);

    delete [] textureMemory;

    return textID;
}


bool FTTextureFontImpl::FaceSize(const unsigned int size, const unsigned int res)
{
    if(!textureIDList.empty() && (faceSize != size) && (faceRes != res))
    {
        /**
         * Textures deleting may result in cached data corruption!
         * Should never be called in dynamic!
         * User should create another instance of TextureFont if
         * need several sizes atonce.
         */
        glDeleteTextures((GLsizei)textureIDList.size(), (const GLuint*)&textureIDList[0]);
        textureIDList.clear();
        textVertices.clear();
        textTextCoords.clear();
        remGlyphs = numGlyphs = face.GlyphCount();
    }
    faceSize = size;
    faceRes = res;
    return FTFontImpl::FaceSize(size, res);
}


template <typename T>
inline FTPoint FTTextureFontImpl::RenderI(const T* string, const int len,
                                          FTPoint position, FTPoint spacing,
                                          int renderMode)
{
    glEnable(GL_TEXTURE_2D);
    FTPoint tmp = FTFontImpl::Render(string, len,
                                     position, spacing, renderMode);
    return tmp;
}


FTPoint FTTextureFontImpl::Render(const char * string, const int len,
                                  FTPoint position, FTPoint spacing,
                                  int renderMode)
{
    return RenderI(string, len, position, spacing, renderMode);
}


FTPoint FTTextureFontImpl::Render(const wchar_t * string, const int len,
                                  FTPoint position, FTPoint spacing,
                                  int renderMode)
{
    return RenderI(string, len, position, spacing, renderMode);
}

