/**
 * Copyright © 2007-2011 Kirill Gavrilov <kirill@sview.ru>
 *
 * StCore library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * StCore library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library.
 * If not, see <http://www.gnu.org/licenses/>.
 */

#include "StWindowImpl.h"

#include <StStrings/StLogger.h>

static const stUtf8_t WINDOW_TITLE_DEFAULT[] = "StWindow";

StWindowImpl::StWindowImpl()
: hMaster(),
  hSlave(),
  nativeParent(),
  windowName(WINDOW_TITLE_DEFAULT),
  lastError(STWIN_INITNOTSTART),
  evInitWin(false),
  evInitGl(false),
  mousePt(0.5, 0.5),
  rectNorm(128, 512, 128, 512),
  rectFull(128, 512, 128, 512),
  rectNormPrev(0, 1, 0, 1),
  monitors(),
  monMasterFullScr(-1),
  monSlave(),
  mySyncCounter(0),
  myWinOnMonitorId(0),
  dataMap(0),
  rendererPlugin(),
  rendererDeviceId(-1),
  targetFps(0.0),
#if (defined(_WIN32) || defined(__WIN32__))
  stMessageThread(NULL),
  hEventQuit(NULL),
  hEventCursorShow(NULL),
  hEventCursorHide(NULL),
#elif (defined(__linux__) || defined(__linux))
  toUseReparentHack(false),
#endif
#if (!defined(__APPLE__))
  event(),
#endif
  dragNDMutex(),
  dragNDCount(0),
  dndList(NULL),
  stMessageList(),
  isUpdate(false),
  stWinAttribs(stDefaultWinAttributes()) {
    //
    dndList = new StString[1];
    monSlave.idMaster = 0;
    monSlave.idSlave = 1;  // use second monitor for slave by default
    monSlave.xAdd = 1;
    monSlave.xSub = 0;
    monSlave.yAdd = 1;
    monSlave.ySub = 0;
    stMemSet(&nativeParent, 0, sizeof(StNativeWin_t));

#if(defined(_WIN32) || defined(__WIN32__))
    // we create Win32 event directly (not StEvent) to use it with MsgWaitForMultipleObjects()
    hEventQuit = CreateEvent(0, true, false, NULL);
    hEventCursorShow = CreateEvent(0, true, false, NULL);
    hEventCursorHide = CreateEvent(0, true, false, NULL);
#elif(defined(__linux__) || defined(__linux))
    toUseReparentHack = !StProcess::getEnv("ST_REPARENT_HACK").isEmpty();
#endif

    // just debug output Monitors' configuration
    monitors.init();
    for(size_t m = 0; m < monitors.size(); ++m) {
        ST_DEBUG_LOG(monitors[m].toString());
    }
}

StWindowImpl::~StWindowImpl() {
    close();
    delete[] dndList;
#if(defined(_WIN32) || defined(__WIN32__))
    CloseHandle(hEventQuit);
    CloseHandle(hEventCursorShow);
    CloseHandle(hEventCursorHide);
#endif
}

void StWindowImpl::close() {
    hide(ST_WIN_MASTER); hide(ST_WIN_SLAVE);
    stMessageList.append(StMessageList::MSG_EXIT);

    hSlave.close(); // close GL contexts
    hMaster.close();
#if(defined(_WIN32) || defined(__WIN32__))
    SetEvent(hEventQuit);
    size_t TIME_LIMIT = 60000;
    size_t timeWait = 10000;
    while(!hMaster.evMsgThread.wait(timeWait)) {
        ST_DEBUG_LOG("WinAPI, wait for Message thread to quit " + (timeWait / 1000) + " seconds!");
        timeWait += 10000;
        if(timeWait > TIME_LIMIT) {
            ST_DEBUG_LOG("WinAPI, Message thread was killed (timeout " + (TIME_LIMIT / 1000) + " seconds exceeded!)");
            stMessageThread->kill();
            break;
        }
    }
    delete stMessageThread; stMessageThread = NULL;
#endif
    stMemSet(&nativeParent, 0, sizeof(StNativeWin_t));
    stWinAttribs.isFullScreen = false; // just hack to return window position after closing
}

#if (!defined(__APPLE__))
void StWindowImpl::setTitle(const StString& theTitle) {
    windowName = theTitle;
#if(defined(_WIN32) || defined(__WIN32__))
    isUpdate = true;
#elif(defined(__linux__) || defined(__linux))
    if(hMaster.hWindow != 0){
        XTextProperty propertyTitle;
        propertyTitle.encoding = None;
        char* utitle = (char* )windowName.toCString();
        Xutf8TextListToTextProperty(hMaster.getDisplay(), &utitle, 1, XUTF8StringStyle,  &propertyTitle);
        XSetWMName(hMaster.getDisplay(), hMaster.hWindow, &propertyTitle);
        XSetWMProperties(hMaster.getDisplay(), hMaster.hWindow, &propertyTitle, &propertyTitle, NULL, 0, NULL, NULL, NULL);
    }
#endif
}
#endif // !__APPLE__

void StWindowImpl::getAttributes(StWinAttributes_t* inOutAttributes) {
    size_t bytesToCopy = (inOutAttributes->nSize > sizeof(StWinAttributes_t)) ? sizeof(StWinAttributes_t) : inOutAttributes->nSize;
    stMemCpy(inOutAttributes, &stWinAttribs, bytesToCopy); // copy as much as possible
    inOutAttributes->nSize = bytesToCopy;
}

void StWindowImpl::setAttributes(const StWinAttributes_t* inAttributes) {
    // TODO (Kirill Gavrilov#5#)
    size_t bytesToCopy = (inAttributes->nSize > sizeof(StWinAttributes_t)) ? sizeof(StWinAttributes_t) : inAttributes->nSize;
    stMemCpy(&stWinAttribs, inAttributes, bytesToCopy); // copy as much as possible
    stWinAttribs.nSize = sizeof(StWinAttributes_t);             // restore own size
    updateSlaveConfig();
    updateWindowPos();
}

#if (!defined(__APPLE__))
void StWindowImpl::show(const int& winNum) {
    if(winNum == ST_WIN_MASTER && stWinAttribs.isHide) {
    #if(defined(_WIN32) || defined(__WIN32__))
        if(hMaster.hWindow != NULL) {
            ShowWindow(hMaster.hWindow, SW_SHOW);
        } else if(hMaster.hWindowGl != NULL) {
            ShowWindow(hMaster.hWindowGl, SW_SHOW);
        }
    #elif(defined(__linux__) || defined(__linux))
        if(!hMaster.stXDisplay.isNull()) {
            if(hMaster.hWindow != 0) {
                XMapWindow(hMaster.getDisplay(), hMaster.hWindow);
            } else if(hMaster.hWindowGl != 0) {
                XMapWindow(hMaster.getDisplay(), hMaster.hWindowGl);
            }
        }
    #endif
        stWinAttribs.isHide = false;
        updateWindowPos();
    } else if(winNum == ST_WIN_SLAVE && stWinAttribs.isSlaveHide) {
    #if(defined(_WIN32) || defined(__WIN32__))
        if(hSlave.hWindowGl != NULL) {
            ShowWindow(hSlave.hWindowGl, SW_SHOW);
        }
    #elif(defined(__linux__) || defined(__linux))
        if(!hSlave.stXDisplay.isNull() && hSlave.hWindowGl != 0) {
            XMapWindow(hSlave.getDisplay(), hSlave.hWindowGl);
        }
    #endif
        stWinAttribs.isSlaveHide = false;
        updateWindowPos();
    }
}

void StWindowImpl::hide(const int& winNum) {
    if(winNum == ST_WIN_MASTER && !stWinAttribs.isHide) {
    #if(defined(_WIN32) || defined(__WIN32__))
        if(hMaster.hWindow != NULL) {
            ShowWindow(hMaster.hWindow, SW_HIDE);
        } else if(hMaster.hWindowGl != NULL) {
            ShowWindow(hMaster.hWindowGl, SW_HIDE);
        }
    #elif(defined(__linux__) || defined(__linux))
        if(!hMaster.stXDisplay.isNull()) {
            if(hMaster.hWindow != 0) {
                XUnmapWindow(hMaster.getDisplay(), hMaster.hWindow);
                isUpdate = true; // ?
            } else if(hMaster.hWindowGl != 0) {
                XUnmapWindow(hMaster.getDisplay(), hMaster.hWindowGl);
                isUpdate = true; // ?
            }
        }
    #endif
        stWinAttribs.isHide = true;
    } else if(winNum == ST_WIN_SLAVE && !stWinAttribs.isSlaveHide) {
    #if(defined(_WIN32) || defined(__WIN32__))
        if(hSlave.hWindowGl != NULL) {
            ShowWindow(hSlave.hWindowGl, SW_HIDE);
        }
    #elif(defined(__linux__) || defined(__linux))
        if(!hSlave.stXDisplay.isNull() && hSlave.hWindowGl != 0) {
            XUnmapWindow(hSlave.getDisplay(), hSlave.hWindowGl);
            isUpdate = true; // ?
        }
    #endif
        stWinAttribs.isSlaveHide = true;
    }
}
#endif // !__APPLE__

void StWindowImpl::showCursor(bool toShow) {
    if(stWinAttribs.isHideCursor != toShow) {
        return; // nothing to update
    }
#if(defined(_WIN32) || defined(__WIN32__))
    // native show / hide function should be called from
    // window-message thread -> we set events to do that
    if(toShow) {
        SetEvent(hEventCursorShow);
    } else {
        SetEvent(hEventCursorHide);
    }
#elif(defined(__linux__) || defined(__linux))
    if(toShow) {
        XUndefineCursor(hMaster.getDisplay(), hMaster.hWindowGl);
    } else {
        static const char noPixData[] = {0, 0, 0, 0, 0, 0, 0, 0};
        XColor black, dummy;
        Colormap cmap = DefaultColormap(hMaster.getDisplay(), DefaultScreen(hMaster.getDisplay()));
        XAllocNamedColor(hMaster.getDisplay(), cmap, "black", &black, &dummy);
        Pixmap noPix = XCreateBitmapFromData(hMaster.getDisplay(), hMaster.hWindowGl, noPixData, 8, 8);
        Cursor noPtr = XCreatePixmapCursor(hMaster.getDisplay(), noPix, noPix, &black, &black, 0, 0);
        XDefineCursor(hMaster.getDisplay(), hMaster.hWindowGl, noPtr);
        XFreeCursor(hMaster.getDisplay(), noPtr);
        if(noPix != None) {
            XFreePixmap(hMaster.getDisplay(), noPix);
        }
        XFreeColors(hMaster.getDisplay(), cmap, &black.pixel, 1, 0);
    }
#endif
    stWinAttribs.isHideCursor = !toShow;
}

#if (!defined(__APPLE__))
void StWindowImpl::setPlacement(const StRectI_t& rect) {
    rectNorm = rect;
    isUpdate = true;
#if(defined(_WIN32) || defined(__WIN32__))
    if(hMaster.hWindow != NULL && !stWinAttribs.isFullScreen) {
        int posLeft   = rectNorm.left() - (!stWinAttribs.isNoDecor ?  GetSystemMetrics(SM_CXSIZEFRAME) : 0);
        int posTop    = rectNorm.top()  - (!stWinAttribs.isNoDecor ? (GetSystemMetrics(SM_CYSIZEFRAME) + GetSystemMetrics(SM_CYCAPTION)) : 0);
        int winWidth  = (!stWinAttribs.isNoDecor ? 2 * GetSystemMetrics(SM_CXSIZEFRAME) : 0) + rectNorm.width();
        int winHeight = (!stWinAttribs.isNoDecor ? (GetSystemMetrics(SM_CYCAPTION) + 2 * GetSystemMetrics(SM_CYSIZEFRAME)) : 0) + rectNorm.height();
        SetWindowPos(hMaster.hWindow,     // window handle
                     HWND_NOTOPMOST,      // top status
                     posLeft, posTop,     // Left-Top position
                     winWidth, winHeight, // Width x Height
                     SWP_NOACTIVATE);
    }
#elif(defined(__linux__) || defined(__linux))
    if(!hMaster.stXDisplay.isNull() && !stWinAttribs.isFullScreen && hMaster.hWindow != 0) {
        XMoveResizeWindow(hMaster.getDisplay(), hMaster.hWindow,
                          rectNorm.left(),  rectNorm.top(),
                          rectNorm.width(), rectNorm.height());
        XFlush(hMaster.getDisplay());
    }
#endif
}
#endif // !__APPLE__

StRectI_t StWindowImpl::getPlacement() {
    if(stWinAttribs.isFullScreen) {
        return rectFull;
    } else {
        return rectNorm;
    }
}

int StWindowImpl::getMouseDown(StPointD_t* point) {
    return stMDownQueue.pop(*point);
}

int StWindowImpl::getMouseUp(StPointD_t* point) {
    return stMUpQueue.pop(*point);
}

StPointD_t StWindowImpl::getMousePos() {
    StRectI_t rectWindow = (stWinAttribs.isFullScreen) ? rectFull : rectNorm;
#if (defined(_WIN32) || defined(__WIN32__))
    if(hMaster.hWindowGl != NULL) {
        CURSORINFO aCursorInfo; //stMemSet(&aCursorInfo, 0, sizeof(aCursorInfo));
        aCursorInfo.cbSize = sizeof(aCursorInfo);
        if(GetCursorInfo(&aCursorInfo) != FALSE
        && ScreenToClient(hMaster.hWindowGl, &aCursorInfo.ptScreenPos) != FALSE) {
            return StPointD_t(double(aCursorInfo.ptScreenPos.x) / (double )rectWindow.width(),
                              double(aCursorInfo.ptScreenPos.y) / (double )rectWindow.height());
        }
    }
#elif (defined(__APPLE__))
    CGEventRef anEvent = CGEventCreate(NULL);
    CGPoint aCursor = CGEventGetLocation(anEvent);
    CFRelease(anEvent);
    return StPointD_t((aCursor.x - double(rectWindow.left())) / double(rectWindow.width()),
                      (aCursor.y - double(rectWindow.top()))  / double(rectWindow.height()));
#elif (defined(__linux__) || defined(__linux))
    if(hMaster.hWindowGl != 0) {
        Window childReturn;
        Window rootReturn = DefaultRootWindow(hMaster.getDisplay());
        int rootX = 0;
        int rootY = 0;
        int winX = 0;
        int winY = 0;
        unsigned int maskReturn;
        XQueryPointer(hMaster.getDisplay(), hMaster.hWindowGl,
                      &rootReturn, &childReturn,
                      &rootX, &rootY, &winX, &winY, &maskReturn);
        return StPointD_t((double )winX / (double )rectWindow.width(),
                          (double )winY / (double )rectWindow.height());
    }
#endif
    // undefined
    return StPointD_t(0.0, 0.0);
}

int StWindowImpl::getDragNDropFile(const int& theIndex, stUtf8_t* theOutFile, const size_t& theBuffSizeBytes) {
    dragNDMutex.lock();
    if(theIndex < 0 || theOutFile == NULL || theBuffSizeBytes == 0) {
        size_t aCount = dragNDCount;
        dragNDMutex.unlock();
        return (int )aCount; // returns files' count
    }
    if(theIndex >= (int )dragNDCount) {
        dragNDMutex.unlock();
        return -1; // returns error
    }
    size_t aSizeBytes = dndList[theIndex].getSize() + 1;
    if(aSizeBytes > theBuffSizeBytes) {
        dragNDMutex.unlock();
        return (int )aSizeBytes; // returns needed buffer size
    }
    stMemCpy(theOutFile, dndList[theIndex].toCString(), aSizeBytes);
    dragNDMutex.unlock();
    return 0; // returns success
}

// Change active GL context in current thread
void StWindowImpl::stglMakeCurrent(const int& winNum){
    switch(winNum) {
          case ST_WIN_MASTER: {
            hMaster.glMakeCurrent();
            break;
        } case ST_WIN_SLAVE: {
            hSlave.glMakeCurrent();
            break;
        }
    }
}

// Swap Buffers (Double Buffering)
void StWindowImpl::stglSwap(const int& winNum) {
    switch(winNum) {
        case ST_WIN_MASTER: {
            hMaster.glSwap();
            break;
        } case ST_WIN_SLAVE: {
            hSlave.glSwap();
            break;
        }
    }
}

stBool_t StWindowImpl::appendMessage(const StMessage_t& stMessage) {
    switch(stMessage.uin) {
        case StMessageList::MSG_MOUSE_DOWN_APPEND: {
            stMUpQueue.clear();
            stMDownQueue.clear();
            StMouseMessage_t* mouseData = (StMouseMessage_t* )stMessage.data;
            stMDownQueue.push(mouseData->point, mouseData->button);
            return stMessageList.append(StMessageList::MSG_MOUSE_DOWN);
        }
        case StMessageList::MSG_MOUSE_UP_APPEND: {
            StMouseMessage_t* mouseData = (StMouseMessage_t* )stMessage.data;
            stMUpQueue.push(mouseData->point, mouseData->button);
            return stMessageList.append(StMessageList::MSG_MOUSE_UP);
        }
        case StMessageList::MSG_KEY_DOWN_APPEND: {
            stMessageList.getKeysMap()[(size_t )stMessage.data] = true;
            return true;
        }
        case StMessageList::MSG_KEY_UP_APPEND: {
            stMessageList.getKeysMap()[(size_t )stMessage.data] = false;
            return true;
        }
        default: return stMessageList.append(stMessage);
    }
}

bool StWindowImpl::getValue(const size_t& key, size_t* value) {
    if(key == ST_WIN_DATAKEYS_RENDERER && dataMap != 0) {
        *value = dataMap;
        return true;
    }
    return false;
}

void StWindowImpl::setValue(const size_t& key, const size_t& value) {
    if(key == ST_WIN_DATAKEYS_RENDERER) {
        dataMap = value;
    } else {
        ST_DEBUG_LOG_AT("NOT IMPLEMENTED!");
    }
}

// Exported class-methods wpappers.
ST_EXPORT StWindowInterface* StWindow_new() {
    return new StWindowImpl();
}

ST_EXPORT void StWindow_del(StWindowInterface* inst) {
    delete (StWindowImpl* )inst;
}

ST_EXPORT void StWindow_close(StWindowInterface* inst) {
    ((StWindowImpl* )inst)->close();
}

ST_EXPORT void StWindow_setTitle(StWindowInterface* inst, const stUtf8_t* theTitle) {
    ((StWindowImpl* )inst)->setTitle(theTitle);
}

ST_EXPORT void StWindow_getAttributes(StWindowInterface* inst, StWinAttributes_t* inOutAttributes) {
    ((StWindowImpl* )inst)->getAttributes(inOutAttributes);
}

ST_EXPORT void StWindow_setAttributes(StWindowInterface* inst, const StWinAttributes_t* inAttributes) {
    ((StWindowImpl* )inst)->setAttributes(inAttributes);
}

ST_EXPORT stBool_t StWindow_isStereoOutput(StWindowInterface* inst) {
    return ((StWindowImpl* )inst)->isStereoOutput();
}

ST_EXPORT void StWindow_setStereoOutput(StWindowInterface* inst, stBool_t stereoState) {
    ((StWindowImpl* )inst)->setStereoOutput(stereoState);
}

ST_EXPORT void StWindow_show(StWindowInterface* inst, const int& win, const stBool_t& show) {
    if(show) {
        ((StWindowImpl* )inst)->show(win);
    } else {
        ((StWindowImpl* )inst)->hide(win);
    }
}

ST_EXPORT void StWindow_showCursor(StWindowInterface* inst, stBool_t toShow) {
    ((StWindowImpl* )inst)->showCursor(toShow);
}

ST_EXPORT stBool_t StWindow_isFullScreen(StWindowInterface* inst) {
    return ((StWindowImpl* )inst)->isFullScreen();
}

ST_EXPORT void StWindow_setFullScreen(StWindowInterface* inst, stBool_t fullscreen) {
    ((StWindowImpl* )inst)->setFullScreen(fullscreen);
}

ST_EXPORT void StWindow_getPlacement(StWindowInterface* inst, StRectI_t* rect) {
    *rect = ((StWindowImpl* )inst)->getPlacement();
}

ST_EXPORT void StWindow_setPlacement(StWindowInterface* inst, const StRectI_t* rect) {
    ((StWindowImpl* )inst)->setPlacement(*rect);
}

ST_EXPORT void StWindow_getMousePos(StWindowInterface* inst, StPointD_t* point) {
    StPointD_t p = ((StWindowImpl* )inst)->getMousePos();
    stMemCpy(point, &p, sizeof(StPointD_t));
}

ST_EXPORT int StWindow_getMouseDown(StWindowInterface* inst, StPointD_t* point) {
    return ((StWindowImpl* )inst)->getMouseDown(point);
}

ST_EXPORT int StWindow_getMouseUp(StWindowInterface* inst, StPointD_t* point) {
    return ((StWindowImpl* )inst)->getMouseUp(point);
}

ST_EXPORT int StWindow_getDragNDropFile(StWindowInterface* inst, const int& theIndex, stUtf8_t* theOutFile, const size_t& theBuffSizeBytes) {
    return ((StWindowImpl* )inst)->getDragNDropFile(theIndex, theOutFile, theBuffSizeBytes);
}

ST_EXPORT void StWindow_stglSwap(StWindowInterface* inst, const int& value) {
    ((StWindowImpl* )inst)->stglSwap(value);
}

ST_EXPORT void StWindow_stglMakeCurrent(StWindowInterface* inst, const int& value) {
    ((StWindowImpl* )inst)->stglMakeCurrent(value);
}

ST_EXPORT stBool_t StWindow_stglCreate(StWindowInterface* inst, const StWinAttributes_t* inAttributes, const StNativeWin_t* nativeParentWindow) {
    return ((StWindowImpl* )inst)->stglCreate(inAttributes, nativeParentWindow);
}

ST_EXPORT double StWindow_stglGetTargetFps(StWindowInterface* inst) {
    return ((StWindowImpl* )inst)->stglGetTargetFps();
}

ST_EXPORT void StWindow_stglSetTargetFps(StWindowInterface* inst, const double& fps) {
    ((StWindowImpl* )inst)->stglSetTargetFps(fps);
}

ST_EXPORT void StWindow_callback(StWindowInterface* inst, StMessage_t* stMessages) {
    ((StWindowImpl* )inst)->callback(stMessages);
}

ST_EXPORT stBool_t StWindow_appendMessage(StWindowInterface* inst, const StMessage_t& stMessage) {
    return ((StWindowImpl* )inst)->appendMessage(stMessage);
}

ST_EXPORT stBool_t StWindow_getValue(StWindowInterface* inst, const size_t& key, size_t* value) {
    return ((StWindowImpl* )inst)->getValue(key, value);
}

ST_EXPORT void StWindow_setValue(StWindowInterface* inst, const size_t& key, const size_t& value) {
    ((StWindowImpl* )inst)->setValue(key, value);
}

ST_EXPORT void* StWindow_memAlloc(const size_t& bytes) {
    return stMemAlloc(bytes);
}

ST_EXPORT void StWindow_memFree(void* ptr) {
    stMemFree(ptr);
}
