/**
 * Copyright © 2007-2011 Kirill Gavrilov <kirill@sview.ru>
 *
 * StCore library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * StCore library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library.
 * If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __StWindowImpl_h_
#define __StWindowImpl_h_

#include <StCore/StWindowInterface.h> // interface declaration
#include "StSearchMonitors.h"
#include "StWinHandles.h"             // os-specific handles class for windows
#include "StMouseClickQueue.h"

#if (defined(__APPLE__))
    #include <StCocoa/StCocoaCoords.h>
#endif

class NSOpenGLContext;
class StThread;

/**
 * This class represents implementation of
 * all main routines for GLwindow managment.
 */
class ST_LOCAL StWindowImpl : public StWindowInterface {

        private:

    StWinHandles           hMaster; // special handles
    StWinHandles            hSlave;
    StNativeWin_t     nativeParent; // native parent struct

    StString            windowName; // GUI window name
    int                  lastError; // last error code

    StEvent              evInitWin; // special event waited from createWindows() thread
    StEvent               evInitGl;

    StPointD_t             mousePt; // mouse coordinates on each iteration
    StRectI_t             rectNorm; // window's coordinates in normal state
    StRectI_t             rectFull; // window's coordinates in fullscreen state
    StRectI_t         rectNormPrev; // rectangle cache to track changes

    StSearchMonitors      monitors; // available monitors list
    int           monMasterFullScr; // if -1 - used current
    StSlaveWindowCfg_t    monSlave;
    size_t           mySyncCounter;
    int           myWinOnMonitorId; // monitor id where window is placed

    size_t                 dataMap;
    StString     rendererPlugin; // active StRendererPlugin's path
    int           rendererDeviceId;

    double               targetFps;

#if (defined(_WIN32) || defined(__WIN32__))
    POINT              myPointTest; // temporary point object to verify cached window position
    StThread*      stMessageThread;
    HANDLE              hEventQuit; // quit message thread event
    HANDLE        hEventCursorShow;
    HANDLE        hEventCursorHide;
    MSG                      event; // message for windows' message loop
#elif (defined(__APPLE__))
        public:
    StCocoaCoords    myCocoaCoords;
#elif (defined(__linux__) || defined(__linux))
    bool         toUseReparentHack; // hack variable ST_REPARENT_HACK
    XEvent                   event;
#endif

    StMutex            dragNDMutex; // access mustex for Drag&Drop file list
    size_t             dragNDCount; // files' count
    StString*           dndList; // Drag&Drop list

    StMouseClickQueue stMDownQueue;
    StMouseClickQueue   stMUpQueue;
    StMessageList    stMessageList; // callback list
    bool                  isUpdate; // helper flag on window movements updates

    StWinAttributes_t stWinAttribs;

    void updateWindowPos();
#if (defined(__linux__) || defined(__linux))
    void parseXDNDClientMsg();
    void parseXDNDSelectionMsg();
#endif

    void updateSlaveConfig() {
        monSlave.idSlave = int(stWinAttribs.slaveMonId);
        if(stWinAttribs.isSlaveXMirrow) {
            monSlave.xAdd = 0; monSlave.xSub = 1;
            monSlave.yAdd = 1; monSlave.ySub = 0;
        } else if(stWinAttribs.isSlaveYMirrow) {
            monSlave.xAdd = 1; monSlave.xSub = 0;
            monSlave.yAdd = 0; monSlave.ySub = 1;
        } else {
            monSlave.xAdd = 1; monSlave.xSub = 0;
            monSlave.yAdd = 1; monSlave.ySub = 0;
        }
    }

    int getMasterLeft() {
        return monitors[monSlave.idMaster].getVRect().left();
    }

    int getMasterTop() {
        return monitors[monSlave.idMaster].getVRect().top();
    }

    /**
     * @return true if slave window should be displayed on independent monitor.
     */
    bool isSlaveIndependent() {
        return !stWinAttribs.isSlaveHLineTop && !stWinAttribs.isSlaveHTop2Px && !stWinAttribs.isSlaveHLineBottom;
    }

    int getSlaveLeft() {
        if(!isSlaveIndependent()) {
            return monitors[getPlacement().center()].getVRect().left();
        } else if(stWinAttribs.isFullScreen) {
            return monitors[monSlave.idSlave].getVRect().left();
        } else {
            StMonitor& stMonMaster = monitors[getPlacement().center()]; // detect from current location
            return monSlave.xAdd * (monitors[monSlave.idSlave].getVRect().left()  + rectNorm.left()  - stMonMaster.getVRect().left())
                 + monSlave.xSub * (monitors[monSlave.idSlave].getVRect().right() - rectNorm.right() + stMonMaster.getVRect().left());
        }
    }

    int getSlaveWidth() {
        if(stWinAttribs.isSlaveHTop2Px) {
            return 2;
        } else if(stWinAttribs.isSlaveHLineTop || stWinAttribs.isSlaveHLineBottom) {
            return monitors[getPlacement().center()].getVRect().width();
        } else if(stWinAttribs.isFullScreen) {
            return monitors[monSlave.idSlave].getVRect().width();
        } else {
            return rectNorm.width();
        }
    }

    int getSlaveTop() {
        if(stWinAttribs.isSlaveHLineBottom) {
            return monitors[getPlacement().center()].getVRect().bottom() - 1;
        } else if(stWinAttribs.isSlaveHLineTop || stWinAttribs.isSlaveHTop2Px) {
            return monitors[getPlacement().center()].getVRect().top();
        } else if(stWinAttribs.isFullScreen) {
            return monitors[monSlave.idSlave].getVRect().top();
        } else {
            StMonitor& stMonMaster = monitors[getPlacement().center()]; // detect from current location
            return monSlave.yAdd * (monitors[monSlave.idSlave].getVRect().top()    + rectNorm.top()    - stMonMaster.getVRect().top())
                 + monSlave.ySub * (monitors[monSlave.idSlave].getVRect().bottom() - rectNorm.bottom() + stMonMaster.getVRect().top());
        }
    }

    int getSlaveHeight() {
        if(stWinAttribs.isSlaveHLineBottom || stWinAttribs.isSlaveHTop2Px) {
            return 1;
        } else if(stWinAttribs.isSlaveHLineTop) {
            return 10;
        } else if(stWinAttribs.isFullScreen) {
            return monitors[monSlave.idSlave].getVRect().height();
        } else {
            return rectNorm.height();
        }
    }

        public:

    // exported methods
    StWindowInterface* getLibImpl() { return this; }
    StWindowImpl();
    ~StWindowImpl();
    void close();
    void setTitle(const StString& title);
    void getAttributes(StWinAttributes_t* inOutAttributes);
    void setAttributes(const StWinAttributes_t* inAttributes);
    bool isStereoOutput() { return stWinAttribs.isStereoOutput; }
    void setStereoOutput(bool stereoState) { stWinAttribs.isStereoOutput = stereoState; }
    void show(const int& );
    void hide(const int& );
    void showCursor(bool toShow);
    bool isFullScreen() { return stWinAttribs.isFullScreen; }
    void setFullScreen(bool fullscreen);
    StRectI_t getPlacement();
    void setPlacement(const StRectI_t& rect);
    StPointD_t getMousePos();
    int getMouseDown(StPointD_t* point);
    int getMouseUp(StPointD_t* point);
    int getDragNDropFile(const int& id, stUtf8_t* outFile, const size_t& buffSizeBytes);
    bool stglCreate(const StWinAttributes_t* inAttributes, const StNativeWin_t* nativeParentWindow);
    void stglSwap(const int& winNum);
    void stglMakeCurrent(const int& winNum);
    double stglGetTargetFps() { return targetFps; }
    void stglSetTargetFps(const double& fps) { targetFps = fps; }
    void callback(StMessage_t* stMessages);
    stBool_t appendMessage(const StMessage_t& stMessage);
    bool getValue(const size_t& key, size_t* value);
    void setValue(const size_t& key, const size_t& value);

    // additional
    void updateChildRect();
#if (defined(_WIN32) || defined(__WIN32__))
    bool wndRegisterClass(const StStringUtfWide& theClassName);
    bool wndCreateWindows(); // called from non-main thread
    LRESULT stWndProc(HWND , UINT , WPARAM , LPARAM );
#elif (defined(__APPLE__))
    void doCreateWindows(NSOpenGLContext* theGLContextMaster,
                         NSOpenGLContext* theGLContextSlave);
#endif

};

#endif //__StWindowImpl_h_
