/**
 * Copyright © 2007-2011 Kirill Gavrilov <kirill@sview.ru>
 *
 * StCore library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * StCore library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library.
 * If not, see <http://www.gnu.org/licenses/>.
 */

#if(defined(__linux__) || defined(__linux))

#include "StWindowImpl.h"
#include "stvkeysxarray.h" // X keys to VKEYs lookup array

#include <StStrings/StLogger.h>

#include <X11/extensions/Xrandr.h>
#include <gdk/gdkx.h>
#include <math.h>

static int dblBuff[] = {
    GLX_RGBA,
    GLX_DEPTH_SIZE, 16,
    GLX_DOUBLEBUFFER,
    None
};

static int quadBuff[] = {
    GLX_RGBA,
    GLX_DEPTH_SIZE, 16,
    GLX_DOUBLEBUFFER,
    GLX_STEREO,
    None
};

/**
 * Our own XError handle function.
 * Default handlers just show the error and exit application immediately.
 * This is very horrible for application, because many errors not critical and could be ignored.
 * The main target for this function creation - prevent sView to quit after
 * multiple BadMatch errors happens on fullscreen->windowed switching
 * (looking like a vendors OpenGL driver bug).
 * Thus, now we just show up the error description and just ignore it - hopes we will well.
 * Function behaviour could be extended in future.
 */
int stXErrorHandler(Display* hDisplay, XErrorEvent* xErrorEvent) {
    char buffer[4096];
    XGetErrorText(hDisplay, xErrorEvent->error_code, buffer, 4096);
    ST_DEBUG_LOG("XError happend: " + buffer + "; ignored");
    // have no idea WHAT we should return here....
    return 0;
}

// function create GUI window
bool StWindowImpl::stglCreate(const StWinAttributes_t* inAttributes, const StNativeWin_t* nativeParentWindow) {
    if(nativeParentWindow != NULL) {
        stMemCpy(&(this->nativeParent), nativeParentWindow, sizeof(StNativeWin_t));
        nativeParent.stWinPtr = this;
    }

    // initialize helper GDK
    static bool isGdkInitialized = false;
    if(!isGdkInitialized) {
        if(!gdk_init_check(NULL, NULL)) {
            stError("GDK, init failed");
            //lastError =
            return false;
        }
        gdk_rgb_init(); // only guess sets up the true colour colour map
        isGdkInitialized = true;
    }

    size_t bytesToCopy = (inAttributes->nSize > sizeof(StWinAttributes_t)) ? sizeof(StWinAttributes_t) : inAttributes->nSize;
    stMemCpy(&stWinAttribs, inAttributes, bytesToCopy); // copy as much as possible
    stWinAttribs.nSize = sizeof(StWinAttributes_t);             // restore own size
    updateSlaveConfig();

    // replace default XError handler to ignore some errors
    /// TODO (Kirill Gavrilov#1) - GTK+ (re)initialization seems to be override our error handler!
    XSetErrorHandler(stXErrorHandler);

    lastError = STWIN_INITNOTSTART;
    // X-server implementation
    // create window on unix systems throw X-server
    XSetWindowAttributes swa;
    int                dummy;

    // open a connection to the X server
    StXDisplayH stXDisplay = new StXDisplay();
    if(!stXDisplay->isOpened()) {
        stXDisplay.nullify();
        stError("X, could not open display");
        lastError = STWIN_ERROR_X_OPENDISPLAY;
        return false;
    }
    hMaster.stXDisplay = stXDisplay;
    Display* hDisplay = stXDisplay->hDisplay;

    // make sure OpenGL's GLX extension supported
    if(!glXQueryExtension(hDisplay, &dummy, &dummy)) {
        hMaster.close();
        stError("X, server has no OpenGL GLX extension");
        lastError = STWIN_ERROR_X_NOGLX;
        return false;
    }

    if(stWinAttribs.isGlStereo) {
        // find an appropriate visual
        stXDisplay->hVisInfo = glXChooseVisual(hDisplay, DefaultScreen(hDisplay), quadBuff);
        if(stXDisplay->hVisInfo == NULL) {
            stError("X, no Quad Buffered visual");
            stXDisplay->hVisInfo = glXChooseVisual(hDisplay, DefaultScreen(hDisplay), dblBuff);
            if(stXDisplay->hVisInfo == NULL) {
                hMaster.close();
                stError("X, no RGB visual with depth buffer");
                lastError = STWIN_ERROR_X_NORGB;
                return false;
            }
        }
    } else {
        // find an appropriate visual
        // find an OpenGL-capable RGB visual with depth buffer
        stXDisplay->hVisInfo = glXChooseVisual(hDisplay, DefaultScreen(hDisplay), dblBuff);
        if(stXDisplay->hVisInfo == NULL) {
            hMaster.close();
            stError("X, no RGB visual with depth buffer");
            lastError = STWIN_ERROR_X_NORGB;
            return false;
        }
    }
    if(stWinAttribs.isSlave) {
        // just copy handle
        hSlave.stXDisplay = stXDisplay;
    }

    // create an X window with the selected visual
    // create an X colormap since probably not using default visual
    swa.colormap = XCreateColormap(hDisplay,
                                   RootWindow(hDisplay, stXDisplay->getScreen()),
                                   stXDisplay->getVisual(), AllocNone);
    swa.border_pixel = 0;

    // what events we want to recive:
    swa.event_mask =  KeyPressMask | KeyReleaseMask       // receive keyboard events
                    | ButtonPressMask | ButtonReleaseMask // receive mouse events
                    | StructureNotifyMask;                // receive ConfigureNotify event on resize and move
                    //| ResizeRedirectMask                // receive ResizeRequest event on resize (instead of common ConfigureNotify)
                    //| ExposureMask
                    //|EnterWindowMask|LeaveWindowMask
                    //|PointerMotionMask|PointerMotionHintMask|Button1MotionMask|Button2MotionMask|Button3MotionMask|Button4MotionMask|Button5MotionMask|ButtonMotionMask
                    //|KeymapStateMask|ExposureMask|VisibilityChangeMask
                    //|SubstructureNotifyMask|SubstructureRedirectMask
                    //|FocusChangeMask|PropertyChangeMask|ColormapChangeMask|OwnerGrabButtonMask

    updateChildRect();

    Window aParentWin = (Window )nativeParent.winHandle;
    if(aParentWin == 0 && !stWinAttribs.isNoDecor) {
        swa.override_redirect = False;
        hMaster.hWindow = XCreateWindow(hDisplay,
            stXDisplay->getRootWindow(),          // parent window
            rectNorm.left(),  rectNorm.top(),     // top left coordinates (pixels)
            rectNorm.width(), rectNorm.height(),  // width x height (in pixels)
            0,                                    // int border width (in pixels)
            stXDisplay->getDepth(),               // window depth (means top-bottom??)
            InputOutput,                          // unsigned int class - window's class
            stXDisplay->getVisual(),
            CWBorderPixel | CWColormap | CWEventMask | CWOverrideRedirect,  // unsigned long valuemask
            &swa);                                // attributes

        if(hMaster.hWindow == 0) {
            hMaster.close();
            stError("X, XCreateWindow failed for Master");
            lastError = STWIN_ERROR_X_CREATEWIN;
            return false;
        }
        aParentWin = hMaster.hWindow;

        XSetStandardProperties(hDisplay,
                               hMaster.hWindow,        // Window window - window handle
                               windowName.toCString(), // window name in title?
                               windowName.toCString(), // window name in task bar?
                               None,                   // Pixmap icon_pixmap - bitmap that is to be used for the icon or None
                               NULL, 0,                // char** argv, int argc
                               NULL);                  // XSizeHints* hints - size hints for the window in its normal state
    }

    swa.override_redirect = True; // GL window always undecorated
    hMaster.hWindowGl = XCreateWindow(hDisplay,
        (aParentWin != 0) ? aParentWin : stXDisplay->getRootWindow(), // parent window
        0, 0,                                // top left coordinates (pixels)
        rectNorm.width(), rectNorm.height(), // width x height (in pixels)
        0,                                   // int border width (in pixels)
        stXDisplay->getDepth(),              // window depth (means top-bottom??)
        InputOutput,                         // unsigned int class - window's class
        stXDisplay->getVisual(),
        CWBorderPixel | CWColormap | CWEventMask | CWOverrideRedirect,  // unsigned long valuemask
        &swa);                               // attributes

    if(hMaster.hWindowGl == 0) {
        hMaster.close();
        stError("X, XCreateWindow failed for Master");
        lastError = STWIN_ERROR_X_CREATEWIN;
        return false;
    }

    XSetStandardProperties(hDisplay,
                           hMaster.hWindowGl,   // Window window - window handle
                           "master window",     // window name in title?
                           "master window",     // window name in task bar?
                           None,                // Pixmap icon_pixmap - bitmap that is to be used for the icon or None
                           NULL, 0,             // char** argv, int argc
                           NULL);               // XSizeHints* hints - size hints for the window in its normal state

    if(stWinAttribs.isSlave) {
        swa.event_mask = NoEventMask; // we do not parse any events to slave window!
        swa.override_redirect = True; // slave window always undecorated
        hSlave.hWindowGl = XCreateWindow(hDisplay,
                                       stXDisplay->getRootWindow(),       // parent window
                                       getSlaveLeft(), getSlaveTop(),
                                       getSlaveWidth(), getSlaveHeight(),
                                       0,                                 // border width (in pixels)
                                       stXDisplay->getDepth(),            // window depth (means top-bottom??)
                                       InputOutput,                       // window's class
                                       stXDisplay->getVisual(),           // visual type
                                       CWBorderPixel | CWColormap | CWEventMask | CWOverrideRedirect, // valuemask
                                       &swa);                             // attributes

        if(hSlave.hWindowGl == 0) {
            hMaster.close();
            hSlave.close();
            stError("X, XCreateWindow failed for Slave");
            lastError = STWIN_ERROR_X_CREATEWIN;
            return false;
        }

        XSetStandardProperties(hDisplay,
                               hSlave.hWindowGl, // Window window - window handle
                               "slave window",
                               "slave window",
                               None,           // Pixmap icon_pixmap - bitmap that is to be used for the icon or None
                               NULL, 0,        // char** argv, int argc
                               NULL);          // XSizeHints* hints - size hints for the window in its normal state
    }

    int isGlCtx = hMaster.glCreateContext(stWinAttribs.isSlave ? &hSlave : NULL, stWinAttribs.isGlStereo);
    if(isGlCtx != STWIN_INIT_SUCCESS) {
        hMaster.close();
        hSlave.close();
        lastError = isGlCtx;
        return false;
    }

    // handle close window event
    if(hMaster.hWindow != 0) {
        XSetWMProtocols(hDisplay, hMaster.hWindow, &(stXDisplay->wndDestroyAtom), 1);
    }

    // Announce XDND support
    Atom version = 5;
    XChangeProperty(hDisplay, hMaster.hWindowGl, stXDisplay->xDNDAware, XA_ATOM, 32, PropModeReplace, (unsigned char* )&version, 1);
    if(hMaster.hWindow != 0) {
        XChangeProperty(hDisplay, hMaster.hWindow, stXDisplay->xDNDAware, XA_ATOM, 32, PropModeReplace, (unsigned char* )&version, 1);
    }

    // Initialize XRandr events reception
    if(XRRQueryExtension(hDisplay, &hMaster.xrandrEventBase, &dummy)) {
        XRRSelectInput(hDisplay,
                       stXDisplay->getRootWindow(),
                       RRScreenChangeNotifyMask |
                       RRCrtcChangeNotifyMask   |
                       RROutputPropertyNotifyMask);
        hMaster.isRecXRandrEvents = true;
    } else {
        hMaster.isRecXRandrEvents = false;
    }

    // request the X window to be displayed on the screen
    if(stWinAttribs.isSlave) {

        // request the X window to be displayed on the screen
        if(!stWinAttribs.isSlaveHide && (!isSlaveIndependent() || monitors.size() > 1)) {
            XMapWindow(hDisplay, hSlave.hWindowGl);
        }
        // always hise mouse cursor on slave window
        const char noPixData[] = {0, 0, 0, 0, 0, 0, 0, 0};
        XColor black, dummy;
        Colormap cmap = DefaultColormap(hDisplay, DefaultScreen(hDisplay));
        XAllocNamedColor(hDisplay, cmap, "black", &black, &dummy);
        Pixmap noPix = XCreateBitmapFromData(hDisplay, hSlave.hWindowGl, noPixData, 8, 8);
        Cursor noPtr = XCreatePixmapCursor(hDisplay, noPix, noPix, &black, &black, 0, 0);
        XDefineCursor(hDisplay, hSlave.hWindowGl, noPtr);
        XFreeCursor(hDisplay, noPtr);
        if(noPix != None) {
            XFreePixmap(hDisplay, noPix);
        }
        XFreeColors(hDisplay, cmap, &black.pixel, 1, 0);
    }
    if(!stWinAttribs.isHide) {
        if(hMaster.hWindow != 0) {
            XMapWindow(hDisplay, hMaster.hWindow);
        }
        XMapWindow(hDisplay, hMaster.hWindowGl);
    }

    // we need this call to go around bugs
    if(!stWinAttribs.isFullScreen && hMaster.hWindow != 0) {
        XMoveResizeWindow(hDisplay, hMaster.hWindow,
                          rectNorm.left(),  rectNorm.top(),
                          rectNorm.width(), rectNorm.height());
    }
    // flushes the output buffer, most client apps needn't use this cause buffer is automatically flushed as needed by calls to XNextEvent()...
    XFlush(hDisplay);
    isUpdate = true;
    lastError = STWIN_INIT_SUCCESS;
    return true;
}

/**
 * Update StWindow position according to native parent position.
 */
void StWindowImpl::updateChildRect() {
    if(!stWinAttribs.isFullScreen && (Window )nativeParent.winHandle != 0 && !hMaster.stXDisplay.isNull()) {
        Display* hDisplay = hMaster.stXDisplay->hDisplay;
        Window dummyWin;
        int xReturn, yReturn;
        unsigned int widthReturn, heightReturn, uDummy;
        XGetGeometry(hDisplay, (Window )nativeParent.winHandle, &dummyWin,
                     &xReturn, &yReturn, &widthReturn, &heightReturn,
                     &uDummy, &uDummy);
        XTranslateCoordinates(hDisplay, (Window )nativeParent.winHandle,
                              hMaster.stXDisplay->getRootWindow(),
                              0, 0, &rectNorm.left(), &rectNorm.top(), &dummyWin);
        rectNorm.right()  = rectNorm.left() + widthReturn;
        rectNorm.bottom() = rectNorm.top() + heightReturn;

        // update only when changes
        if(rectNorm != rectNormPrev) {
            rectNormPrev = rectNorm;
            isUpdate = true;
            stMessageList.append(StMessageList::MSG_RESIZE);

            /// TODO (Kirill Gavrilov#4) remove this workaround!
            // call this to be sure master window showed well
            // this useful only when compiz on - seems to be bug in it!
            if(toUseReparentHack && hMaster.hWindowGl != 0) {
                XReparentWindow(hDisplay, hMaster.hWindowGl, (Window )nativeParent.winHandle, 0, 0);
            }
        }
    }
}

void StWindowImpl::setFullScreen(bool fullscreen) {
    stWinAttribs.isFullScreen = fullscreen;
    if(stWinAttribs.isHide) {
        // TODO (Kirill Gavrilov#9) parse correctly
        // do nothing, just set the flag
        return;
    } else if(hMaster.stXDisplay.isNull()) {
        return;
    }

    Display* hDisplay = hMaster.stXDisplay->hDisplay;
    if(stWinAttribs.isFullScreen) {
        const StMonitor& stMon = (monMasterFullScr == -1) ? monitors[rectNorm.center()] : monitors[monMasterFullScr];
        rectFull = stMon.getVRect();
        XUnmapWindow(hDisplay, hMaster.hWindowGl); // workaround for strange bugs

        if((Window )nativeParent.winHandle != 0 || hMaster.hWindow != 0) {
            XReparentWindow(hDisplay, hMaster.hWindowGl, hMaster.stXDisplay->getRootWindow(), 0, 0);
            XMoveResizeWindow(hDisplay, hMaster.hWindowGl,
                              rectFull.left(),  rectFull.top(),
                              rectFull.width(), rectFull.height());
            XFlush(hDisplay);
            XMapWindow(hDisplay, hMaster.hWindowGl);
        } else {
            XMoveResizeWindow(hDisplay, hMaster.hWindowGl,
                              rectFull.left(),  rectFull.top(),
                              rectFull.width(), rectFull.height());
            XFlush(hDisplay);
            XMapWindow(hDisplay, hMaster.hWindowGl);
        }

        if(stWinAttribs.isSlave  && (!isSlaveIndependent() || monitors.size() > 1)) {
            XMoveResizeWindow(hDisplay, hSlave.hWindowGl,
                              getSlaveLeft(), getSlaveTop(),
                              getSlaveWidth(), getSlaveHeight());
        }
    } else {
        Window aParent = ((Window )nativeParent.winHandle != 0) ? (Window )nativeParent.winHandle : hMaster.hWindow;
        if(aParent != 0) {
            XReparentWindow(hDisplay, hMaster.hWindowGl, aParent, 0, 0);
            isUpdate = true;
        } else {
            XUnmapWindow(hDisplay, hMaster.hWindowGl); // workaround for strange bugs
            XResizeWindow(hDisplay, hMaster.hWindowGl, 256, 256);
            if(stWinAttribs.isSlave && (!isSlaveIndependent() || monitors.size() > 1)) {
                XUnmapWindow(hDisplay, hSlave.hWindowGl);
                XResizeWindow(hDisplay, hSlave.hWindowGl, 256, 256);
            }
            XFlush(hDisplay);
            XMapWindow(hDisplay, hMaster.hWindowGl);
            if(stWinAttribs.isSlave && (!isSlaveIndependent() || monitors.size() > 1)) {
                XMapWindow(hDisplay, hSlave.hWindowGl);
            }
            XFlush(hDisplay);
            XMoveResizeWindow(hDisplay, hMaster.hWindowGl,
                              rectNorm.left(),  rectNorm.top(),
                              rectNorm.width(), rectNorm.height());
        }
    }
    XSetInputFocus(hDisplay, hMaster.hWindowGl, RevertToParent, CurrentTime);

    stMessageList.append(StMessageList::MSG_RESIZE); // add event to update GL rendering scape
    stMessageList.append(StMessageList::MSG_FULLSCREEN_SWITCH);
    // flushes the output buffer, most client apps needn't use this cause buffer is automatically flushed as needed by calls to XNextEvent()...
    XFlush(hDisplay);
}

void StWindowImpl::parseXDNDClientMsg() {
    const StXDisplayH& stXDisplay = hMaster.stXDisplay;
    // hMaster.hWindow or hMaster.hWindowGl
    Window aWinReciever = ((XClientMessageEvent* )&event)->window;
    if(event.xclient.message_type == stXDisplay->xDNDEnter) {
        hMaster.xDNDVersion = (event.xclient.data.l[1] >> 24);
        bool moreThan3 = event.xclient.data.l[1] & 1;
        Window hSrcWin = event.xclient.data.l[0];
        /*ST_DEBUG_LOG(
            "Xdnd: Source window = 0x" + (int )event.xclient.data.l[0] + "\n"
            + "Supports > 3 types = " + (more_than_3) + "\n"
            + "Protocol version = " + hMaster.xDNDVersion + "\n"
            + "Type 1 = " + hMaster.getAtomName(event.xclient.data.l[2]) + "\n"
            + "Type 2 = " + hMaster.getAtomName(event.xclient.data.l[3]) + "\n"
            + "Type 3 = " + hMaster.getAtomName(event.xclient.data.l[4]) + "\n"
        );*/
        if(moreThan3) {
            Property prop = stXDisplay->readProperty(hSrcWin, stXDisplay->xDNDTypeList);
            Atom* atomList = (Atom* )prop.data;
            for(int i = 0; i < prop.nitems; ++i) {
                if(atomList[i] == stXDisplay->xDNDPlainText) {
                    hMaster.xDNDRequestType = stXDisplay->xDNDPlainText;
                    break;
                }
            }
            XFree(prop.data);
        } else if((Atom )event.xclient.data.l[2] == stXDisplay->xDNDPlainText
               || (Atom )event.xclient.data.l[3] == stXDisplay->xDNDPlainText
               || (Atom )event.xclient.data.l[4] == stXDisplay->xDNDPlainText) {
            hMaster.xDNDRequestType = stXDisplay->xDNDPlainText;
        } else {
            hMaster.xDNDRequestType = XA_STRING;
        }
    } else if(event.xclient.message_type == stXDisplay->xDNDPosition) {
        /*ST_DEBUG_LOG(
            "Source window = 0x" + (int )event.xclient.data.l[0] + "\n"
            + "Position: x=" + (int )(event.xclient.data.l[2] >> 16) + " y=" + (int )(event.xclient.data.l[2] & 0xffff)  + "\n"
            + "Timestamp = " + (int )event.xclient.data.l[3] + " (Version >= 1 only)\n"
        );*/

        // Xdnd: reply with an XDND status message
        XClientMessageEvent m;
        stMemSet(&m, 0, sizeof(m));
        m.type = ClientMessage;
        m.display = event.xclient.display;
        m.window  = event.xclient.data.l[0];
        m.message_type = stXDisplay->xDNDStatus;
        m.format = 32;
        m.data.l[0] = aWinReciever;
        m.data.l[1] = True; //
        m.data.l[2] = 0;    // specify an empty rectangle
        m.data.l[3] = 0;
        m.data.l[4] = stXDisplay->xDNDActionCopy; // we only accept copying anyway.

        XSendEvent(stXDisplay->hDisplay, event.xclient.data.l[0], False, NoEventMask, (XEvent* )&m);
        XFlush(stXDisplay->hDisplay);
    } else if(event.xclient.message_type == stXDisplay->xDNDLeave) {
        //
    } else if(event.xclient.message_type == stXDisplay->xDNDDrop) {
        hMaster.xDNDSrcWindow = event.xclient.data.l[0];
        Atom sel = XInternAtom(stXDisplay->hDisplay, "PRIMARY", 0);
        if(hMaster.xDNDVersion >= 1) {
            XConvertSelection(stXDisplay->hDisplay, stXDisplay->xDNDSelection, hMaster.xDNDRequestType,
                              sel, aWinReciever, event.xclient.data.l[2]);
        } else {
            XConvertSelection(stXDisplay->hDisplay, stXDisplay->xDNDSelection, hMaster.xDNDRequestType,
                              sel, aWinReciever, CurrentTime);
        }
    }
}

void StWindowImpl::parseXDNDSelectionMsg() {
    Atom target = event.xselection.target;
    const StXDisplayH& stXDisplay = hMaster.stXDisplay;
    // hMaster.hWindow or hMaster.hWindowGl
    Window aWinReciever = ((XClientMessageEvent* )&event)->window;
    /*ST_DEBUG_LOG(
        "A selection notify has arrived!\n"
        + "Requestor = 0x" + (int )event.xselectionrequest.requestor + "\n"
        + "Selection atom = " + hMaster.getAtomName(event.xselection.selection) + "\n"
        + "Target atom    = " + hMaster.getAtomName(target) + "\n"
        + "Property atom  = " + hMaster.getAtomName(event.xselection.property) + "\n"
    );*/
    if(event.xselection.property == None) {
        return;
    } else {
        Atom sel = XInternAtom(stXDisplay->hDisplay, "PRIMARY", 0);
        Property prop = stXDisplay->readProperty(aWinReciever, sel);
        //If we're being given a list of targets (possible conversions)
        if(target == stXDisplay->XA_TARGETS) {
            XConvertSelection(stXDisplay->hDisplay, sel, XA_STRING, sel, aWinReciever, CurrentTime);
        } else if(target == hMaster.xDNDRequestType) {
            StString aData = StString((char* )prop.data);

            dragNDMutex.lock();
            dragNDCount = 1;
            delete[] dndList;
            dndList = new StString[1];

            size_t anEndChar = aData.getLength();
            for(StUtf8Iter anIter = aData.iterator(); *anIter != 0; ++anIter) {
                // cut only first filename, separated with CR/LF
                if(*anIter == stUtf32_t('\n') || *anIter == stUtf32_t(13)) {
                    anEndChar = anIter.getIndex();
                    break;
                }
            }

            // TODO (Kirill Gavrilov#5#) improve functionality
            // notes:
            // 1) dragndrop can move not only strings
            // 2) filename converted to links (maybe non-local!)
            const StString ST_FILE_PROTOCOL("file://");
            size_t aCutFrom = aData.isStartsWith(ST_FILE_PROTOCOL) ? ST_FILE_PROTOCOL.getLength() : 0;
            aData = aData.subString(aCutFrom, anEndChar);
            if(hMaster.xDNDRequestType != XA_STRING) {
                dndList[0].fromUrl(aData);
            } else {
                dndList[0] = aData;
            }
            dragNDMutex.unlock();
            stMessageList.append(StMessageList::MSG_DRAGNDROP_IN);
            //ST_DEBUG_LOG(data);

            // Reply OK
            XClientMessageEvent m;
            stMemSet(&m, 0, sizeof(m));
            m.type = ClientMessage;
            m.display = stXDisplay->hDisplay;
            m.window = hMaster.xDNDSrcWindow;
            m.message_type = stXDisplay->xDNDFinished;
            m.format = 32;
            m.data.l[0] = aWinReciever;
            m.data.l[1] = 1;
            m.data.l[2] = stXDisplay->xDNDActionCopy; //We only ever copy.

            // Reply that all is well.
            XSendEvent(stXDisplay->hDisplay, hMaster.xDNDSrcWindow, False, NoEventMask, (XEvent* )&m);
            XSync(stXDisplay->hDisplay, False);
        }
        XFree(prop.data);
    }
}

void StWindowImpl::updateWindowPos() {
    const StXDisplayH& stXDisplay = hMaster.stXDisplay;
    if(stXDisplay.isNull() || hMaster.hWindowGl == 0) {
        return;
    }
    if(!stWinAttribs.isFullScreen) {
        if(rectNorm.left() == 0 && rectNorm.top() == 0 && hMaster.hWindow != 0) {
            int width  = rectNorm.width();
            int height = rectNorm.height();
            Window hChildWin;
            XTranslateCoordinates(stXDisplay->hDisplay, hMaster.hWindow,
                                  stXDisplay->getRootWindow(),
                                  0, 0, &rectNorm.left(), &rectNorm.top(), &hChildWin);
            rectNorm.right()  = rectNorm.left() + width;
            rectNorm.bottom() = rectNorm.top() + height;
        }
        if(hMaster.hWindow != 0) {
            XMoveResizeWindow(stXDisplay->hDisplay, hMaster.hWindowGl,
                              0, 0,
                              rectNorm.width(), rectNorm.height());
        }
        if(stWinAttribs.isSlave && (!isSlaveIndependent() || monitors.size() > 1)) {
            XMoveResizeWindow(stXDisplay->hDisplay, hSlave.hWindowGl,
                              getSlaveLeft(),  getSlaveTop(),
                              getSlaveWidth(), getSlaveHeight());
        }
    }
    stMessageList.append(StMessageList::MSG_RESIZE); // add event to update GL rendering scape

    // force input focus to Master
    XSetInputFocus(stXDisplay->hDisplay, hMaster.hWindowGl, RevertToParent, CurrentTime);

    // detect when window moved to another monitor
    if(!stWinAttribs.isFullScreen && monitors.size() > 1) {
        int aNewMonId = monitors[rectNorm.center()].getId();
        if(myWinOnMonitorId != aNewMonId) {
            stMessageList.append(StMessageList::MSG_WIN_ON_NEW_MONITOR);
            myWinOnMonitorId = aNewMonId;
        }
    }
}

// Function set to argument-buffer given events
void StWindowImpl::callback(StMessage_t* stMessages) {
    // get callback from Master window
    const StXDisplayH& stXDisplay = hMaster.stXDisplay;
    if(!stXDisplay.isNull()) {
        // detect embedded window was moved
        if(!stWinAttribs.isFullScreen && (Window )nativeParent.winHandle != 0 && hMaster.hWindowGl != 0
        &&  stWinAttribs.isSlave
        && !stWinAttribs.isSlaveHLineTop && !stWinAttribs.isSlaveHTop2Px && !stWinAttribs.isSlaveHLineBottom
        && ++mySyncCounter > 4) {
            // sadly, this Xlib call takes a lot of time
            // perform it rarely
            Window aDummyWin;
            int aCurrPosX, aCurrPosY;
            if(XTranslateCoordinates(stXDisplay->hDisplay, hMaster.hWindowGl,
                                     stXDisplay->getRootWindow(),
                                     0, 0, &aCurrPosX, &aCurrPosY, &aDummyWin)
            && (aCurrPosX != rectNorm.left() || aCurrPosY != rectNorm.top())) {
                //ST_DEBUG_LOG("need updateChildRect= " + aCurrPosX + " x " + aCurrPosY);
                updateChildRect();
            }
            mySyncCounter = 0;
        }

        int indMax = XPending(stXDisplay->hDisplay);
        for(int ind = 0; ind < indMax && XPending(stXDisplay->hDisplay) > 0; ++ind) {
            XNextEvent(stXDisplay->hDisplay, &event);
            switch(event.type) {
                case ClientMessage: {
                    /*ST_DEBUG_LOG("A ClientMessage has arrived:\n"
                        + "Type = " + stXDisplay->getAtomName(event.xclient.message_type)
                        + " (" + event.xclient.format + ")\n"
                        + " message_type " + event.xclient.message_type + "\n"
                    );*/
                    parseXDNDClientMsg();
                    if(event.xclient.data.l[0] == (int )stXDisplay->wndDestroyAtom) {
                        stMessageList.append(StMessageList::MSG_CLOSE);
                    }
                    break;
                }
                case SelectionNotify: {
                    parseXDNDSelectionMsg();
                    break;
                }
                case DestroyNotify: {
                    // something else...
                    break;
                }
                case KeyPress: {
                    XKeyEvent* kevent = (XKeyEvent* )&event;
                    // It is necessary to convert the keycode to a keysym
                    KeySym keysym = XLookupKeysym(kevent, 0);
                    /**ST_DEBUG_LOG("KeyPress,   keycode= " + kevent->keycode
                             + "; KeySym = " + (unsigned int )keysym + "\n");*/
                    if(keysym >= ST_XK2ST_VK_SIZE) {
                        ST_DEBUG_LOG("KeyPress,   keycode= " + kevent->keycode
                                 + "; KeySym = " + (unsigned int )keysym + " ignored!\n");
                        break;
                    }
                    stMessageList.getKeysMap()[ST_XK2ST_VK[keysym]] = true;
                    break;
                }
                case KeyRelease: {
                    XKeyEvent* kevent = (XKeyEvent* )&event;
                    KeySym keysym = XLookupKeysym(kevent, 0);
                    /**ST_DEBUG_LOG("KeyRelease, keycode= " + kevent->keycode
                             + "; KeySym = " + (unsigned int )keysym + "\n");*/
                    if(keysym >= ST_XK2ST_VK_SIZE) {
                        ST_DEBUG_LOG("KeyRelease, keycode= " + kevent->keycode
                                 + "; KeySym = " + (unsigned int )keysym + " ignored!\n");
                        break;
                    }
                    stMessageList.getKeysMap()[ST_XK2ST_VK[keysym]] = false;
                    break;
                }
                case ButtonPress:
                case ButtonRelease: {
                    XButtonEvent* buttonEvent = &event.xbutton;
                    StPointD_t point(double(buttonEvent->x) / double(getPlacement().width()),
                                     double(buttonEvent->y) / double(getPlacement().height()));
                    int mbtn = ST_NOMOUSE;
                    // TODO (Kirill Gavrilov#6#) parse extra mouse buttons
                    switch(buttonEvent->button) {
                        case 1: mbtn = ST_MOUSE_LEFT; break;
                        case 3: mbtn = ST_MOUSE_RIGHT; break;
                        case 2: mbtn = ST_MOUSE_MIDDLE; break;
                        default: mbtn = buttonEvent->button; break;
                    }
                    switch(event.type) {
                        case ButtonRelease: {
                            stMUpQueue.push(point, mbtn);
                            stMessageList.append(StMessageList::MSG_MOUSE_UP);
                            break;
                        }
                        case ButtonPress: {
                            stMDownQueue.push(point, mbtn);
                            stMessageList.append(StMessageList::MSG_MOUSE_DOWN);
                            break;
                        }
                    }
                    /*ST_DEBUG_LOG(((event.type == ButtonPress) ? "ButtonPress " : "Release ")
                        + mbtn + ", x= " + point.x + ", y= " + point.y
                    );*/
                    // force input focus to the master window
                    XSetInputFocus(stXDisplay->hDisplay, hMaster.hWindowGl, RevertToParent, CurrentTime);
                    break;
                }
                //case ResizeRequest:
                //case ConfigureRequest:
                case ConfigureNotify: {
                    //ST_DEBUG_LOG("ConfigureNotify");
                    XConfigureEvent* configEvent = &event.xconfigure;
                    if(!stWinAttribs.isFullScreen && configEvent->window == hMaster.hWindow) {
                        StRectI_t aNewRect;
                        if(configEvent->x > 64 && configEvent->y > 64) {
                            // we got real GLorigin position
                            // worked when we in Compiz mode and on MoveWindow
                            aNewRect.left()   = configEvent->x;
                            aNewRect.right()  = configEvent->x + configEvent->width;
                            aNewRect.top()    = configEvent->y;
                            aNewRect.bottom() = configEvent->y + configEvent->height;
                        } else {
                            // we got psevdo x and y (~window decorations?)
                            // on ResizeWindow without Compiz
                            Window child;
                            aNewRect.left() = 0;
                            aNewRect.top()  = 0;
                            XTranslateCoordinates(stXDisplay->hDisplay, hMaster.hWindow,
                                                  stXDisplay->getRootWindow(),
                                                  0, 0, &aNewRect.left(), &aNewRect.top(), &child);
                            aNewRect.right()  = aNewRect.left() + configEvent->width;
                            aNewRect.bottom() = aNewRect.top() + configEvent->height;
                        }
                        if(rectNorm != aNewRect) {
                            // new compiz send messages when placement not really changed
                            rectNorm = aNewRect;
                            isUpdate = true; // call updateWindowPos() to update position
                        }
                    }
                    break;
                }
                default: {
                    // process XRandr events
                    if(hMaster.isRecXRandrEvents) {
                        int xrandrEventType = event.type - hMaster.xrandrEventBase;
                        if(xrandrEventType == RRNotify || xrandrEventType == RRScreenChangeNotify) {
                            ST_DEBUG_LOG("XRandr update event");
                            monitors.init(); // reinitialize monitors list
                            for(size_t m = 0; m < monitors.size(); ++m) {
                                ST_DEBUG_LOG(monitors[m].toString()); // just debug output
                            }
                        }
                    }
                }
            }
        }
    }

    StPointD_t newMousePt = getMousePos();
    if(newMousePt.x() >= 0.0 && newMousePt.x() <= 1.0 && newMousePt.y() >= 0.0 && newMousePt.y() <= 1.0) {
        StPointD_t aDspl = newMousePt - mousePt;
        if(std::abs(aDspl.x()) >= 0.0008 || std::abs(aDspl.y()) >= 0.0008) {
            stMessageList.append(StMessageList::MSG_MOUSE_MOVE);
        }
    }
    mousePt = newMousePt;

    if(isUpdate) {
        // update position only when all messages are parsed
        updateWindowPos();
        isUpdate = false;
    }

    // TODO (Kirill Gavrilov#5#) parse multimedia keys
    stMessageList.popList(stMessages);
}

#endif
