/**
 * Copyright © 2011 Kirill Gavrilov <kirill@sview.ru>
 *
 * StCore library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * StCore library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library.
 * If not, see <http://www.gnu.org/licenses/>.
 */

#if (defined(__APPLE__))

#include "StWindowImpl.h"
#include "stvkeysxarray.h" // X keys to VKEYs lookup array

#include <StStrings/StLogger.h>

#include "StCocoaView.h"
#include "StCocoaWin.h"

#include <StCocoa/StCocoaLocalPool.h>
#include <StCocoa/StCocoaString.h>

static const NSOpenGLPixelFormatAttribute THE_DOUBLE_BUFF[] = {
    //NSOpenGLPFAColorSize, (NSOpenGLPixelFormatAttribute) 32,
    //NSOpenGLPFADepthSize, (NSOpenGLPixelFormatAttribute) 0,
    //NSOpenGLPFAStencilSize, 0,
    NSOpenGLPFADoubleBuffer,
    NSOpenGLPFAAccelerated,
    0
};

static const NSOpenGLPixelFormatAttribute THE_QUAD_BUFF[] = {
    NSOpenGLPFAStereo, 1,
    NSOpenGLPFADoubleBuffer,
    NSOpenGLPFAAccelerated,
    0
};

void StWindowImpl::setTitle(const StString& theTitle) {
    windowName = theTitle;
    if(hMaster.hWindow != NULL) {
        StCocoaLocalPool aLocalPool;
        StCocoaString aTitle(windowName);
        [hMaster.hWindow setTitle: aTitle.toStringNs()];
    }
}

void StWindowImpl::setPlacement(const StRectI_t& theRect) {
    rectNorm = theRect;
    isUpdate = true;

    if(hMaster.hWindow != NULL
    && !stWinAttribs.isFullScreen) {
        StCocoaLocalPool aLocalPool;
        NSRect aFrameRect = [hMaster.hWindow frameRectForContentRect: myCocoaCoords.normalToCocoa(rectNorm)];
        [hMaster.hWindow setFrame: aFrameRect display: YES];
    }
}

void StWindowImpl::show(const int& winNum) {
    if(winNum == ST_WIN_MASTER && stWinAttribs.isHide) {
        if(hMaster.hWindow != NULL) {
            [hMaster.hWindow orderFront: NULL];
        }
        stWinAttribs.isHide = false;
        updateWindowPos();
    } else if(winNum == ST_WIN_SLAVE && stWinAttribs.isSlaveHide) {
        if(hSlave.hWindow != NULL) {
            [hSlave.hWindow orderFront: NULL];
        }
        stWinAttribs.isSlaveHide = false;
        updateWindowPos();
    }
}

void StWindowImpl::hide(const int& winNum) {
    StCocoaLocalPool aLocalPool;
    if(winNum == ST_WIN_MASTER && !stWinAttribs.isHide) {
        if(hMaster.hWindow != NULL) {
            [hMaster.hWindow orderOut: NULL];
        }
        stWinAttribs.isHide = true;
    } else if(winNum == ST_WIN_SLAVE && !stWinAttribs.isSlaveHide) {
        if(hSlave.hWindow != NULL) {
            [hSlave.hWindow orderOut: NULL];
        }
        stWinAttribs.isSlaveHide = true;
    }
}

@interface StWinInitInfo : NSObject
    {
        StWindowImpl*    myWinSt;
        NSOpenGLContext* myGLContextMaster;
        NSOpenGLContext* myGLContextSlave;
    }

    - (id ) init: (StWindowImpl* )    theWinSt
          master: (NSOpenGLContext* ) theGLContextMaster
           slave: (NSOpenGLContext* ) theGLContextSlave;

    - (void ) doCreateWindows: (id ) theSender;
@end

@implementation StWinInitInfo

    - (id ) init: (StWindowImpl* )    theWinSt
          master: (NSOpenGLContext* ) theGLContextMaster
           slave: (NSOpenGLContext* ) theGLContextSlave {
        self = [super init];
        if(self == NULL) {
            return NULL;
        }
        myWinSt = theWinSt;
        myGLContextMaster = theGLContextMaster;
        myGLContextSlave  = theGLContextSlave;
        return self;
    }

    - (void ) doCreateWindows: (id ) theSender {
        myWinSt->doCreateWindows(myGLContextMaster, myGLContextSlave);
    }

@end

void StWindowImpl::doCreateWindows(NSOpenGLContext* theGLContextMaster,
                                   NSOpenGLContext* theGLContextSlave) {
    StCocoaLocalPool aLocalPool;

    // create the Master window
    NSUInteger aWinStyle = stWinAttribs.isNoDecor
                         ? NSBorderlessWindowMask
                         : NSTitledWindowMask | NSClosableWindowMask | NSResizableWindowMask;
    hMaster.hWindow = [[StCocoaWin alloc] initWithStWin: this
                                                   rect: rectNorm
                                              styleMask: aWinStyle];
    if(hMaster.hWindow == NULL) {
        stError("Cocoa, Master Window Creation Error");
        lastError = STWIN_ERROR_COCOA_CREATEWIN;
        return;
    }
    StCocoaString aTitle(windowName);
    [hMaster.hWindow setTitle: aTitle.toStringNs()];

    // create the view in Master window
    hMaster.hViewGl = [[StCocoaView alloc] initWithStWin: this
                                                   nsWin: hMaster.hWindow];
    if(hMaster.hViewGl == NULL) {
        hMaster.close();
        stError("Cocoa, Master GL View Creation Error");
        lastError = STWIN_ERROR_COCOA_CREATEWIN;
        return;
    }
    // setup GL context (NSOpenGLView create context on first call
    // so nothing is released here)
    [hMaster.hViewGl setOpenGLContext: theGLContextMaster];
    [theGLContextMaster setView: hMaster.hViewGl];

    if(stWinAttribs.isSlave) {
        // create the Slave window
        StRectI_t aRectSlave(getSlaveTop(),  getSlaveTop() + getSlaveHeight(),
                             getSlaveLeft(), getSlaveLeft() + getSlaveWidth());
        hSlave.hWindow = [[StCocoaWin alloc] initWithStWin: this
                                                      rect: aRectSlave
                                                 styleMask: NSBorderlessWindowMask];
        if(hSlave.hWindow == NULL) {
            hMaster.close();
            hSlave.close();
            stError("Cocoa, Slave Window Creation Error");
            lastError = STWIN_ERROR_COCOA_CREATEWIN;
            return;
        }
        [hSlave.hWindow setTitle: @"Slave window"];

        // create the view in Slave window
        hSlave.hViewGl = [[StCocoaView alloc] initWithStWin: this
                                                      nsWin: hSlave.hWindow];
        if(hSlave.hViewGl == NULL) {
            stError("Cocoa, Slave GL View Creation Error");
            hMaster.close();
            hSlave.close();
            lastError = STWIN_ERROR_COCOA_CREATEWIN;
            return;
        }
        // setup GL context (NSOpenGLView create context on first call
        // so nothing is released here)
        [hSlave.hViewGl setOpenGLContext: theGLContextSlave];
        [theGLContextSlave setView: hSlave.hViewGl];

        // make slave window topmost
        [hSlave.hWindow setLevel: kCGMaximumWindowLevel];
        if(!stWinAttribs.isSlaveHide && (!isSlaveIndependent() || monitors.size() > 1)) {
            [hSlave.hWindow orderFront: NULL];
        }
    }

    [hMaster.hWindow makeKeyAndOrderFront: NSApp];
}

bool StWindowImpl::stglCreate(const StWinAttributes_t* inAttributes,
                              const StNativeWin_t*     nativeParentWindow) {
    lastError = STWIN_INITNOTSTART;
    if(nativeParentWindow != NULL) {
        ///stMemCpy(&(this->nativeParent), nativeParentWindow, sizeof(StNativeWin_t));
        ///nativeParent.stWinPtr = this;
    }
    size_t bytesToCopy = (inAttributes->nSize > sizeof(StWinAttributes_t)) ? sizeof(StWinAttributes_t) : inAttributes->nSize;
    stMemCpy(&stWinAttribs, inAttributes, bytesToCopy); // copy as much as possible
    stWinAttribs.nSize = sizeof(StWinAttributes_t);     // restore own size
    updateSlaveConfig();
    updateChildRect();

    if(NSApp == NULL) {
        stError("Cocoa, main application doesn't exists");
        lastError = STWIN_ERROR_COCOA_NO_APP;
        return false;
    }

    StCocoaLocalPool aLocalPool;

    // create the Master GL context
    ///NSOpenGLPixelFormat* aGlFormatMaster = [[[NSOpenGLView class] defaultPixelFormat] retain];
    NSOpenGLPixelFormat* aGLFormat    = NULL;
    NSOpenGLContext* aGLContextMaster = NULL;
    NSOpenGLContext* aGLContextSlave  = NULL;
    if(stWinAttribs.isGlStereo) {
        aGLFormat = [[[NSOpenGLPixelFormat alloc] initWithAttributes: THE_QUAD_BUFF] autorelease];
        aGLContextMaster = [[[NSOpenGLContext alloc] initWithFormat: aGLFormat
                                                       shareContext: NULL] autorelease];
        if(aGLContextMaster == NULL) {
            aGLFormat = [[[NSOpenGLPixelFormat alloc] initWithAttributes: THE_DOUBLE_BUFF] autorelease];
            aGLContextMaster = [[[NSOpenGLContext alloc] initWithFormat: aGLFormat
                                                           shareContext: NULL] autorelease];
            if(aGLContextMaster == NULL) {
                stError("Cocoa, fail to create Double Buffered OpenGL context");
                lastError = STWIN_ERROR_COCOA_NO_GL;
                return false;
            } else {
                stError("Cocoa, fail to create Quad Buffered OpenGL context");
            }
        }
    } else {
        aGLFormat = [[[NSOpenGLPixelFormat alloc] initWithAttributes: THE_DOUBLE_BUFF] autorelease];
        aGLContextMaster = [[[NSOpenGLContext alloc] initWithFormat: aGLFormat
                                                       shareContext: NULL] autorelease];
        if(aGLContextMaster == NULL) {
            stError("Cocoa, fail to create Double Buffered OpenGL context");
            lastError = STWIN_ERROR_COCOA_NO_GL;
            return false;
        }
    }

    // create the Slave GL context
    if(stWinAttribs.isSlave) {
        aGLContextSlave = [[[NSOpenGLContext alloc] initWithFormat: aGLFormat
                                                      shareContext: aGLContextMaster] autorelease];
        if(aGLContextSlave == NULL) {
            stError("Cocoa, fail to create Slave OpenGL context");
            lastError = STWIN_ERROR_COCOA_NO_GL;
            return false;
        }
    }

    // create special object to invoke in main thread
    StWinInitInfo* aWinInit = [[StWinInitInfo alloc] init: this
                                                   master: aGLContextMaster
                                                    slave: aGLContextSlave];
    if(aWinInit == NULL) {
        lastError = STWIN_ERROR_COCOA_CREATEWIN;
        return false;
    }
    if([NSThread isMainThread]) {
        [aWinInit doCreateWindows: NULL];
    } else {
        [aWinInit performSelectorOnMainThread: @selector(doCreateWindows:)
                                   withObject: NULL
                                waitUntilDone: YES];
    }
    [aWinInit release];

    if(lastError != STWIN_INITNOTSTART) {
        return false;
    }

    hMaster.glMakeCurrent();

    isUpdate = true;
    lastError = STWIN_INIT_SUCCESS;
    return true;
}

/**
 * Update StWindow position according to native parent position.
 */
void StWindowImpl::updateChildRect() {
    ///
}

void StWindowImpl::setFullScreen(bool fullscreen) {
    stWinAttribs.isFullScreen = fullscreen;
    if(stWinAttribs.isFullScreen) {
        const StMonitor& aMon = (monMasterFullScr == -1) ? monitors[rectNorm.center()] : monitors[monMasterFullScr];
        rectFull = aMon.getVRect();

        if(stWinAttribs.isSlave && hSlave.hViewGl != NULL) {
            [hSlave.hViewGl goToFullscreen];
        }
        if(hMaster.hViewGl != NULL) {
            [hMaster.hViewGl goToFullscreen];
        }
    } else {
        if(stWinAttribs.isSlave && hSlave.hViewGl != NULL) {
            [hSlave.hViewGl goToWindowed];
        }
        if(hMaster.hViewGl != NULL) {
            [hMaster.hViewGl goToWindowed];
        }
    }
}

void StWindowImpl::updateWindowPos() {
    if(stWinAttribs.isSlave && !stWinAttribs.isSlaveHide
    && (!isSlaveIndependent() || monitors.size() > 1)
    && !stWinAttribs.isFullScreen) {
        StRectI_t aRectSlave(getSlaveTop(),  getSlaveTop() + getSlaveHeight(),
                             getSlaveLeft(), getSlaveLeft() + getSlaveWidth());
        NSRect aFrameRect = [hSlave.hWindow frameRectForContentRect: myCocoaCoords.normalToCocoa(aRectSlave)];
        [hSlave.hWindow setFrame: aFrameRect display: YES];
    }

    // detect when window moved to another monitor
    if(!stWinAttribs.isFullScreen && monitors.size() > 1) {
        int aNewMonId = monitors[rectNorm.center()].getId();
        if(myWinOnMonitorId != aNewMonId) {
            stMessageList.append(StMessageList::MSG_WIN_ON_NEW_MONITOR);
            myWinOnMonitorId = aNewMonId;
        }
    }
}

// Function set to argument-buffer given events
void StWindowImpl::callback(StMessage_t* stMessages) {
    // detect master window movements
    if(stWinAttribs.isFullScreen) {
        if(rectNormPrev != rectFull) {
            rectNormPrev = rectFull;
            isUpdate = true;
            stMessageList.append(StMessageList::MSG_RESIZE);
        }
    } else {
        StRectI_t aWinRectNew = myCocoaCoords.cocoaToNormal([hMaster.hWindow contentRectForFrameRect: [hMaster.hWindow frame]]);
        if(rectNormPrev != aWinRectNew) {
            rectNorm = aWinRectNew;
            rectNormPrev = aWinRectNew;
            isUpdate = true;
            stMessageList.append(StMessageList::MSG_RESIZE);
        }
    }

    // detect mouse movements
    StPointD_t newMousePt = getMousePos();
    if(newMousePt.x() >= 0.0 && newMousePt.x() <= 1.0 && newMousePt.y() >= 0.0 && newMousePt.y() <= 1.0) {
        StPointD_t aDspl = newMousePt - mousePt;
        if(std::abs(aDspl.x()) >= 0.0008 || std::abs(aDspl.y()) >= 0.0008) {
            stMessageList.append(StMessageList::MSG_MOUSE_MOVE);
        }
    }
    mousePt = newMousePt;

    if(isUpdate) {
        // update position only when all messages are parsed
        updateWindowPos();
        isUpdate = false;
    }

    stMessageList.popList(stMessages);
}

#endif // __APPLE__
