/**
 * Copyright © 2007-2011 Kirill Gavrilov <kirill@sview.ru>
 *
 * StCore library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * StCore library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library.
 * If not, see <http://www.gnu.org/licenses/>.
 */

#if(defined(_WIN32) || defined(__WIN32__))

#include "StWindowImpl.h"

#include <StStrings/StLogger.h>

#include <math.h>

static SV_THREAD_FUNCTION threadCreateWindows(void* inStWin);

/**
 * Static proc function just do recall to StWindowImpl class method.
 */
static LRESULT CALLBACK stWndProcWrapper(HWND in_hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
    if(uMsg == WM_CREATE) {
        // save pointer to our class instance (sended on window create) to window storage
        LPCREATESTRUCT pCS = (LPCREATESTRUCT )lParam;
        SetWindowLongPtr(in_hWnd, int(GWLP_USERDATA), (LONG_PTR )pCS->lpCreateParams);
    }
    // get pointer to our class instance
    StWindowImpl* pThis = (StWindowImpl* )GetWindowLongPtr(in_hWnd, int(GWLP_USERDATA));
    if(pThis != NULL) {
        return pThis->stWndProc(in_hWnd, uMsg, wParam, lParam);
    } else {
        return DefWindowProc(in_hWnd, uMsg, wParam, lParam);
    }
}

// function create GUI window
bool StWindowImpl::stglCreate(const StWinAttributes_t* inAttributes, const StNativeWin_t* nativeParentWindow) {
    if(nativeParentWindow != NULL) {
        stMemCpy(&(this->nativeParent), nativeParentWindow, sizeof(StNativeWin_t));
    }

    // parse attributes
    size_t bytesToCopy = (inAttributes->nSize > sizeof(StWinAttributes_t)) ? sizeof(StWinAttributes_t) : inAttributes->nSize;
    stMemCpy(&stWinAttribs, inAttributes, bytesToCopy); // copy as much as possible
    stWinAttribs.nSize = sizeof(StWinAttributes_t);             // restore own size
    updateSlaveConfig();

    lastError = STWIN_INITNOTSTART;

    evInitWin.reset();
    evInitGl.reset();
    stMessageThread = new StThread(threadCreateWindows, (void* )this);
    // wait for thread to create window
    evInitWin.wait();
    if(lastError != STWIN_INIT_SUCCESS) {
        return false;
    }
    int isGlCtx = hMaster.glCreateContext(stWinAttribs.isSlave ? &hSlave : NULL, stWinAttribs.isGlStereo);
    evInitGl.set();

    return (isGlCtx == STWIN_INIT_SUCCESS);
}

bool StWindowImpl::wndRegisterClass(const StStringUtfWide& theClassName) {
    HINSTANCE hInstance = GetModuleHandle(NULL); // Holds The Instance Of The Application
    WNDCLASSW wndClass;                                           // Windows Class Structure
    wndClass.style         = CS_HREDRAW | CS_VREDRAW | CS_OWNDC; // Redraw On Size, And Own DC For Window.
    wndClass.lpfnWndProc   = (WNDPROC )stWndProcWrapper;         // WndProc Handles Messages
    wndClass.cbClsExtra    = 0;                                  // No Extra Window Data
    wndClass.cbWndExtra    = 0;                                  // No Extra Window Data
    wndClass.hInstance     = hInstance;                          // Set The Instance
    // TODO (Kirill Gavrilov) change icon setup
    wndClass.hIcon         = LoadIcon(hInstance, L"A");          // Load The Icon A
    wndClass.hCursor       = LoadCursor(NULL, IDC_ARROW);        // Load The Arrow Pointer
    wndClass.hbrBackground = NULL;                               // No Background Required For GL
    wndClass.lpszMenuName  = NULL;                               // No menu
    wndClass.lpszClassName = theClassName.toCString();           // Set The Class Name
    return (RegisterClassW(&wndClass) != 0);
}

// Function create windows' handles in another thread (to take events impact-less to GL-rendering)
static SV_THREAD_FUNCTION threadCreateWindows(void* inStWin) {
    StWindowImpl* stWin = (StWindowImpl* )inStWin;
    stWin->wndCreateWindows();
    return SV_THREAD_RETURN 0;
}

bool StWindowImpl::wndCreateWindows() {
    StStringUtfWide aNewClassName = StWinHandles::getNewClassName();

    // need we register class for Master window?
    if(nativeParent == NULL) {
        if(!wndRegisterClass(aNewClassName)) {
            stError("WinAPI: Failed to register Master window class");
            lastError = STWIN_ERROR_WIN32_REGCLASS;
            return false;
        } else {
            hMaster.className = aNewClassName;
        }
    }

    aNewClassName += "Gl";
    if(!wndRegisterClass(aNewClassName)) {
        stError("WinAPI: Failed to register Master window class");
        lastError = STWIN_ERROR_WIN32_REGCLASS;
        return false;
    } else {
        hMaster.classNameGl = aNewClassName;
    }
    if(stWinAttribs.isSlave) {
        aNewClassName = StWinHandles::getNewClassName() + "Gl";
        if(!wndRegisterClass(aNewClassName)) {
            stError("WinAPI: Failed to register Slave window class");
            lastError = STWIN_ERROR_WIN32_REGCLASS;
            return false;
        } else {
            hSlave.classNameGl = aNewClassName;
        }
    }
    hMaster.threadIdWnd = hSlave.threadIdWnd = StThread::getCurrentThreadId();

    // ========= At first - create windows' instances =========
    // parent Master window could be decorated - we parse this situation to get true coordinates
    int posLeft   = rectNorm.left() - (!stWinAttribs.isNoDecor ? GetSystemMetrics(SM_CXSIZEFRAME) : 0);
    int posTop    = rectNorm.top() - (!stWinAttribs.isNoDecor ? (GetSystemMetrics(SM_CYSIZEFRAME) + GetSystemMetrics(SM_CYCAPTION)) : 0);
    int winWidth  = (!stWinAttribs.isNoDecor ? 2 * GetSystemMetrics(SM_CXSIZEFRAME) : 0) + rectNorm.width();
    int winHeight = (!stWinAttribs.isNoDecor ? (GetSystemMetrics(SM_CYCAPTION) + 2 * GetSystemMetrics(SM_CYSIZEFRAME)) : 0) + rectNorm.height();
    HINSTANCE hInstance = GetModuleHandle(NULL); // Holds The Instance Of The Application

    if(nativeParent == NULL) {
        // WS_EX_ACCEPTFILES - Drag&Drop support
        hMaster.hWindow = CreateWindowExW(WS_EX_APPWINDOW | WS_EX_WINDOWEDGE | WS_EX_ACCEPTFILES,
                                         hMaster.className.toCString(),
                                         windowName.toUtfWide().toCString(),
                                         (!stWinAttribs.isNoDecor ? WS_OVERLAPPEDWINDOW : WS_POPUP) | WS_CLIPSIBLINGS | WS_CLIPCHILDREN,
                                         posLeft, posTop,
                                         winWidth, winHeight,
                                         NULL, NULL, hInstance, this); // put pointer to class, getted on WM_CREATE
        if(hMaster.hWindow == NULL) {
            hMaster.close();
            stError("WinAPI: Master Window Creation Error");
            lastError = STWIN_ERROR_WIN32_CREATE;
            return false;
        }
    } else {
        // we have external native parent window
        stWinAttribs.isNoDecor = true;

        RECT rect;
        GetWindowRect(nativeParent, &rect);

        rectNorm.left()   = rect.left;
        rectNorm.right()  = rect.right;
        rectNorm.top()    = rect.top;
        rectNorm.bottom() = rect.bottom;

        posLeft   = rectNorm.left();
        posTop    = rectNorm.top();
        winWidth  = rectNorm.width();
        winHeight = rectNorm.height();
        isUpdate = true;
    }

    // we use WS_EX_NOPARENTNOTIFY style to prevent to send notify on destroing our child window (NPAPI plugin -> deadlock)
    DWORD masterWindowGl_dwExStyle = (nativeParent == NULL) ? WS_EX_NOACTIVATE : (WS_EX_NOACTIVATE | WS_EX_NOPARENTNOTIFY);
    hMaster.hWindowGl = CreateWindowExW(masterWindowGl_dwExStyle,
                                        hMaster.classNameGl.toCString(),
                                        windowName.toUtfWide().toCString(),
                                        WS_VISIBLE | WS_CHILD,  // this is child sub-window!
                                        0, 0, rectNorm.width(), rectNorm.height(),
                                        (nativeParent == NULL) ? hMaster.hWindow : nativeParent,
                                        NULL, hInstance, this); // put pointer to class, getted on WM_CREATE
    if(hMaster.hWindowGl == NULL) {
        hMaster.close();
        stError("WinAPI: Master GL Window Creation Error");
        lastError = STWIN_ERROR_WIN32_CREATE;
        return false;
    }

    if(stWinAttribs.isSlave) {
        hSlave.hWindowGl = CreateWindowExW(WS_EX_TOOLWINDOW | WS_EX_WINDOWEDGE | WS_EX_NOACTIVATE,
                                           hSlave.classNameGl.toCString(),
                                           L"Slave window",
                                           WS_POPUP | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_DISABLED, // slave always disabled (hasn't input focus)!
                                           posLeft, posTop, // initialize slave window at same screen as master to workaround bugs in drivers that may prevent GL context sharing
                                           rectNorm.width(), rectNorm.height(),
                                           NULL, NULL, hInstance, NULL);
        if(hSlave.hWindowGl == NULL) {
            hMaster.close();
            hSlave.close();
            stError(L"WinAPI: Slave Window Creation Error");
            lastError = STWIN_ERROR_WIN32_CREATE;
            return false;
        }
    }

    // ========= Synchronization barrier - wait until MAIN thread create GL contexts =========
    lastError = STWIN_INIT_SUCCESS; // 'send' it's OK for main thread
    evInitWin.set();
    evInitGl.wait();

    // ========= Now show up the windows =========
    if(stWinAttribs.isSlave && !stWinAttribs.isSlaveHide && (!isSlaveIndependent() || monitors.size() > 1)) {
        SetWindowPos(hSlave.hWindowGl,
                     HWND_NOTOPMOST,
                     getSlaveLeft(), getSlaveTop(),
                     getSlaveWidth(), getSlaveHeight(),
                     SWP_SHOWWINDOW);
    }

    if(!stWinAttribs.isHide) {
        SetWindowPos(hMaster.hWindow,     // window handle
                     HWND_NOTOPMOST,      // top status
                     posLeft, posTop,     // Left-Top position
                     winWidth, winHeight, // Width x Height
                     SWP_SHOWWINDOW);

        SetWindowPos(hMaster.hWindowGl, // window handle
                     HWND_NOTOPMOST,    // top status
                     0, 0, rectNorm.width(), rectNorm.height(),
                     SWP_SHOWWINDOW);   // show window
    }

    stMessageList.append(StMessageList::MSG_RESIZE);

    // always wait for message thread exit before quit
    hMaster.evMsgThread.reset();

    // ========= Start callback procedure =========
    if(!stWinAttribs.isHide && nativeParent == NULL) {
        // TODO (Kirill Gavrilov#4) need we special StWindow function to make window on top?
        SetForegroundWindow(hMaster.hWindow); // make sure Master window on top and has input focus
    }
    HANDLE waitEvents[3] = {hEventQuit, hEventCursorShow, hEventCursorHide};
    for(;;) {
        switch(MsgWaitForMultipleObjects(3, waitEvents, FALSE, INFINITE, QS_ALLINPUT)) {
            case WAIT_OBJECT_0: {
                // Event 1 (hEventQuit) has been set. If the event was created as autoreset, it has also been reset
                ///ST_DEBUG_LOG("WinAPI, End of the message thread... (TID= " + StThread::getCurrentThreadId() + ")");
                hSlave.close();  // close window handles
                hMaster.close();

                hSlave.threadIdWnd = 0;
                hMaster.threadIdWnd = 0;

                hMaster.evMsgThread.set(); // thread now exit, nothing should be after!
                return true;
            }
            case WAIT_OBJECT_0 + 1: {
                // Event 2 (hEventCursorShow) has been set. If the event was created as autoreset, it has also been reset
                // show / hide cursor - SHOULD be called from window thread...
                ShowCursor(TRUE);
                ///ST_DEBUG_LOG("ShowCursor(TRUE)");
                ResetEvent(hEventCursorShow);
                break;
            }
            case WAIT_OBJECT_0 + 2: {
                // Event 3 (hEventCursorHide) has been set. If the event was created as autoreset, it has also been reset
                // warning - this is NOT force hide / show function;
                // this call decrease or increase show counter
                ShowCursor(FALSE);
                ///ST_DEBUG_LOG("ShowCursor(FALSE)");
                ResetEvent(hEventCursorHide);
                break;
            }
            case WAIT_OBJECT_0 + 3: {
                // A thread's (window's) message(s) has arrived
                // We should process ALL messages cause MsgWaitForMultipleObjects
                // will NOT triggered for new messages already in stack!!!
                // Means - do not replace 'while' with 'if(PeekMessage(...))'.
                while(PeekMessage(&event, NULL, 0U, 0U, PM_REMOVE)) {
                    TranslateMessage(&event); // Translate The Message
                    DispatchMessage(&event);  // Dispatch The Message
                }
                break; // break from switch
            }
            default: {
                // no input or messages waiting
                break;
            }
        }
    }
}

/**
 * Update StWindow position according to native parent position.
 */
void StWindowImpl::updateChildRect() {
    if(!stWinAttribs.isFullScreen && nativeParent != NULL) {
        isUpdate = true;
        RECT aRect;
        GetClientRect(nativeParent, &aRect);
        ClientToScreen(nativeParent, (POINT* )&aRect);
        ClientToScreen(nativeParent, ((POINT* )&aRect) + 1);

        // TODO (Kirill Gavrilov#2) not thread safe assignment!
        rectNorm.left()   = aRect.left;
        rectNorm.top()    = aRect.top;
        rectNorm.right()  = aRect.right;
        rectNorm.bottom() = aRect.bottom;

        stMessageList.append(StMessageList::MSG_RESIZE);
    }
}

LRESULT StWindowImpl::stWndProc(HWND in_hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
    switch(uMsg) {                 // Check For Windows Messages
        case WM_ACTIVATE: {        // Watch For Window Activate Message
            if(!HIWORD(wParam)) {
                //active = true;   // Check Minimization State
            } else {
                //active = false;  // Program Is No Longer Active
                //active = true;
            }
            return 0;              // Return To The Message Loop
        }

        case WM_SYSCOMMAND: {      // Intercept System Commands
            switch(wParam) {       // Check System Calls
            case 0xF032: // TODO (Kirill Gavrilov#7#) parse double click on title in 'right' way
            case SC_MAXIMIZE:
                setFullScreen(true);
                isUpdate = true;
                stMessageList.append(StMessageList::MSG_RESIZE);
            case SC_SCREENSAVE:    // Screensaver Trying To Start?
            case SC_MONITORPOWER:  // Monitor Trying To Enter Powersave?
                return 0;          // Prevent From Happening
            }
            break;                 // Exit
        }

        case WM_CLOSE: {           // Did We Receive A Close Message?
            stMessageList.append(StMessageList::MSG_CLOSE);
            return 0;              // Jump Back
        }

        case WM_DROPFILES: {
            HDROP aDrops = (HDROP )wParam;
            UINT aFilesCount = DragQueryFileW(aDrops, 0xFFFFFFFF, NULL, 0);
            stUtfWide_t aFileBuff[MAX_PATH];
            dragNDMutex.lock();
            if(aFilesCount < 1) {
                DragFinish(aDrops);
                break;
            }
            dragNDCount = aFilesCount;
            delete[] dndList;
            dndList = new StString[dragNDCount];
            for(UINT aFileId = 0; aFileId < aFilesCount; ++aFileId) {
                if(DragQueryFileW(aDrops, aFileId, aFileBuff, MAX_PATH) > 0) {
                    dndList[aFileId] = StString(aFileBuff);
                }
            }
            dragNDMutex.unlock();
            DragFinish(aDrops); // do not forget
            stMessageList.append(StMessageList::MSG_DRAGNDROP_IN);
            break;
        }
        case WM_MOVE: {
            if(stWinAttribs.isFullScreen || nativeParent != NULL) {
                break;
            } else if(in_hWnd == hMaster.hWindow) {
                isUpdate = true;
                // TODO (Kirill GAvrilov#2) not thread safe assignment!
                rectNorm.left() = (int )(short )LOWORD(lParam);
                rectNorm.top()  = (int )(short )HIWORD(lParam);
                stMessageList.append(StMessageList::MSG_RESIZE);
                //ST_DEBUG_LOG("WM_MOVE, XxY= " + x + "x" + y);
                break;
            }
            // ignore GL subwindow resize messages!
            break;
        }
        case WM_SIZE: {
            if(stWinAttribs.isFullScreen || nativeParent != NULL) {
                break;
            } else if(in_hWnd == hMaster.hWindow) {
                isUpdate = true;
                int w = LOWORD(lParam);
                int h = HIWORD(lParam);
                // TODO (Kirill GAvrilov#2) not thread safe assignment!
                rectNorm.right()  = rectNorm.left() + w;
                rectNorm.bottom() = rectNorm.top()  + h;
                stMessageList.append(StMessageList::MSG_RESIZE);
                //ST_DEBUG_LOG("WM_SIZE, WxH= " + w + "x" + h);
                break;
            }
            break;
        }
        case WM_MOVING:
        case WM_SIZING: {
            if(stWinAttribs.isFullScreen || nativeParent != NULL) {
                break;
            } else if(in_hWnd == hMaster.hWindow) {
                RECT* rec = (RECT* )(LPARAM )lParam;
                // TODO (Kirill GAvrilov#2) not thread safe assignment!
                rectNorm.left()   = rec->left   + GetSystemMetrics(SM_CXSIZEFRAME);
                rectNorm.right()  = rec->right  - GetSystemMetrics(SM_CXSIZEFRAME);
                rectNorm.top()    = rec->top    + GetSystemMetrics(SM_CYSIZEFRAME) + GetSystemMetrics(SM_CYCAPTION);
                rectNorm.bottom() = rec->bottom - GetSystemMetrics(SM_CYSIZEFRAME);
                isUpdate = true;
                stMessageList.append(StMessageList::MSG_RESIZE);
                break;
            }
            break;
        }
        // keys lookup
        case WM_KEYDOWN: {
            stMessageList.getKeysMap()[wParam] = true; break;
        }
        case WM_KEYUP: {         // has a key been released?
            stMessageList.getKeysMap()[wParam] = false; break;
        }
        // mouse lookup
        // TODO (Kirill Gavrilov#6#) parse double click mouse messages
        //case WM_LBUTTONDBLCLK: // left double click
        //case WM_MBUTTONDBLCLK: // right double click
        case WM_MOUSEWHEEL:    // vertical wheel
        //case WM_MOUSEHWHEEL:   // horizontal wheel (only Vista+)
        case WM_LBUTTONUP:
        case WM_RBUTTONUP:
        case WM_MBUTTONUP:
        case WM_XBUTTONUP:
        case WM_LBUTTONDOWN:
        case WM_RBUTTONDOWN:
        case WM_MBUTTONDOWN:
        case WM_XBUTTONDOWN: {
            int mouseXPx = int(short(LOWORD(lParam)));
            int mouseYPx = int(short(HIWORD(lParam)));
            if(uMsg == WM_MOUSEWHEEL) {
                // special case - WinAPI give us position relative to screen!
                mouseXPx -= getPlacement().left();
                mouseYPx -= getPlacement().top();
            }
            StPointD_t point(double(mouseXPx) / double(getPlacement().width()),
                             double(mouseYPx) / double(getPlacement().height()));
            int mbtn = ST_NOMOUSE;
            switch(uMsg) {
                case WM_LBUTTONUP:
                case WM_LBUTTONDOWN: mbtn = ST_MOUSE_LEFT; break;
                case WM_RBUTTONUP:
                case WM_RBUTTONDOWN: mbtn = ST_MOUSE_RIGHT; break;
                case WM_MBUTTONUP:
                case WM_MBUTTONDOWN: mbtn = ST_MOUSE_MIDDLE; break;
                case WM_XBUTTONUP:
                case WM_XBUTTONDOWN: mbtn = (HIWORD(wParam) == XBUTTON1) ? ST_MOUSE_X1 : ST_MOUSE_X2; break;
                case WM_MOUSEWHEEL: {
                    int zDelta = GET_WHEEL_DELTA_WPARAM(wParam);
                    mbtn = (zDelta > 0) ? ST_MOUSE_SCROLL_V_UP : ST_MOUSE_SCROLL_V_DOWN;
                    break;
                }
            }
            switch(uMsg) {
                case WM_LBUTTONUP:
                case WM_RBUTTONUP:
                case WM_MBUTTONUP:
                case WM_XBUTTONUP: {
                    // TODO (Kirill Gavrilov#9) what if we have some another not unclicked mouse button?
                    ReleaseCapture();

                    stMUpQueue.push(point, mbtn);
                    stMessageList.append(StMessageList::MSG_MOUSE_UP);
                    break;
                }
                case WM_LBUTTONDOWN:
                case WM_RBUTTONDOWN:
                case WM_MBUTTONDOWN:
                case WM_XBUTTONDOWN: {
                    // to receive out-of-window unclick message
                    SetFocus(in_hWnd);
                    SetCapture(in_hWnd);

                    stMDownQueue.push(point, mbtn);
                    stMessageList.append(StMessageList::MSG_MOUSE_DOWN);
                    break;
                }
                case WM_MOUSEWHEEL: {
                    // TODO (Kirill Gavrilov#9#) delta ignored
                    stMDownQueue.push(point, mbtn);
                    stMessageList.append(StMessageList::MSG_MOUSE_DOWN);
                    stMUpQueue.push(point, mbtn);
                    stMessageList.append(StMessageList::MSG_MOUSE_UP);
                    break;
                }
            }
            break;
        }
    }
    // Pass All Unhandled Messages To DefWindowProc
    return DefWindowProc(in_hWnd, uMsg, wParam, lParam);
}

/**
 * In this function we fullscreen only Master window,
 * Slave window resized in update procedure.
 */
void StWindowImpl::setFullScreen(bool fullscreen) {
    stWinAttribs.isFullScreen = fullscreen;

    ShowWindow(hMaster.hWindow, SW_HIDE);
    if(stWinAttribs.isFullScreen) {
        // embedded
        if(nativeParent != NULL) {
            // TODO (Kirill Gavrilov#6) check safity
            SetParent(hMaster.hWindowGl, NULL);
            SetFocus(hMaster.hWindowGl);
        }

        // generic
        const StMonitor& stMon = (monMasterFullScr == -1) ? monitors[rectNorm.center()] : monitors[monMasterFullScr];
        rectFull = stMon.getVRect();
        SetWindowLongPtr(hMaster.hWindow, GWL_STYLE, WS_POPUP);
        SetWindowPos(hMaster.hWindow, // window handle
                     HWND_TOP,        // top status
                     rectFull.left(),  rectFull.top(),
                     rectFull.width(), rectFull.height(),
                     !stWinAttribs.isHide ? SWP_SHOWWINDOW : SWP_NOACTIVATE); // show window
        if(!stWinAttribs.isHide) {
            SetFocus(hMaster.hWindow);
        }
    } else {
        // embedded
        if(nativeParent != NULL) {
            SetParent(hMaster.hWindowGl, (HWND )nativeParent);
        }

        // generic
        if(!stWinAttribs.isNoDecor) {
            SetWindowLongPtr(hMaster.hWindow, GWL_STYLE, WS_OVERLAPPEDWINDOW);
        }
        SetWindowPos(hMaster.hWindow, // window handle
                     HWND_NOTOPMOST,  // top status
                     rectNorm.left() - GetSystemMetrics(SM_CXSIZEFRAME), // left position
                     rectNorm.top()  - GetSystemMetrics(SM_CYSIZEFRAME) - GetSystemMetrics(SM_CYCAPTION), // top position
                     2 * GetSystemMetrics(SM_CXSIZEFRAME) + rectNorm.width(), // width
                     GetSystemMetrics(SM_CYCAPTION) + 2* GetSystemMetrics(SM_CYSIZEFRAME) + rectNorm.height(), // height
                     !stWinAttribs.isHide ? SWP_SHOWWINDOW : SWP_NOACTIVATE); // show window
        if(!stWinAttribs.isHide) {
            SetFocus(hMaster.hWindow);
        }
    }
    isUpdate = true;
    stMessageList.append(StMessageList::MSG_RESIZE);
    stMessageList.append(StMessageList::MSG_FULLSCREEN_SWITCH);
}

void StWindowImpl::updateWindowPos() {
    if(stWinAttribs.isSlave && !stWinAttribs.isSlaveHide && (!isSlaveIndependent() || monitors.size() > 1)) {
        HWND afterHWND = hMaster.hWindow;
        if(stWinAttribs.isSlaveHLineBottom || stWinAttribs.isSlaveHTop2Px || stWinAttribs.isSlaveHLineTop) {
            afterHWND = HWND_TOPMOST;
        }

        // resize Slave GL-window
        SetWindowPos(hSlave.hWindowGl, afterHWND,
                     getSlaveLeft(), getSlaveTop(),
                     getSlaveWidth(), getSlaveHeight(),
                     SWP_NOACTIVATE);
    }

    if(!stWinAttribs.isHide) {
        // resize Master GL-subwindow
        GLsizei sizeX = (stWinAttribs.isFullScreen) ? rectFull.width()  : rectNorm.width();
        GLsizei sizeY = (stWinAttribs.isFullScreen) ? rectFull.height() : rectNorm.height();
        SetWindowPos(hMaster.hWindowGl,
                     (nativeParent != NULL && stWinAttribs.isFullScreen) ? HWND_TOPMOST : HWND_TOP,
                     0, 0, sizeX, sizeY,
                     SWP_NOACTIVATE);
    }

    // detect when window moved to another monitor
    if(!stWinAttribs.isFullScreen && monitors.size() > 1) {
        int aNewMonId = monitors[rectNorm.center()].getId();
        if(myWinOnMonitorId != aNewMonId) {
            stMessageList.append(StMessageList::MSG_WIN_ON_NEW_MONITOR);
            myWinOnMonitorId = aNewMonId;
        }
    }
}

// Function set to argument-buffer given events and return events number
void StWindowImpl::callback(StMessage_t* stMessages) {
    // detect embedded window was moved
    if(nativeParent != NULL && hMaster.hWindowGl != NULL && !stWinAttribs.isFullScreen) {
        myPointTest.x = 0;
        myPointTest.y = 0;
        if(ClientToScreen(hMaster.hWindowGl, &myPointTest) != FALSE
        && (myPointTest.x != rectNorm.left() || myPointTest.y != rectNorm.top())) {
            updateChildRect();
        }
    }

    if(isUpdate) {
        SetWindowTextW(hMaster.hWindow, windowName.toUtfWide().toCString());
        updateWindowPos();
        isUpdate = false;
    }

    // get callback for current thread - just null cycle...
    if(PeekMessage(&event, NULL, 0, 0, PM_REMOVE)) {// Is There A Message Waiting in MAIN (non-GUI) thread?
        if(event.message == WM_QUIT) {              // Have We Received A Quit Message?
            //isOut = true;                         // If So done=TRUE
        } else {
            TranslateMessage(&event);               // Translate The Message
            DispatchMessage(&event);                // Dispatch The Message
        }
    }

    StPointD_t newMousePt = getMousePos();
    if(newMousePt.x() >= 0.0 && newMousePt.x() <= 1.0 && newMousePt.y() >= 0.0 && newMousePt.y() <= 1.0) {
        StPointD_t aDspl = newMousePt - mousePt;
        if(std::abs(aDspl.x()) >= 0.0008 || std::abs(aDspl.y()) >= 0.0008) {
            stMessageList.append(StMessageList::MSG_MOUSE_MOVE);
        }
    }
    mousePt = newMousePt;

    // TODO (Kirill Gavrilov#5#) parse multimedia keys
    //if(GetAsyncKeyState(VK_MEDIA_NEXT_TRACK ) != 0) {
    //
    //}
    stMessageList.popList(stMessages);
}

#endif
