/**
 * Copyright © 2009-2011 Kirill Gavrilov <kirill@sview.ru>
 *
 * Distributed under the Boost Software License, Version 1.0.
 * See accompanying file license-boost.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt
 */

#include <StGLWidgets/StGLMenu.h>
#include <StGLWidgets/StGLMenuCheckbox.h>
#include <StGLWidgets/StGLMenuItem.h>
#include <StGLWidgets/StGLMenuProgram.h>
#include <StGLWidgets/StGLMenuRadioButton.h>
#include <StGLWidgets/StGLRootWidget.h>

namespace {
    static const StString CLASS_NAME("StGLMenu");
    static const size_t SHARE_PROGRAM_ID = StGLRootWidget::generateShareId();
};

void StGLMenu::DeleteWithSubMenus(StGLMenu* stMenu) {
    if(stMenu == NULL) {
        return;
    }
    for(StGLWidget* child = stMenu->getChildren()->getStart(); child != NULL; child = child->getNext()) {
        StGLMenuItem* item = (StGLMenuItem* )child;
        if(item->getSubMenu() != NULL) {
            DeleteWithSubMenus(item->getSubMenu());
        }
    }
    delete stMenu;
}

StGLMenu::StGLMenu(StGLWidget* theParent,
                   const int theLeft,
                   const int theTop,
                   const int theOrient,
                   const bool theIsRootMenu)
: StGLWidget(theParent,
             theLeft, theTop,
             StGLCorner(ST_VCORNER_TOP, ST_HCORNER_LEFT),
             32, 32),
  myProgram(getRoot()->getShare(SHARE_PROGRAM_ID)),
  stVertexBuf(),
  stColorVec(0.855f, 0.855f, 0.855f, 1.0f),
  orient(theOrient),
  itemHeight(32), width(0),
  bRootMenu(theIsRootMenu),
  bActive(!theIsRootMenu),
  isInitialized(false) {
    //
}

const StString& StGLMenu::getClassName() {
    return CLASS_NAME;
}

void StGLMenu::setVisibility(bool isVisible, bool isForce) {
    StGLWidget::setVisibility(isVisible, isForce);
    for(StGLWidget* child = getChildren()->getStart(); child != NULL; child = child->getNext()) {
        child->setVisibility(isVisible, isForce);
        if(!isVisible) {
            ((StGLMenuItem* )child)->setSelected(false);
        }
    }
}

void StGLMenu::stglResize() {
    StArray<StGLVec2> aVertices(4);
    getRectGl(aVertices);
    stVertexBuf.init(aVertices);

    myProgram->use();
        myProgram->setProjMat(getRoot()->getScreenProjection());
    myProgram->unuse();
}

void StGLMenu::stglResize(const StRectI_t& winRectPx) {
    StGLWidget::stglResize(winRectPx);
    stglResize();
}

void StGLMenu::stglUpdateSubmenuLayout() {
    stglInit();
    for(StGLWidget* child = getChildren()->getStart(); child != NULL; child = child->getNext()) {
        StGLMenuItem* aMenuItem = (StGLMenuItem* )child;
        if(aMenuItem->getSubMenu() != NULL) {
            aMenuItem->getSubMenu()->stglUpdateSubmenuLayout();
        }
    }
}

bool StGLMenu::stglInit() {
    width = 0;
    isInitialized = StGLWidget::stglInit();
    if(!isInitialized) {
        return false;
    }
    for(StGLWidget* child = getChildren()->getStart(); child != NULL; child = child->getNext()) {
        StGLMenuItem* item = (StGLMenuItem* )child;
        int itemW = item->computeTextWidth();
        if(orient == MENU_HORIZONTAL) {
            item->changeRectPx().moveLeftTo(width);
            item->changeRectPx().right() = item->getRectPx().left() + itemW;
            item->setTextWidth(itemW);
            width += itemW;
        } else {
            width = (itemW > width) ? itemW : width;
        }
        if(item->getSubMenu() != NULL) {
            if(orient == MENU_HORIZONTAL) {
                item->getSubMenu()->changeRectPx().moveTopLeftTo(item->getRectPxAbsolute().left(), item->getRectPxAbsolute().bottom());
            } else {
                item->getSubMenu()->changeRectPx().moveTopLeftTo(item->getRectPxAbsolute().right() - 10, item->getRectPxAbsolute().top());
            }
        }
    }
    StGLWidget* childLast = getChildren()->getLast();
    if(childLast != NULL) {
        changeRectPx().right()  = getRectPx().left() + childLast->getRectPx().right();
        changeRectPx().bottom() = getRectPx().top()  + childLast->getRectPx().bottom();
    }
    if(orient == MENU_VERTICAL) {
        width += 32; // icon offset
        changeRectPx().right() = getRectPx().left() + width;
        int itemCount = 0;
        for(StGLWidget* child = getChildren()->getStart(); child != NULL; child = child->getNext(), ++itemCount) {
            StGLMenuItem* item = (StGLMenuItem* )child;
            item->changeRectPx().moveTopTo(itemCount * itemHeight);
            item->changeRectPx().right() = item->getRectPx().left() + width;
            item->setTextWidth(width);
            if(item->getSubMenu() != NULL) {
                item->getSubMenu()->changeRectPx().moveTopLeftTo(getRectPxAbsolute().right() - 10, item->getRectPxAbsolute().top());
            }
        }
        changeRectPx().bottom() = getRectPx().top() + itemCount * itemHeight;
    }

    // already initialized?
    if(stVertexBuf.isValid()) {
        return true;
    }

    // initialize GLSL program
    if(myProgram.isNull()) {
        myProgram.create(new StGLMenuProgram());
        if(!myProgram->init()) {
            isInitialized = false;
            return isInitialized;
        }
    }

    StArray<StGLVec2> aDummyVert(4);
    stVertexBuf.init(aDummyVert);
    stglResize();

    return isInitialized;
}

void StGLMenu::stglDraw(unsigned int view) {
    if(!isInitialized || !isVisible()) {
        return;
    }

    if(isResized) {
        stglResize();
        isResized = false;
    }

    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);
        myProgram->use();
        myProgram->setColor(stColorVec, GLfloat(opacityValue));

            stVertexBuf.bindVertexAttrib(myProgram->getVVertexLoc());
            glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
            stVertexBuf.unBindVertexAttrib(myProgram->getVVertexLoc());

        myProgram->unuse();
    glDisable(GL_BLEND);

    StGLWidget::stglDraw(view);
}

bool StGLMenu::tryUnClick(const StPointD_t& cursorZo, const int& mouseBtn, bool& isItemUnclicked) {
    bool someClickedBefore = isItemUnclicked;
    bool selfClicked = StGLWidget::tryUnClick(cursorZo, mouseBtn, isItemUnclicked);
    bool selfItemClicked = !someClickedBefore && isItemUnclicked;
    if(isRootMenu() && !selfItemClicked) {
        setActive(false); // disactivate root menu
    }
    for(StGLWidget* child = getChildren()->getStart(); child != NULL; child = child->getNext()) {
        StGLMenuItem* item = (StGLMenuItem* )child;
        item->setSelected(false);
    }
    return selfClicked;
}

StGLMenuItem* StGLMenu::addItem(const StString& label, const size_t userData) {
    StGLMenuItem* newMenuItem = new StGLMenuItem(this, 0, 0);
    newMenuItem->setText(label);
    newMenuItem->setUserData(userData);
    return newMenuItem;
}

StGLMenuItem* StGLMenu::addItem(const StString& label, StGLMenu* subMenu) {
    StGLMenuItem* newMenuItem = new StGLMenuItem(this, 0, 0, subMenu);
    newMenuItem->setText(label);
    return newMenuItem;
}

StGLMenuItem* StGLMenu::addItem(const StString& theLabel,
                                const StHandle<StBoolParam>& theTrackedValue) {
    StGLMenuItem* aMenuItem = new StGLMenuCheckbox(this, theTrackedValue);
    aMenuItem->setText(theLabel);
    return aMenuItem;
}

StGLMenuItem* StGLMenu::addItem(const StString& theLabel,
                                const StHandle<StInt32Param>& theTrackedValue,
                                const int32_t theOnValue) {
    StGLMenuItem* aMenuItem = new StGLMenuRadioButton(this, theTrackedValue, theOnValue);
    aMenuItem->setText(theLabel);
    return aMenuItem;
}

StGLMenuItem* StGLMenu::addItem(const StString& theLabel,
                                const StHandle<StFloat32Param>& theTrackedValue,
                                const float theOnValue) {
    StGLMenuItem* aMenuItem = new StGLMenuRadioButton(this, theTrackedValue, theOnValue);
    aMenuItem->setText(theLabel);
    return aMenuItem;
}
