/**
 * Copyright © 2009-2011 Kirill Gavrilov <kirill@sview.ru>
 *
 * Distributed under the Boost Software License, Version 1.0.
 * See accompanying file license-boost.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt
 */

#include <StGLWidgets/StGLMenu.h>
#include <StGLWidgets/StGLMenuItem.h>
#include <StGLWidgets/StGLMenuProgram.h>
#include <StGLWidgets/StGLRootWidget.h>

namespace {
    static const StString CLASS_NAME("StGLMenuItem");
    static const size_t SHARE_PROGRAM_ID = StGLRootWidget::generateShareId();
};

void StGLMenuItem::DeleteWithSubMenus(StGLMenuItem* stMenuItem) {
    if(stMenuItem == NULL) {
        return;
    }
    if(stMenuItem->getSubMenu() != NULL) {
        StGLMenu::DeleteWithSubMenus(stMenuItem->getSubMenu());
        stMenuItem->setSubMenu(NULL);
    }
    delete stMenuItem;
}

StGLMenuItem::StGLMenuItem(StGLMenu* theParent,
                           const int theLeft, const int theTop,
                           StGLMenu* theSubMenu)
: StGLTextArea(theParent,
               theLeft, theTop,
               StGLCorner(ST_VCORNER_TOP, ST_HCORNER_LEFT)),
  subMenu(theSubMenu),
  myProgram(getRoot()->getShare(SHARE_PROGRAM_ID)),
  stBackVertexBuf(),
  stBackColor(),
  bItemSelected(false) {
    //
    StGLWidget::signals.onMouseUnclick.connect(this, &StGLMenuItem::doMouseUnclick);

    setTextAligment(ALIGN_LEFT);
    setBorder(false);
    stBackColor[StGLMenuItem::HIGHLIGHT] = StGLVec4(0.765f, 0.765f, 0.765f, 1.0f);
    stBackColor[StGLMenuItem::CLICKED]   = StGLVec4(0.500f, 0.500f, 0.500f, 1.0f);
}

const StString& StGLMenuItem::getClassName() {
    return CLASS_NAME;
}

const int StGLMenuItem::computeTextWidth() {
    // TODO (Kirill Gavrilov#1) browser crashed when multiple plugin's instances loaded on one page!
    /**if(stglInit()) {
        // fontLayout must be initialized!
        FTBBox textBox = fontLayout.BBox(text.utfText());
        GLdouble textAreaW = textBox.Upper().X() - textBox.Lower().X();
        GLdouble textAreaH = textBox.Upper().Y() - textBox.Lower().Y();
        return (int )((8.0 + textAreaW) * (GLdouble )getFontSize() / textAreaH);
    }
    return 1;*/
    return int(10 * (text.getLength() + 2));
}

void StGLMenuItem::stglResize() {
    // back vertices
    StArray<StGLVec2> aVertices(4);
    getRectGl(aVertices);
    stBackVertexBuf.init(aVertices);

    // update projection matrix
    myProgram->use();
        myProgram->setProjMat(getRoot()->getScreenProjection());
    myProgram->unuse();
}

void StGLMenuItem::stglResize(const StRectI_t& winRectPx) {
    StGLTextArea::stglResize(winRectPx);
    stglResize();
}

void StGLMenuItem::stglUpdate(const StPointD_t& cursorZo) {
    StGLTextArea::stglUpdate(cursorZo);
    if(!isInitialized || !isVisible()) {
        return;
    }
    if(isClicked(ST_MOUSE_LEFT) || (isSelected() && hasSubMenu())) {
        return;
    }
    if(isPointIn(getRoot()->getCursorZo())) {
        if(getParentMenu()->isActive()) {
            setSelected(true);
        }
    }
}

bool StGLMenuItem::stglInit() {
    isInitialized = StGLTextArea::stglInit();
    if(!isInitialized) {
        return false;
    }

    // already initialized?
    if(stBackVertexBuf.isValid()) {
        return true;
    }

    // initialize GLSL program
    if(myProgram.isNull()) {
        myProgram.create(new StGLMenuProgram());
        if(!myProgram->init()) {
            isInitialized = false;
            return isInitialized;
        }
    }

    StArray<StGLVec2> aDummyVert(4);
    stBackVertexBuf.init(aDummyVert);

    stglResize();
    return isInitialized;
}

void StGLMenuItem::stglDraw(unsigned int view) {
    if(!isInitialized || !isVisible()) {
        return;
    }

    textPoint.y() = (getRectPx().height() - getFontSize()) / 2;
    textPoint.x() = getParentMenu()->getOrient() == StGLMenu::MENU_VERTICAL ? 32 : 0;

    StGLMenuItem::State state = StGLMenuItem::PASSIVE;
    if(isClicked(ST_MOUSE_LEFT) || (isSelected() && hasSubMenu())) {
        state = StGLMenuItem::CLICKED;
    } else if(isPointIn(getRoot()->getCursorZo())) {
        state = StGLMenuItem::HIGHLIGHT;
    }

    if(isResized) {
        stglResize();
        isResized = false;
    }

    if(state != StGLMenuItem::PASSIVE) {
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        glEnable(GL_BLEND);
        myProgram->use();

        if(state != StGLMenuItem::PASSIVE) {
            myProgram->setColor(stBackColor[state], GLfloat(opacityValue));
            stBackVertexBuf.bindVertexAttrib(myProgram->getVVertexLoc());
                glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
            stBackVertexBuf.unBindVertexAttrib(myProgram->getVVertexLoc());
        }

        myProgram->unuse();
        glDisable(GL_BLEND);
    }

    StGLTextArea::stglDraw(view);
}

void StGLMenuItem::setSelected(bool isItemSelected) {
    if(isItemSelected) {
        for(StGLWidget* child = getParent()->getChildren()->getStart(); child != NULL; child = child->getNext()) {
            if(child != this) {
                ((StGLMenuItem* )child)->setSelected(false);
            }
        }
        if(hasSubMenu()) {
            subMenu->setVisibility(true, true);
        }
    } else {
        if(hasSubMenu()) {
            subMenu->setVisibility(false, true);
        }
    }
    this->bItemSelected = isItemSelected;
}

bool StGLMenuItem::tryClick(const StPointD_t& cursorZo, const int& mouseBtn, bool& isItemClicked) {
    if(StGLWidget::tryClick(cursorZo, mouseBtn, isItemClicked)) {
        isItemClicked = true; // always clickable widget
        return true;
    }
    return false;
}

bool StGLMenuItem::tryUnClick(const StPointD_t& cursorZo, const int& mouseBtn, bool& isItemUnclicked) {
    if(StGLWidget::tryUnClick(cursorZo, mouseBtn, isItemUnclicked)) {
        isItemUnclicked = true; // always clickable widget
        if(getParentMenu()->isRootMenu()) {
            getParentMenu()->setActive(true); // activate root menu
        }
        return true;
    }
    return false;
}

void StGLMenuItem::doMouseUnclick(const int btnId) {
    if(btnId == ST_MOUSE_LEFT) {
        signals.onItemClick(getUserData());
    }
}
