/**
 * Copyright © 2009-2010 Kirill Gavrilov <kirill@sview.ru
 *
 * Distributed under the Boost Software License, Version 1.0.
 * See accompanying file license-boost.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt
 */

#include <StGLWidgets/StGLMessageBox.h>

StGLMessageBox::StGLMessageBox(StGLWidget* parent, const StString& text,
                               const int& width, const int& height)
: StGLWidget(parent, 32, 32, StGLCorner(ST_VCORNER_TOP, ST_HCORNER_LEFT), width, height),
  stTextArea(NULL),
  stProgram(),
  stVertexBuf() {
    //
    StGLWidget::signals.onMouseUnclick.connect(this, &StGLMessageBox::doMouseUnclick);

    static const int OFFSET_PIXELS = 32;
    int offsetX = width  > 2 * OFFSET_PIXELS ? OFFSET_PIXELS : 0;
    int offsetY = height > 2 * OFFSET_PIXELS ? OFFSET_PIXELS : 0;
    stTextArea = new StGLTextArea(this, offsetX, offsetY, StGLCorner(ST_VCORNER_TOP, ST_HCORNER_LEFT),
                                  width - 2 * offsetX, height - 2 * offsetY, true);
    stTextArea->setText(text);
    stTextArea->setBorder(false);
    stTextArea->setTextColor(1.0f, 1.0f, 1.0f);
}

StGLMessageBox::~StGLMessageBox() {
    //
}

bool StGLMessageBox::stglInit() {
    if(!StGLWidget::stglInit()) {
        return false;
    }

    const GLfloat QUAD_VERTICES[4 * 4] = {
         1.0f, -1.0f, 0.0f, 1.0f, // top-right
         1.0f,  1.0f, 0.0f, 1.0f, // bottom-right
        -1.0f, -1.0f, 0.0f, 1.0f, // top-left
        -1.0f,  1.0f, 0.0f, 1.0f  // bottom-left
    };
    stVertexBuf.init(4, 4, QUAD_VERTICES);
    stglResize();

    if(!stProgram.init()) {
        return false;
    }
    return true;
}

void StGLMessageBox::stglResize() {
    // move to the center
    int origW = getRectPx().width();
    int origH = getRectPx().height();
    int parentW = getParent()->getRectPx().width();
    int parentH = getParent()->getRectPx().height();
    changeRectPx().left()   = parentW / 2 - origW / 2;
    changeRectPx().right()  = getRectPx().left() + origW;
    changeRectPx().top()    = parentH / 2 - origH / 2;
    changeRectPx().bottom() = getRectPx().top() + origH;

    GLfloat toZScreen = -getCamera()->getZScreen();

    StRectD_t butRectGl = getRectGl();
    const GLfloat quadVertices[4 * 4] = {
        GLfloat(butRectGl.right()), GLfloat(butRectGl.top()),    toZScreen, 1.0f, // top-right
        GLfloat(butRectGl.right()), GLfloat(butRectGl.bottom()), toZScreen, 1.0f, // bottom-right
        GLfloat(butRectGl.left()),  GLfloat(butRectGl.top()),    toZScreen, 1.0f, // top-left
        GLfloat(butRectGl.left()),  GLfloat(butRectGl.bottom()), toZScreen, 1.0f  // bottom-left
    };
    stVertexBuf.init(4, 4, quadVertices);
}

void StGLMessageBox::stglResize(const StRectI_t& winRectPx) {
    StGLWidget::stglResize(winRectPx);
    stglResize();
}

void StGLMessageBox::stglDraw(unsigned int view) {
    if(!isVisible()) {
        return;
    }

    if(isResized) {
        stglResize();
        isResized = false;
    }

    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);

    stProgram.use();
    stProgram.setProjMat(getCamera()->getProjMatrix());
    stProgram.setColor(StGLVec4(0.06f, 0.06f, 0.06f, 1.0f), GLfloat(opacityValue) * 0.8f);

        stVertexBuf.bindVertexAttrib(stProgram.getVVertexLoc());
        glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
        stVertexBuf.unBindVertexAttrib(stProgram.getVVertexLoc());

    stProgram.unuse();

    glDisable(GL_BLEND);

    StGLWidget::stglDraw(view); // draw children
}

void StGLMessageBox::setVisibility(bool isVisible, bool isForce) {
    StGLWidget::setVisibility(isVisible, isForce);
    for(StGLWidget* child = getChildren()->getStart(); child != NULL; child = child->getNext()) {
        child->setVisibility(isVisible, isForce);
    }
}

bool StGLMessageBox::tryClick(const StPointD_t& cursorZo, const int& mouseBtn, bool& isItemClicked) {
    if(StGLWidget::tryClick(cursorZo, mouseBtn, isItemClicked)) {
        isItemClicked = true;
        return true;
    }
    return false;
}

bool StGLMessageBox::tryUnClick(const StPointD_t& cursorZo, const int& mouseBtn, bool& isItemUnclicked) {
    if(StGLWidget::tryUnClick(cursorZo, mouseBtn, isItemUnclicked)) {
        isItemUnclicked = true;
        return true;
    }
    return false;
}

void StGLMessageBox::doKillSelf(const size_t ) {
    delete this; /// TODO (Kirill Gavrilov#2) danger call!!!
}

void StGLMessageBox::doMouseUnclick(const int btnId) {
    if(btnId == ST_MOUSE_LEFT) {
        signals.onClickLeft(getUserData());
    } else if(btnId == ST_MOUSE_RIGHT) {
        signals.onClickRight(getUserData());
    }
}
