/**
 * Copyright © 2009-2011 Kirill Gavrilov <kirill@sview.ru>
 *
 * Distributed under the Boost Software License, Version 1.0.
 * See accompanying file license-boost.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt
 */

#include <StGLWidgets/StGLTextArea.h>
#include <StGLWidgets/StGLRootWidget.h>

#include <StGL/StGLProgram.h>
#include <StFile/StFileNode.h>
#include <StThreads/StProcess.h>

class StGLTextArea::StTextProgram : public StGLProgram {

        private:

    StGLVarLocation uniProjMatLoc;
    StGLVarLocation uniModelMatLoc;
    StGLVarLocation uniTextColorLoc;

    StGLVarLocation atrVVertexLoc;
    StGLVarLocation atrVTexCoordLoc;

        public:

    StTextProgram()
    : StGLProgram("StGLTextArea, Text Program"),
      uniProjMatLoc(),
      uniModelMatLoc(),
      uniTextColorLoc(),
      atrVVertexLoc(),
      atrVTexCoordLoc() {
        //
    }

    virtual ~StTextProgram() {
        //
    }

    StGLVarLocation getVVertexLoc() const {
        return atrVVertexLoc;
    }

    StGLVarLocation getVTexCoordLoc() const {
        return atrVTexCoordLoc;
    }

    void setProjMat(const StGLMatrix& stProjMat) {
        glUniformMatrix4fv(uniProjMatLoc, 1, GL_FALSE, stProjMat);
    }

    void setModelMat(const StGLMatrix& stModelMat) {
        glUniformMatrix4fv(uniModelMatLoc, 1, GL_FALSE, stModelMat);
    }

    void setTextColor(const StGLVec4& textColorVec) {
        glUniform4fv(uniTextColorLoc, 1, textColorVec);
    }

    virtual bool init() {
        const char VERTEX_SHADER[] =
           "uniform mat4 uProjMat; \
            uniform mat4 uModelMat; \
            attribute vec4 vVertex; \
            attribute vec2 vTexCoord; \
            varying vec2 fTexCoord; \
            void main(void) { \
                fTexCoord = vTexCoord; \
                gl_Position = uProjMat * uModelMat * vVertex; \
            }";

        const char FRAGMENT_SHADER[] =
           "uniform sampler2D uTexture; \
            uniform vec4 uTextColor; \
            varying vec2 fTexCoord; \
            void main(void) { \
                vec4 color = uTextColor; \
                color.a *= texture2D(uTexture, fTexCoord).a; \
                gl_FragColor = color; \
            }";

        StGLVertexShader stVertexShader(StGLProgram::getTitle());
        stVertexShader.init(VERTEX_SHADER);

        StGLFragmentShader stFragmentShader(StGLProgram::getTitle());
        stFragmentShader.init(FRAGMENT_SHADER);
        if(!StGLProgram::create().attachShader(stVertexShader).attachShader(stFragmentShader).link()) {
            return false;
        }

        uniProjMatLoc   = StGLProgram::getUniformLocation("uProjMat");
        uniModelMatLoc  = StGLProgram::getUniformLocation("uModelMat");
        uniTextColorLoc = StGLProgram::getUniformLocation("uTextColor");
        atrVVertexLoc   = StGLProgram::getAttribLocation("vVertex");
        atrVTexCoordLoc = StGLProgram::getAttribLocation("vTexCoord");

        StGLVarLocation uniTextureLoc = StGLProgram::getUniformLocation("uTexture");
        if(uniTextureLoc.isValid()) {
            StGLProgram::use();
            glUniform1i(uniTextureLoc, StGLProgram::TEXTURE_SAMPLE_0);
            StGLProgram::unuse();
        }

        return uniProjMatLoc.isValid()
            && uniModelMatLoc.isValid()
            && uniTextColorLoc.isValid()
            && atrVVertexLoc.isValid()
            && atrVTexCoordLoc.isValid()
            && uniTextureLoc.isValid();
    }

};

class StGLTextArea::StBorderProgram : public StGLProgram {

        private:

    StGLVarLocation uniProjMatLoc;
    StGLVarLocation uniModelMatLoc;
    StGLVarLocation uniColorLoc;

    StGLVarLocation atrVVertexLoc;

        public:

    StBorderProgram()
    : StGLProgram("StGLTextArea, Border Program"),
      uniProjMatLoc(),
      uniModelMatLoc(),
      uniColorLoc(),
      atrVVertexLoc() {
        //
    }

    virtual ~StBorderProgram() {
        //
    }

    StGLVarLocation getVVertexLoc() const {
        return atrVVertexLoc;
    }

    void setProjMat(const StGLMatrix& stProjMat) {
        glUniformMatrix4fv(uniProjMatLoc, 1, GL_FALSE, stProjMat);
    }

    void setModelMat(const StGLMatrix& stModelMat) {
        glUniformMatrix4fv(uniModelMatLoc, 1, GL_FALSE, stModelMat);
    }

    void setColor(const StGLVec4& colorVec) {
        glUniform4fv(uniColorLoc, 1, colorVec);
    }

    virtual bool init() {
        const char VERTEX_SHADER[] =
           "uniform mat4 uProjMat; \
            uniform mat4 uModelMat; \
            attribute vec4 vVertex; \
            void main(void) { \
                gl_Position = uProjMat * uModelMat * vVertex; \
            }";

        const char FRAGMENT_SHADER[] =
           "uniform vec4 uColor; \
            void main(void) { \
                gl_FragColor = uColor; \
            }";

        StGLVertexShader stVertexShader(StGLProgram::getTitle());
        stVertexShader.init(VERTEX_SHADER);

        StGLFragmentShader stFragmentShader(StGLProgram::getTitle());
        stFragmentShader.init(FRAGMENT_SHADER);
        if(!StGLProgram::create().attachShader(stVertexShader).attachShader(stFragmentShader).link()) {
            return false;
        }

        uniProjMatLoc  = StGLProgram::getUniformLocation("uProjMat");
        uniModelMatLoc = StGLProgram::getUniformLocation("uModelMat");
        uniColorLoc    = StGLProgram::getUniformLocation("uColor");
        atrVVertexLoc  = StGLProgram::getAttribLocation("vVertex");
        return uniProjMatLoc.isValid() && uniModelMatLoc.isValid() && uniColorLoc.isValid() && atrVVertexLoc.isValid();
    }

};

namespace {
    static const StString CLASS_NAME("StGLTextArea");
    static const size_t SHARE_TEXT_PROGRAM_ID   = StGLRootWidget::generateShareId();
    static const size_t SHARE_BORDER_PROGRAM_ID = StGLRootWidget::generateShareId();
    static const size_t SHARE_FONT_NORMAL_ID    = StGLRootWidget::generateShareId();
    static const size_t SHARE_FONT_SMALL_ID     = StGLRootWidget::generateShareId();
    static const size_t SHARE_FONT_BIG_ID       = StGLRootWidget::generateShareId();
};

StGLTextArea::StGLTextArea(StGLWidget* theParent,
                           const int theLeft, const int theTop,
                           const StGLCorner theCorner,
                           const int theWidth, const int theHeight, bool isCutView,
                           const FontSize theSize,
                           const bool toDrawShadow,
                           const bool toScale)
: StGLWidget(theParent, theLeft, theTop, theCorner, theWidth, theHeight),
  stTextVertBuf(),
  stTextTCrdBuf(),
  stBorderIVertBuf(),
  stBorderOVertBuf(),
  stTextProgram(getRoot()->getShare(SHARE_TEXT_PROGRAM_ID)),
  stBorderProgram(getRoot()->getShare(SHARE_BORDER_PROGRAM_ID)),
  ftFont(getRoot()->getShare((theSize == SIZE_NORMAL) ? SHARE_FONT_NORMAL_ID
                          : ((theSize == SIZE_SMALL)  ? SHARE_FONT_SMALL_ID : SHARE_FONT_BIG_ID))),
  fontLayout(),
  text(),
  areaDirection(TO_RIGHT | TO_DOWN),
  mySize(theSize),
  myTextColor(0.0f, 0.0f, 0.0f, 1.0f),
  myShadowColor(0.0f, 0.0f, 0.0f, 1.0f),
  backColor(0.365f, 0.722f, 1.0f, 1.0f),
  borderColor(0.0f, 0.0f, 0.0f, 1.0f),
  textPoint(3, 3),
  toRecomputeText(true),
  isCutView(isCutView),
  isShowAreaBorder(false),
  myToDrawShadow(toDrawShadow),
  myToScale(toScale),
  isInitialized(false) {
    //
    fontLayout.SetAlignment(FTGL::ALIGN_CENTER);
}

StGLTextArea::~StGLTextArea() {
    //
}

const StString& StGLTextArea::getClassName() {
    return CLASS_NAME;
}

const StString& StGLTextArea::getText() const {
    return text;
}

void StGLTextArea::setText(const StString& text) {
    if(this->text != text) {
        this->text = text;
        toRecomputeText = true;
    }
}

void StGLTextArea::setTextAligment(const int& aligment) {
    fontLayout.SetAlignment(FTGL::TextAlignment(aligment));
}

void StGLTextArea::setTextWidth(const int theWidth) {
    fontLayout.SetLineLength((float )theWidth);
    toRecomputeText = true;
}

enum {
    ST_FONT_SERIF,
    ST_FONT_SANS,
    ST_FONT_MONO,
};

static StString getFont(const int theFontType) {
#if (defined(_WIN32) || defined(__WIN32__))
    switch(theFontType) {
        case ST_FONT_SERIF:
            return "times.ttf";
        case ST_FONT_SANS:
            return "micross.ttf";
        case ST_FONT_MONO:
        default:
            return "tahoma.ttf";
    }
#elif (defined(__APPLE__))
    switch(theFontType) {
        case ST_FONT_SERIF:
            return "Times.dfont";
        case ST_FONT_SANS:
            return "Trebuchet MS.ttf";
            //return "LucidaGrande.ttc";
            //return "Geneva.dfont";
            //return "HelveticaNeue.dfont";
            //return "Helvetica.dfont";
        case ST_FONT_MONO:
        default:
            return "Monaco.dfont";
    }
#elif (defined(__linux__) || defined(__linux))
    switch(theFontType) {
        case ST_FONT_SERIF:
            return "DejaVuSerif.ttf";
        case ST_FONT_SANS:
            return "DejaVuSans.ttf";
        case ST_FONT_MONO:
        default:
            return "DejaVuSansMono.ttf";
    }
#else
    #error undefined fonts
#endif
}

bool StGLTextArea::stglInit() {
    if(isInitialized) {
        return true;
    }

    // initialize FTGL font
    static const StString ST_FONT_SANS_PATH = StProcess::getFontsRoot() + getFont(ST_FONT_SANS);
    if(ftFont.isNull()) {
        StStringUtf8 aFontPath = StFileNode::getCompatibleName(ST_FONT_SANS_PATH);
        ftFont.create(new FTTextureFont(aFontPath.toCString()));
        if(ftFont->Error()) {
            ST_ERROR_LOG("Could not load font '" + ST_FONT_SANS_PATH + '\'');
            return false;
        }
        // setup font size only once!
        // to use another size you should create new shared resource!
        ftFont->FaceSize(getFontSize());
        ftFont->CharMap(ft_encoding_unicode);
    }

    if(ftFont->Error()) {
        return false;
    }

    fontLayout.SetFont(ftFont.operator->());
    if(fontLayout.Error()) {
        stError(StString("Font layout error"));
        return false;
    }
    fontLayout.SetLineLength((float )getRectPx().width());

    // initialize text program
    if(stTextProgram.isNull()) {
        stTextProgram.create(new StTextProgram());
        if(!stTextProgram->init()) {
            return false;
        }
    } else if(!stTextProgram->isValid()) {
        return false;
    }

    // initialize border program
    if(stBorderProgram.isNull()) {
        stBorderProgram.create(new StBorderProgram());
        if(!stBorderProgram->init()) {
            return false;
        }
    } else if(!stBorderProgram->isValid()) {
        return false;
    }

    isInitialized = true;

    stBorderIVertBuf.init();
    stBorderOVertBuf.init();

    return StGLWidget::stglInit();
}

void StGLTextArea::recomputeBorder() {
    GLfloat textAreaW = GLfloat(getRectPx().width());
    const GLfloat quadVerticesInner[4 * 4] = {
        3.0f + textAreaW, ftBox.Upper().Yf() + 3.0f, 0.0f, 1.0f, // top-right
        3.0f + textAreaW, ftBox.Lower().Yf() - 3.0f, 0.0f, 1.0f, // bottom-right
                   -3.0f, ftBox.Upper().Yf() + 3.0f, 0.0f, 1.0f, // top-left
                   -3.0f, ftBox.Lower().Yf() - 3.0f, 0.0f, 1.0f  // bottom-left
    };
    const GLfloat quadVerticesOuter[4 * 4] = {
        4.0f + textAreaW, ftBox.Upper().Yf() + 4.0f, 0.0f, 1.0f, // top-right
        4.0f + textAreaW, ftBox.Lower().Yf() - 4.0f, 0.0f, 1.0f, // bottom-right
                   -4.0f, ftBox.Upper().Yf() + 4.0f, 0.0f, 1.0f, // top-left
                   -4.0f, ftBox.Lower().Yf() - 4.0f, 0.0f, 1.0f  // bottom-left
    };
    stBorderIVertBuf.init(4, 4, quadVerticesInner);
    stBorderOVertBuf.init(4, 4, quadVerticesOuter);
}

void StGLTextArea::drawText() {
    for(size_t index = 0; index < stTextVertBuf.size(); ++index) {
        if(stTextVertBuf[index].getElemsCount() < 1) {
            continue;
        }
        glBindTexture(GL_TEXTURE_2D, ftFont->getTextureId(index));

        stTextVertBuf[index].bindVertexAttrib(stTextProgram->getVVertexLoc());
        stTextTCrdBuf[index].bindVertexAttrib(stTextProgram->getVTexCoordLoc());

            glDrawArrays(GL_TRIANGLES, 0, GLsizei(stTextVertBuf[index].getElemsCount()));

        stTextTCrdBuf[index].unBindVertexAttrib(stTextProgram->getVTexCoordLoc());
        stTextVertBuf[index].unBindVertexAttrib(stTextProgram->getVVertexLoc());
    }
}

void StGLTextArea::stglDraw(unsigned int theView) {
    if(!isInitialized || !isVisible()) {
        return;
    }
    myTextColor.a() = myShadowColor.a() = backColor.a() = borderColor.a() = (GLfloat )opacityValue;

    if(toRecomputeText) {
        ftFont->reset();
        fontLayout.Render(text.toCString());

        while(stTextVertBuf.size() < ftFont->getTexturesCount()) {
            stTextVertBuf.add(StGLVertexBuffer());
            stTextTCrdBuf.add(StGLVertexBuffer());
            stTextVertBuf[stTextVertBuf.size() - 1].init();
            stTextTCrdBuf[stTextTCrdBuf.size() - 1].init();
        }

        size_t vertElemSize = 0;
        size_t texCoordElemSize = 0;
        size_t elemsCount = 0;
        for(size_t index = 0; index < ftFont->getTexturesCount(); ++index) {
            const GLfloat* verticesData = ftFont->getVertices(index, &vertElemSize, &elemsCount);
            const GLfloat* texCoordsData = ftFont->getTexCoords(index, &texCoordElemSize, &elemsCount);
            /*ST_DEBUG_LOG(ST_TEXT("vertices count = ") + elemsCount
                         + ST_TEXT("; texture ") + (index + 1) + ST_TEXT(" / ") + ftFont->getTexturesCount()); */
            stTextVertBuf[index].init(vertElemSize, elemsCount, verticesData);
            stTextTCrdBuf[index].init(texCoordElemSize, elemsCount, texCoordsData);
        }

        ftBox = fontLayout.BBox(text.toCString());

        if(myToScale) {
            changeRectPx().bottom() = changeRectPx().top() + int(ftBox.Upper().Yf() - ftBox.Lower().Yf()) + 2;
        }

        if(isShowAreaBorder) {
            recomputeBorder();
        }
        toRecomputeText = false;
    }

    StRectI_t textRectPx = getRectPx();
    textRectPx.left() += textPoint.x();
    textRectPx.top()  += textPoint.y();
    StRectD_t textRectGl = getRoot()->getRectGl(getAbsolute(textRectPx));

    GLfloat textAreaW = GLfloat(getRectPx().width());
    GLfloat textAreaH = ftBox.Upper().Yf() - ftBox.Lower().Yf();

    GLfloat xx = ((areaDirection & TO_RIGHT) == TO_RIGHT) ? 0.0f : -textAreaW;
    GLfloat yy = ((areaDirection & TO_DOWN ) == TO_DOWN ) ? -ftBox.Upper().Yf() : textAreaH;

    // size corrector for FTGL
    StRectD_t zparams; getCamera()->getZParams(zparams);
    GLfloat sizeOut = 2.0f * GLfloat(zparams.top()) / GLfloat(getRoot()->getRectPx().height());

    StGLMatrix stModelMat;
    stModelMat.translate(StGLVec3(0.0f, 0.0f, -getCamera()->getZScreen()));
    stModelMat.translate(StGLVec3(GLfloat(textRectGl.left()) + xx * sizeOut,
                                  GLfloat(textRectGl.top())  + yy * sizeOut,
                                  0.0f));
    stModelMat.scale(sizeOut, sizeOut, 0.0f);

    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);

    // draw borders
    if(isShowAreaBorder) {
        if(stBorderOVertBuf.getElemsCount() == 0) {
            recomputeBorder();
        }

        stBorderProgram->use();
            stBorderProgram->setProjMat(getCamera()->getProjMatrix());
            stBorderProgram->setModelMat(stModelMat);

            stBorderProgram->setColor(borderColor);
            stBorderOVertBuf.bindVertexAttrib(stBorderProgram->getVVertexLoc());
                glDrawArrays(GL_TRIANGLE_STRIP, 0, GLsizei(stBorderOVertBuf.getElemsCount()));
            stBorderOVertBuf.unBindVertexAttrib(stBorderProgram->getVVertexLoc());

            stBorderProgram->setColor(backColor);
            stBorderIVertBuf.bindVertexAttrib(stBorderProgram->getVVertexLoc());
                glDrawArrays(GL_TRIANGLE_STRIP, 0, GLsizei(stBorderIVertBuf.getElemsCount()));
            stBorderIVertBuf.unBindVertexAttrib(stBorderProgram->getVVertexLoc());
        stBorderProgram->unuse();
    }

    // draw text
    glActiveTexture(GL_TEXTURE0); // our shader is bound to first texture unit
    stTextProgram->use();
        stTextProgram->setProjMat(getCamera()->getProjMatrix());
        stTextProgram->setModelMat(stModelMat);
        stTextProgram->setTextColor(myToDrawShadow ? myShadowColor : myTextColor);

        drawText();

        if(myToDrawShadow) {
            stModelMat.initIdentity();
            textRectPx.left() -= 1;
            textRectPx.top()  -= 1;
            textRectGl = getRoot()->getRectGl(getAbsolute(textRectPx));
            stModelMat.translate(StGLVec3(0.0f, 0.0f, -getCamera()->getZScreen()));
            stModelMat.translate(StGLVec3(GLfloat(textRectGl.left()) + xx * sizeOut,
                                          GLfloat(textRectGl.top())  + yy * sizeOut,
                                          0.0f));
            stModelMat.scale(sizeOut, sizeOut, 0.0f);

            stTextProgram->setModelMat(stModelMat);
            stTextProgram->setTextColor(myTextColor);

            drawText();
        }

    stTextProgram->unuse();

    glDisable(GL_BLEND);

    StGLWidget::stglDraw(theView);
}
