/**
 * This source is a part of sView program.
 *
 * Copyright © Kirill Gavrilov, 2009-2011
 */

#include "StAudioQueue.h"

#include <StGL/StGLVec.h>

/**
 * Check OpenAL state.
 */
#ifdef __ST_DEBUG__
void stalCheckErrors(const StString& procedure) {
#else
void stalCheckErrors(const StString& ) {
#endif
    ALenum error = alGetError();
    switch(error) {
        case AL_NO_ERROR: return; // alright
        case AL_INVALID_NAME:      ST_DEBUG_LOG(procedure + ": AL_INVALID_NAME");      return;
        case AL_INVALID_ENUM:      ST_DEBUG_LOG(procedure + ": AL_INVALID_ENUM");      return;
        case AL_INVALID_VALUE:     ST_DEBUG_LOG(procedure + ": AL_INVALID_VALUE");     return;
        case AL_INVALID_OPERATION: ST_DEBUG_LOG(procedure + ": AL_INVALID_OPERATION"); return;
        case AL_OUT_OF_MEMORY:     ST_DEBUG_LOG(procedure + ": AL_OUT_OF_MEMORY");     return;
        default:                   ST_DEBUG_LOG(procedure + ": OpenAL unknown error"); return;
    }
}

static const StGLVec3 POSITION_CENTER       ( 0.0f, 0.0f,  0.0f);
static const StGLVec3 POSITION_FRONT_LEFT   (-1.0f, 0.0f, -1.0f);
static const StGLVec3 POSITION_FRONT_CENTER ( 0.0f, 0.0f, -1.0f);
static const StGLVec3 POSITION_FRONT_RIGHT  ( 1.0f, 0.0f, -1.0f);
static const StGLVec3 POSITION_LFE          ( 0.0f, 0.0f,  0.0f);
static const StGLVec3 POSITION_REAR_LEFT    (-1.0f, 0.0f,  1.0f);
static const StGLVec3 POSITION_REAR_RIGHT   ( 1.0f, 0.0f,  1.0f);

/**
 * Check if dynamically linked version of FFmpeg libraries
 * is old or not.
 */
inline bool isReoderingNeededInit() {
    stLibAV::Version aVersion = stLibAV::Version::libavcodec();
    // I have no idea which version introduce this feature for ac3 and ogg streams...
    // but experimentally detect that FFmpeg 0.5.1 is old and FFmpeg 0.6 include this
    // We check libavcodec version here and hopes this will true for most packages
    bool isUpToDate = ( aVersion.myMajor  > 52) ||
                      ((aVersion.myMajor == 52) && (aVersion.myMinor >= 72));
    if(!isUpToDate) {
        ST_DEBUG_LOG("Used old FFmpeg, enabled sView channel reorder for multichannel AC3 and OGG Vorbis streams!");
    }
    return !isUpToDate;
}

inline bool isReoderingNeeded() {
    static bool isNeeded = isReoderingNeededInit();
    return isNeeded;
}

void StAudioQueue::stalConfigureSources1() {
    alSourcefv(alSources[0], AL_POSITION,  POSITION_CENTER);
    stalCheckErrors("alSource*0");
}

void StAudioQueue::stalConfigureSources4_0() {
    alSourcefv(alSources[0], AL_POSITION, POSITION_FRONT_LEFT);
    alSourcefv(alSources[1], AL_POSITION, POSITION_FRONT_RIGHT);
    alSourcefv(alSources[2], AL_POSITION, POSITION_REAR_LEFT);
    alSourcefv(alSources[3], AL_POSITION, POSITION_REAR_RIGHT);
    stalCheckErrors("alSource*0123");
}

void StAudioQueue::stalConfigureSources5_1() {
    alSourcefv(alSources[0], AL_POSITION, POSITION_FRONT_LEFT);
    alSourcefv(alSources[1], AL_POSITION, POSITION_FRONT_RIGHT);
    alSourcefv(alSources[2], AL_POSITION, POSITION_FRONT_CENTER);
    alSourcefv(alSources[3], AL_POSITION, POSITION_LFE);
    alSourcefv(alSources[4], AL_POSITION, POSITION_REAR_LEFT);
    alSourcefv(alSources[5], AL_POSITION, POSITION_REAR_RIGHT);
    stalCheckErrors("alSource*012345");
}

bool StAudioQueue::stalInit() {
    if(!stalContext.create(alDeviceName)) {
        if(!stalContext.create()) {
            // retry with default device
            return false;
        }
    }
    stalContext.makeCurrent();

    alGetError(); // clear error code

    // generate the buffers
    for(size_t srcId = 0; srcId < NUM_AL_SOURCES; ++srcId) {
        alGenBuffers(NUM_AL_BUFFERS, &alBuffers[srcId][0]);
        stalCheckErrors(StString("alGenBuffers") + srcId);
    }

    // generate the sources
    alGenSources(NUM_AL_SOURCES, alSources);
    stalCheckErrors("alGenSources");

    // configure sources
    StGLVec3 stZeroVec(0.0f);
    for(size_t srcId = 0; srcId < NUM_AL_SOURCES; ++srcId) {
        alSourcefv(alSources[srcId], AL_POSITION,  stZeroVec);
        alSourcefv(alSources[srcId], AL_VELOCITY,  stZeroVec);
        alSourcefv(alSources[srcId], AL_DIRECTION, stZeroVec);
        alSourcef (alSources[srcId], AL_ROLLOFF_FACTOR,  0.0f);
        alSourcei (alSources[srcId], AL_SOURCE_RELATIVE, AL_TRUE);
        stalCheckErrors(StString("alSource*") + srcId);
    }

    // configure listener
    ALfloat listenerOri[] = {0.0f, 0.0f,-1.0f,  // forward
                             0.0f, 1.0f, 0.0f}; // up
    alListenerfv(AL_POSITION,    stZeroVec);
    alListenerfv(AL_VELOCITY,    stZeroVec);
    alListenerfv(AL_ORIENTATION, listenerOri);
    return true;
}

void StAudioQueue::stalDeinit() {
    // clear buffers
    stalEmpty();
    alSourceStopv(NUM_AL_SOURCES, alSources);

    alDeleteSources(NUM_AL_SOURCES, alSources);
    stalCheckErrors("alDeleteSources");

    for(size_t srcId = 0; srcId < NUM_AL_SOURCES; ++srcId) {
        alDeleteBuffers(NUM_AL_BUFFERS, &alBuffers[srcId][0]);
        stalCheckErrors(StString("alDeleteBuffers") + srcId);
    }

    // close device
    stalContext.destroy();
}

void StAudioQueue::stalEmpty() {
ST_DEBUG_LOG("   ___ stalEmpty()");
    alSourceStopv(NUM_AL_SOURCES, alSources);

    ALint aBufQueued = 0;
    ALuint alBuffIdToUnqueue = 0;
    for(size_t aSrcId = 0; aSrcId < NUM_AL_SOURCES; ++aSrcId) {
        alGetSourcei(alSources[aSrcId], AL_BUFFERS_QUEUED, &aBufQueued);
        for(ALint aBufIter = 0; aBufIter < aBufQueued; ++aBufIter) {
            alSourceUnqueueBuffers(alSources[aSrcId], 1, &alBuffIdToUnqueue);
            stalCheckErrors(StString("alSourceUnqueueBuffers") + aSrcId);
        }
        alSourcei(alSources[aSrcId], AL_BUFFER, 0);
    }
    ///alSourceRewindv(NUM_AL_SOURCES, alSources);
}

ALenum StAudioQueue::stalGetSourceState() {
    ALenum aState;
    alGetSourcei(alSources[0], AL_SOURCE_STATE, &aState);
    if(myDebugPrevSrcState != aState) {
        switch(aState) {
            case AL_INITIAL: ST_DEBUG_LOG("OpenAL source state: INITIAL"); break;
            case AL_PLAYING: ST_DEBUG_LOG("OpenAL source state: PLAYING"); break;
            case AL_PAUSED:  ST_DEBUG_LOG("OpenAL source state: PAUSED") ; break;
            case AL_STOPPED: ST_DEBUG_LOG("OpenAL source state: STOPPED"); break;
            default:         ST_DEBUG_LOG("OpenAL source state: UNKNOWN"); break;
        }
    }
    myDebugPrevSrcState = aState;
    return aState;
}

/**
 * Simple thread function which just call decodeLoop().
 */
static SV_THREAD_FUNCTION threadFunction(void* audioQueue) {
    StAudioQueue* stAudioQueue = (StAudioQueue* )audioQueue;
    stAudioQueue->decodeLoop();
    return SV_THREAD_RETURN 0;
}

StAudioQueue::StAudioQueue(const StString& alDeviceName)
: StAVPacketQueue(512),
  alDataLoop(),
  myPlaybackTimer(false),
  evDowntime(true),
  pcmBufferSrc(PCM16_SIGNED, AVCODEC_MAX_AUDIO_FRAME_SIZE),
  pcmBufferOut(PCM16_SIGNED, AVCODEC_MAX_AUDIO_FRAME_SIZE),
  alDeviceName(alDeviceName),
  stalContext(),
  alFormat(AL_FORMAT_STEREO16),
  myPrevFormat(AL_FORMAT_STEREO16),
  myPrevFrequency(0),
  toQuit(false),
  myDebugPrevQueued(-1),
  myDebugPrevSrcState(-1) {
    stMemSet(alSources, 0, sizeof(alSources));
    stalInit();

    // launch thread parse incoming packets from queue
    myThread = new StThread(threadFunction, (void* )this);
}

StAudioQueue::~StAudioQueue() {
    toQuit = true;
    pushQuit();

    myThread->wait();
    myThread.nullify();

    stalDeinit();

    deinit();
}

bool StAudioQueue::init(AVFormatContext*   theFormatCtx,
                        const unsigned int theStreamId) {
    if(!StAVPacketQueue::init(theFormatCtx, theStreamId)
    || myCodecCtx->codec_type != AVMEDIA_TYPE_AUDIO) {
        signals.onError("FFmpeg: invalid stream");
        deinit();
        return false;
    }

    // get AUDIO codec
    myCodec = avcodec_find_decoder(myCodecCtx->codec_id);
    if(myCodec == NULL) {
        signals.onError("FFmpeg: audio codec not found");
        deinit();
        return false;
    }
#if(LIBAVCODEC_VERSION_INT >= AV_VERSION_INT(53, 8, 0))
    if(avcodec_open2(myCodecCtx, myCodec, NULL) < 0) {
#else
    if(avcodec_open(myCodecCtx, myCodec) < 0) {
#endif
        signals.onError("FFmpeg: could not open audio codec");
        deinit();
        return false;
    }

    // setup source sample format (bitness)
    if(myCodecCtx->sample_fmt == stLibAV::audio::SAMPLE_FMT::U8) {
        pcmBufferSrc.setFormat(PCM8_UNSIGNED);
    } else if(myCodecCtx->sample_fmt == stLibAV::audio::SAMPLE_FMT::S16) {
        pcmBufferSrc.setFormat(PCM16_SIGNED);
    } else if(myCodecCtx->sample_fmt == stLibAV::audio::SAMPLE_FMT::S32) {
        pcmBufferSrc.setFormat(PCM32_SIGNED);
    } else if(myCodecCtx->sample_fmt == stLibAV::audio::SAMPLE_FMT::FLT) {
        pcmBufferSrc.setFormat(PCM32FLOAT);
    } else if(myCodecCtx->sample_fmt == stLibAV::audio::SAMPLE_FMT::DBL) {
        pcmBufferSrc.setFormat(PCM64FLOAT);
    } else {
        signals.onError(StString("Audio sample format '") + stLibAV::audio::getSampleFormatString(myCodecCtx)
                      + "' not supported");
        deinit();
        return false;
    }

    // setup frequency
    pcmBufferSrc.setFreq(myCodecCtx->sample_rate);
    pcmBufferOut.setFreq(myCodecCtx->sample_rate);

    // setup channel order
    if(myCodecCtx->channels == 1) {
        if(pcmBufferSrc.getFormat() == PCM8_UNSIGNED) {
            // uint8_t is supported by OpenAL core
            alFormat = AL_FORMAT_MONO8;
            pcmBufferOut.setFormat(PCM8_UNSIGNED);
        } else if(pcmBufferSrc.getFormat() == PCM64FLOAT && stalContext.hasExtFloat64) {
            // use float64 extension
            alFormat = alGetEnumValue("AL_FORMAT_MONO_DOUBLE_EXT");
            pcmBufferOut.setFormat(PCM64FLOAT);
        } else if((pcmBufferSrc.getFormat() == PCM32_SIGNED ||
                   pcmBufferSrc.getFormat() == PCM32FLOAT ||
                   pcmBufferSrc.getFormat() == PCM64FLOAT)
               && stalContext.hasExtFloat32) {
            // use float32 extension to reduce quality degradation
            alFormat = alGetEnumValue("AL_FORMAT_MONO_FLOAT32");
            pcmBufferOut.setFormat(PCM32FLOAT);
        } else {
            // default - int16_t
            alFormat = AL_FORMAT_MONO16;
            pcmBufferOut.setFormat(PCM16_SIGNED);
        }

        pcmBufferSrc.setupChannels(StChannelMap::CH10, StChannelMap::PCM, 1);
        pcmBufferOut.setupChannels(StChannelMap::CH10, StChannelMap::PCM, 1);
        stalConfigureSources1();
    } else if(myCodecCtx->channels == 2) {
        if(pcmBufferSrc.getFormat() == PCM8_UNSIGNED) {
            // uint8_t is supported by OpenAL core
            alFormat = AL_FORMAT_STEREO8;
            pcmBufferOut.setFormat(PCM8_UNSIGNED);
        } else if(pcmBufferSrc.getFormat() == PCM64FLOAT && stalContext.hasExtFloat64) {
            // use float64 extension
            alFormat = alGetEnumValue("AL_FORMAT_STEREO_DOUBLE_EXT");
            pcmBufferOut.setFormat(PCM64FLOAT);
        } else if((pcmBufferSrc.getFormat() == PCM32_SIGNED ||
                   pcmBufferSrc.getFormat() == PCM32FLOAT ||
                   pcmBufferSrc.getFormat() == PCM64FLOAT)
               && stalContext.hasExtFloat32) {
            // use float32 extension to reduce quality degradation
            alFormat = alGetEnumValue("AL_FORMAT_STEREO_FLOAT32");
            pcmBufferOut.setFormat(PCM32FLOAT);
        } else {
            // default - int16_t
            alFormat = AL_FORMAT_STEREO16;
            pcmBufferOut.setFormat(PCM16_SIGNED);
        }

        pcmBufferSrc.setupChannels(StChannelMap::CH20, StChannelMap::PCM, 1);
        pcmBufferOut.setupChannels(StChannelMap::CH20, StChannelMap::PCM, 1);
        stalConfigureSources1();
    } else if(myCodecCtx->channels == 4) {
        if(stalContext.hasExtMultiChannel) {
            if(pcmBufferSrc.getFormat() == PCM8_UNSIGNED) {
                alFormat = alGetEnumValue("AL_FORMAT_QUAD8");
                pcmBufferOut.setFormat(PCM8_UNSIGNED);
            } else if(pcmBufferSrc.getFormat() == PCM32_SIGNED ||
                      pcmBufferSrc.getFormat() == PCM32FLOAT ||
                      pcmBufferSrc.getFormat() == PCM64FLOAT) {
                // use float32 format to reduce quality degradation
                alFormat = alGetEnumValue("AL_FORMAT_QUAD32");
                pcmBufferOut.setFormat(PCM32FLOAT);
            } else {
                // default - int16_t
                alFormat = alGetEnumValue("AL_FORMAT_QUAD16");
                pcmBufferOut.setFormat(PCM16_SIGNED);
            }

            pcmBufferSrc.setupChannels(StChannelMap::CH40, StChannelMap::PCM, 1);
            pcmBufferOut.setupChannels(StChannelMap::CH40, StChannelMap::PCM, 1);
            stalConfigureSources1();
        } else {
            signals.onError("OpenAL: multichannel extension (AL_FORMAT_QUAD16) not available");
            deinit();
            return false;
        }
    } else if(myCodecCtx->channels == 6) {
        if(stalContext.hasExtMultiChannel) {
            if(pcmBufferSrc.getFormat() == PCM8_UNSIGNED) {
                alFormat = alGetEnumValue("AL_FORMAT_51CHN8");
                pcmBufferOut.setFormat(PCM8_UNSIGNED);
            } else if(pcmBufferSrc.getFormat() == PCM32_SIGNED ||
                      pcmBufferSrc.getFormat() == PCM32FLOAT ||
                      pcmBufferSrc.getFormat() == PCM64FLOAT) {
                // use float32 format to reduce quality degradation
                alFormat = alGetEnumValue("AL_FORMAT_51CHN32");
                pcmBufferOut.setFormat(PCM32FLOAT);
            } else {
                // default - int16_t
                alFormat = alGetEnumValue("AL_FORMAT_51CHN16");
                pcmBufferOut.setFormat(PCM16_SIGNED);
            }

            pcmBufferOut.setupChannels(StChannelMap::CH51, StChannelMap::PCM, 1);
            if(isReoderingNeeded()) {
                // workaround for old FFmpeg
                if(myCodec->id == CODEC_ID_AC3) {
                    pcmBufferSrc.setupChannels(StChannelMap::CH51, StChannelMap::AC3, 1);
                } else if(myCodec->id == CODEC_ID_VORBIS) {
                    pcmBufferSrc.setupChannels(StChannelMap::CH51, StChannelMap::OGG, 1);
                } else {
                    pcmBufferSrc.setupChannels(StChannelMap::CH51, StChannelMap::PCM, 1);
                }
            } else {
                pcmBufferSrc.setupChannels(StChannelMap::CH51, StChannelMap::PCM, 1);
            }
            stalConfigureSources1();
        } else {
            if(pcmBufferSrc.getFormat() == PCM8_UNSIGNED) {
                alFormat = AL_FORMAT_MONO8;
                pcmBufferOut.setFormat(PCM8_UNSIGNED);
            } else if(pcmBufferSrc.getFormat() == PCM64FLOAT && stalContext.hasExtFloat64) {
                // use float64 extension
                alFormat = alGetEnumValue("AL_FORMAT_MONO_DOUBLE_EXT");
                pcmBufferOut.setFormat(PCM64FLOAT);
            } else if((pcmBufferSrc.getFormat() == PCM32_SIGNED ||
                       pcmBufferSrc.getFormat() == PCM32FLOAT ||
                       pcmBufferSrc.getFormat() == PCM64FLOAT)
                   && stalContext.hasExtFloat32) {
                // use float32 extension to reduce quality degradation
                alFormat = alGetEnumValue("AL_FORMAT_MONO_FLOAT32");
                pcmBufferOut.setFormat(PCM32FLOAT);
            } else {
                // default - int16_t
                alFormat = AL_FORMAT_MONO16;
                pcmBufferOut.setFormat(PCM16_SIGNED);
            }

            pcmBufferOut.setupChannels(StChannelMap::CH51, StChannelMap::PCM, 6);
            if(isReoderingNeeded()) {
                // workaround for old FFmpeg
                if(myCodec->id == CODEC_ID_AC3) {
                    pcmBufferSrc.setupChannels(StChannelMap::CH51, StChannelMap::AC3, 1);
                } else if(myCodec->id == CODEC_ID_VORBIS) {
                    pcmBufferSrc.setupChannels(StChannelMap::CH51, StChannelMap::OGG, 1);
                } else {
                    pcmBufferSrc.setupChannels(StChannelMap::CH51, StChannelMap::PCM, 1);
                }
            } else {
                pcmBufferSrc.setupChannels(StChannelMap::CH51, StChannelMap::PCM, 1);
            }
            stalConfigureSources5_1();
            ST_DEBUG_LOG("OpenAL: multichannel extension (AL_FORMAT_51CHN16) not available");
        }
    }
    return true;
}

void StAudioQueue::deinit() {
    pcmBufferSrc.clear();
    pcmBufferOut.clear();
    StAVPacketQueue::deinit();
}

bool StAudioQueue::parseEvents() {
    double aPtsSeek = 0.0;
    switch(popPlayEvent(aPtsSeek)) {
        case ST_PLAYEVENT_PLAY: {
            stalEmpty();
            playTimerStart(0.0);
            playTimerPause();
            return false;
        }
        case ST_PLAYEVENT_STOP: {
            playTimerPause();
            stalEmpty();
            return false;
        }
        case ST_PLAYEVENT_PAUSE: {
            playTimerPause();
            alSourcePausev(NUM_AL_SOURCES, alSources);
            return false;
        }
        case ST_PLAYEVENT_RESUME: {
            playTimerResume();
            alSourcePlayv(NUM_AL_SOURCES, alSources);
            return false;
        }
        case ST_PLAYEVENT_SEEK: {
            // TODO (Kirill Gavrilov#5#) current seeking technics is ugly
            ST_DEBUG_LOG("StControlA::SEEK");
            stalEmpty();
            playTimerStart(aPtsSeek);
            playTimerPause();
            pcmBufferSrc.setDataSize(0);
            pcmBufferOut.setDataSize(0);
            // return special flag to skip "resume playback from" in loop
            return true;
        }
        case ST_PLAYEVENT_NONE:
        default:
            return false;
    }
}

bool StAudioQueue::stalQueue(const double thePts) {
    ALint aQueued = 0;
    ALint aProcessed = 0;
    ALenum aState = stalGetSourceState();
    alGetSourcei(alSources[0], AL_BUFFERS_PROCESSED, &aProcessed);
    alGetSourcei(alSources[0], AL_BUFFERS_QUEUED,    &aQueued);

#if defined(__ST_DEBUG__)
    if(myDebugPrevQueued != aQueued) {
        ST_DEBUG_LOG("OpenAL buffers: " + aQueued + " queued + "
            + aProcessed + " processed from " + NUM_AL_BUFFERS
        );
    }
    myDebugPrevQueued = aQueued;
#endif

    if(myPrevFormat    != alFormat
    ///|| myPrevFrequency != pcmBufferOut.getFreq()
    || (aState  == AL_STOPPED
     && aQueued == NUM_AL_BUFFERS)
    ) {
        ST_DEBUG_LOG("AL, reinitialize buffers per source , size= " + pcmBufferOut.getDataSize(0)
                            + "; freq= " + pcmBufferOut.getFreq());
        stalEmpty();
        stalCheckErrors("reset state");
        aProcessed = 0;
        aQueued = 0;
    }

    bool toTryToPlay = false;
    bool isQueued = false;
    if(aProcessed == 0 && aQueued < NUM_AL_BUFFERS) {
        stalCheckErrors("reset state");
        ///ST_DEBUG_LOG("AL, queue more buffers " + aQueued + " / " + NUM_AL_BUFFERS);
        myPrevFormat = alFormat;
        myPrevFrequency = pcmBufferOut.getFreq();
        for(size_t aSrcId = 0; aSrcId < pcmBufferOut.getSourcesCount(); ++aSrcId) {
            alBufferData(alBuffers[aSrcId][aQueued], alFormat,
                         pcmBufferOut.getData(aSrcId), (ALsizei )pcmBufferOut.getDataSize(aSrcId),
                         pcmBufferOut.getFreq());
            stalCheckErrors("alBufferData");
            alSourceQueueBuffers(alSources[aSrcId], 1, &alBuffers[aSrcId][aQueued]);
            stalCheckErrors("alSourceQueueBuffers");
        }
        toTryToPlay = ((aQueued + 1) == NUM_AL_BUFFERS);
        isQueued = true;
    } else if(aProcessed != 0
           && (aState == AL_PLAYING
            || aState == AL_PAUSED)) {
        ALuint alBuffIdToFill = 0;
        ///ST_DEBUG_LOG("queue buffer " + thePts + "; state= " + stalGetSourceState());
        if(pcmBufferOut.getDataSize(0) == 0) {
            ST_DEBUG_LOG(" EMPTY BUFFER ");
            return true;
        }

        myPrevFormat = alFormat;
        myPrevFrequency = pcmBufferOut.getFreq();
        for(size_t aSrcId = 0; aSrcId < pcmBufferOut.getSourcesCount(); ++aSrcId) {
            alSourceUnqueueBuffers(alSources[aSrcId], 1, &alBuffIdToFill);
            stalCheckErrors("alSourceUnqueueBuffers");
            if(alBuffIdToFill != 0) {
                alBufferData(alBuffIdToFill, alFormat,
                             pcmBufferOut.getData(aSrcId), (ALsizei )pcmBufferOut.getDataSize(aSrcId),
                             pcmBufferOut.getFreq());
                stalCheckErrors("alBufferData");
                alSourceQueueBuffers(alSources[aSrcId], 1, &alBuffIdToFill);
                stalCheckErrors("alSourceQueueBuffers");
            } else {
                ST_DEBUG_LOG("OpenAL, unqueue FAILED");
            }
        }
        toTryToPlay = true;
        isQueued = true;
    }

    if(aState == AL_STOPPED
    && toTryToPlay) {
        double diffSecs = double(alDataLoop.summ() + pcmBufferOut.getDataSizeWhole()) / double(pcmBufferOut.getSecondSize());
        if((thePts - diffSecs) < 100000.0) {
            playTimerStart(thePts - diffSecs);
        } else {
            playTimerStart(0.0);
        }
        alSourcePlayv(NUM_AL_SOURCES, alSources);
        ST_DEBUG_LOG("!!! OpenAL was in stopped state, now resume playback from " + (thePts - diffSecs));
    }

    return isQueued;
}

void StAudioQueue::stalFillBuffers(const double thePts,
                                   const bool   toIgnoreEvents) {
    if(!toIgnoreEvents) {
        parseEvents();
    }

    bool toSkipPlaybackFrom = false;
    while(!stalQueue(thePts)) {
        // AL queue is full
        if(!toIgnoreEvents) {
            toSkipPlaybackFrom = parseEvents();
        }
        if(toQuit) {
            return;
        }

        if(!toSkipPlaybackFrom && !stalIsAudioPlaying() && isPlaying()) {
            // this position means:
            // 1) buffers were empty and playback was stopped
            //    now we have all buffers full and could play them
            double diffSecs = double(alDataLoop.summ() + pcmBufferOut.getDataSizeWhole()) / double(pcmBufferOut.getSecondSize());
            if((thePts - diffSecs) < 100000.0) {
                playTimerStart(thePts - diffSecs);
            } else {
                playTimerStart(0.0);
            }
            alSourcePlayv(NUM_AL_SOURCES, alSources);
            ST_DEBUG_LOG("!!! OpenAL was in stopped state, now resume playback from " + (thePts - diffSecs));
        } else {
            // TODO (Kirill Gavrilov#3#) often updates may prevent normal video playback
            // on files with broken audio/video PTS
            ALfloat aPos = 0.0f;
            alGetSourcef(alSources[0], AL_SEC_OFFSET, &aPos);
            double diffSecs = double(alDataLoop.summ() + pcmBufferOut.getDataSizeWhole()) / double(pcmBufferOut.getSecondSize());
            diffSecs -= aPos;
            if((thePts - diffSecs) < 100000.0) {
                 static double oldPts = 0.0;
                 if(thePts != oldPts) {
                    /**ST_DEBUG_LOG("set AAApts " + (thePts - diffSecs)
                        + " from " + getPts()
                        + "(" + thePts + ", " + diffSecs + ")"
                    );*/
                    playTimerStart(thePts - diffSecs);
                    oldPts = thePts;
                }
                ///playTimerStart(thePts - diffSecs);
            }
        }
        StThread::sleep(1);
    }
}

void StAudioQueue::decodePacket(const StHandle<StAVPacket>& thePacket, double& thePts) {

    const uint8_t* audio_pkt_data = thePacket->getData();
    int audio_pkt_size = thePacket->getSize();
    int len1 = 0;
    int dataSize = 0;
    bool checkMoreFrames = false;
    double aNewPts = 0.0;
    // packet could store multiple frames
    for(;;) {
        while(audio_pkt_size > 0) {
            dataSize = (int )pcmBufferSrc.getBufferSizeWhole();

        #if(LIBAVCODEC_VERSION_INT >= AV_VERSION_INT(52, 23, 0))
            StAVPacket anAvPkt;
            anAvPkt.getAVpkt()->data = (uint8_t* )audio_pkt_data;
            anAvPkt.getAVpkt()->size = audio_pkt_size;
            len1 = avcodec_decode_audio3(myCodecCtx, (int16_t* )pcmBufferSrc.getData(), &dataSize, anAvPkt.getAVpkt());
        #else
            len1 = avcodec_decode_audio2(myCodecCtx,
                                         (int16_t* )pcmBufferSrc.getData(), &dataSize,
                                         audio_pkt_data, audio_pkt_size);
        #endif
            if(len1 < 0) {
                // if error, we skip the frame
                audio_pkt_size = 0;
                break;
            }

            audio_pkt_data += len1;
            audio_pkt_size -= len1;
            if(dataSize <= 0) {
                continue;
            }

            checkMoreFrames = true;
            pcmBufferSrc.setDataSize(dataSize);

            if(pcmBufferOut.addData(pcmBufferSrc)) {
                // 'big buffer' still not full
                break;
            }

            if(thePacket->getPts() != stLibAV::NOPTS_VALUE) {
                aNewPts = unitsToSeconds(thePacket->getPts()) - myPtsStartBase;
                if(aNewPts <= thePts) {
                    ST_DEBUG_LOG("Got the AUDIO packet with pts in past; "
                        + "new PTS= "   + aNewPts
                        + "; old PTS= " + thePts
                    );
                }
                thePts = aNewPts;
            }

            // now fill OpenAL buffers
            stalFillBuffers(thePts, false);
            if(toQuit) {
                return;
            }

            // save the history for filled AL buffers sizes
            alDataLoop.push(pcmBufferOut.getDataSizeWhole());

            // clear 'big' buffer
            pcmBufferOut.setDataSize(0);
            pcmBufferOut.addData(pcmBufferSrc);
            break;
        }

        if(checkMoreFrames) {
            checkMoreFrames = false;
            continue;
        }
        break;
    }

}

void StAudioQueue::decodeLoop() {
    double aPts = 0.0;
    StHandle<StAVPacket> aPacket;
    for(;;) {
        // wait for upcoming packets
        if(isEmpty()) {
            evDowntime.set();
            StThread::sleep(10);
            ///ST_DEBUG_LOG_AT("AQ is empty");
            continue;
        }
        evDowntime.reset();

        aPacket = pop();
        if(aPacket.isNull()) {
            continue;
        }
        switch(aPacket->getType()) {
            case StAVPacket::FLUSH_PACKET: {
                // got the special FLUSH packet - flush FFmpeg codec buffers
                if(myCodecCtx != NULL && myCodec != NULL) {
                    avcodec_flush_buffers(myCodecCtx);
                }
                // at this moment we clear current data from our buffers too
                pcmBufferOut.setDataSize(0);
                pcmBufferSrc.setDataSize(0);
                stalEmpty();
                continue;
            }
            case StAVPacket::START_PACKET: {
                playTimerStart(myPtsStartStream - myPtsStartBase);
                aPts = 0.0;
                continue;
            }
            case StAVPacket::END_PACKET: {
                pushPlayEvent(ST_PLAYEVENT_NONE);
                // TODO (Kirill Gavrilov#3#) improve file-by-file playback
                if(pcmBufferOut.getDataSize(0) != 0) {
                    stalFillBuffers(aPts, true);
                }
                pcmBufferOut.setDataSize(0);
                pcmBufferSrc.setDataSize(0);
                if(toQuit) {
                    return;
                }
                continue;
            }
            case StAVPacket::QUIT_PACKET: {
                return;
            }
        }

        // we got the data packet, so decode it
        decodePacket(aPacket, aPts);
        aPacket.nullify();
    }
}

void StAudioQueue::pushPlayEvent(const StPlayEvent_t theEventId,
                                 const double        theSeekParam) {
    myEventMutex.lock();
    StAVPacketQueue::pushPlayEvent(theEventId, theSeekParam);
    if(theEventId == ST_PLAYEVENT_SEEK) {
        myPlaybackTimer.restart(theSeekParam * 1000000.0);
    }
    myEventMutex.unlock();
}
