/**
 * Copyright © 2009-2011 Kirill Gavrilov <kirill@sview.ru>
 *
 * StMoviePlayer program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * StMoviePlayer program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.
 * If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __StAudioQueue_h_
#define __StAudioQueue_h_

#include <StStrings/StString.h>
#include <StThreads/StThreads.h>

#include "StAVPacketQueue.h"// StAVPacketQueue class
#include "StPCMBuffer.h"    // audio PCM buffer class
#include "StALContext.h"

// forward declarations
class StAudioQueue;

// define StHandle template specialization
ST_DEFINE_HANDLE(StAudioQueue, StAVPacketQueue);

/**
 * This is Audio playback class (OpenAL is used)
 * which feeded with packets (StAVPacket),
 * so it also implements StAVPacketQueue.
 */
class ST_LOCAL StAudioQueue : public StAVPacketQueue {

        private:

    // This constant sets count of OpenAL buffers, used in loop
    // for gapless playback
    #define NUM_AL_BUFFERS 4
    #define NUM_AL_SOURCES 8

    /**
     * This is help class, used to store buffer-sizes
     * that was decoded last.
     */
    class DataLoop {

            private:

        size_t dataSizes[NUM_AL_BUFFERS];
        size_t last;

            public:

        DataLoop() : last(NUM_AL_BUFFERS - 1) {
            stMemSet(dataSizes, 0, NUM_AL_BUFFERS * sizeof(size_t));
        }

        ~DataLoop() {
            //
        }

        void clear() {
            stMemSet(dataSizes, 0, NUM_AL_BUFFERS * sizeof(size_t));
        }

        void push(const size_t& dataSize) {
            ++last;
            if(last >= NUM_AL_BUFFERS) {
               last = 0;
            }
            dataSizes[last] = dataSize;
        }

        size_t summ() const {
            size_t sm = 0;
            for(size_t i = 0; i < NUM_AL_BUFFERS; ++i) {
                sm += dataSizes[i];
            }
            return sm;
        }

    } alDataLoop;

    StHandle<StThread>     myThread; //!< decoding loop thread
    mutable StTimer myPlaybackTimer; //!< timer used for current PTS calculation
    StEvent evDowntime;

    StPCMBuffer pcmBufferSrc; // decoded PCM audio buffer
    StPCMBuffer pcmBufferOut; // output PCM audio buffer

    // OpenAL items
    StString alDeviceName;
    StALContext stalContext;
    ALuint alBuffers[NUM_AL_SOURCES][NUM_AL_BUFFERS];
    ALuint alSources[NUM_AL_SOURCES]; // audio source
    ALenum alFormat; // internal format

    ALenum myPrevFormat;
    ALsizei myPrevFrequency;

    volatile bool  toQuit;
    size_t myInitBuffer;
    ALint  myDebugPrevQueued;   // debug variables
    ALenum myDebugPrevSrcState;

        private:

    bool stalInit();
    void stalDeinit();

    void stalConfigureSources1();
    void stalConfigureSources4_0();
    void stalConfigureSources5_1();

    bool stalQueue(const double thePts);

    /**
     * This function do fill OpenAL buffers.
     * @param pts (const double& ) - PTS for last decoded frame.
     */
    void stalFillBuffers(const double thePts,
                         const bool   toIgnoreEvents);

    void stalEmpty();

    ALenum stalGetSourceState();

    bool parseEvents();

    void decodePacket(const StHandle<StAVPacket>& thePacket,
                      double& thePts);

        private:

    void playTimerStart(const double thePts) {
        myEventMutex.lock();
            // timer operate within microseconds
            myPlaybackTimer.restart(thePts * 1000000.0);
        myEventMutex.unlock();
    }

    void playTimerPause() {
        myEventMutex.lock();
            myPlaybackTimer.pause();
        myEventMutex.unlock();
    }

    void playTimerResume() {
        myEventMutex.lock();
            myPlaybackTimer.resume();
        myEventMutex.unlock();
    }

        public:

    StAudioQueue(const StString& alDeviceName);
    virtual ~StAudioQueue();

    bool isInDowntime() {
        return evDowntime.check();
    }

    /**
     * Initialization function.
     * @param theFormatCtx (AVFormatContext* )  - pointer to video format context;
     * @param theStreamId (const unsigned int ) - stream id in video format context;
     * @return true if no error.
     */
    virtual bool init(AVFormatContext*   theFormatCtx,
                      const unsigned int theStreamId);

    /**
     * Clean function.
     */
    virtual void deinit();

    /**
     * Main decoding and playback loop.
     * Give packets from queue, decode them and fill OpenAL buffers
     * for playback.
     */
    void decodeLoop();

    /**
     * @return true if audio is played.
     */
    bool stalIsAudioPlaying() {
        return (stalGetSourceState() == AL_PLAYING);
    }

        public: //!< playback control methods

    virtual void pushPlayEvent(const StPlayEvent_t theEventId,
                               const double        theSeekParam = 0.0);

    double getPts() const {
        myEventMutex.lock();
            if(!isPlaying()) {
                myPlaybackTimer.pause();
            }
            double aPts = myPlaybackTimer.getElapsedTimeInSec();
        myEventMutex.unlock();
        return aPts;
    }

};

#endif //__StAudioQueue_h_
