/**
 * Copyright © 2007-2010 Kirill Gavrilov <kirill@sview.ru>
 *
 * StOutAnaglyph library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * StOutAnaglyph library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library.
 * If not, see <http://www.gnu.org/licenses/>.
 */

#include "StOutAnaglyph.h"

#include <StGLEW.h>
#include <StGL/StGL.h>
#include <StGL/StGLEnums.h>
#include <StThreads/StThreads.h> // threads header (mutexes, threads,...)
#include <StCore/StWindow.h>
#include <StSettings/StSettings.h>
#include <StSettings/StTranslations.h>

namespace {
    static bool HAS_LOGGER_ID = StLogger::IdentifyModule("StOutAnaglyph");
};

// shaders data
static const StString VSHADER        = "vAnaglyph.shv";
static const StString FSHADER_SIMPLE = "fAnaglyphSimple.shf";
static const StString FSHADER_GRAY   = "fAnaglyphGray.shf";
static const StString FSHADER_TRUE   = "fAnaglyphTrue.shf";
static const StString FSHADER_OPTIM  = "fAnaglyphOptim.shf";
static const StString FSHADER_YELLOW = "fAnaglyphYellow.shf";
static const StString FSHADER_YELLOWD= "fAnaglyphYellowDubois.shf";
static const StString FSHADER_GREEN  = "fAnaglyphGreen.shf";

static const StString ST_OUT_PLUGIN_NAME = "StOutAnaglyph";

static const StString ST_SETTING_GLASSES        = "glasses";
static const StString ST_SETTING_REDCYAN        = "optionRedCyan";
static const StString ST_SETTING_AMBERBLUE      = "optionAmberBlue";
static const StString ST_SETTING_WINDOWPOS      = "windowPos";
static const StString ST_SETTING_VSYNC          = "vsync";
static const StString ST_SETTING_FRBUFF_X       = "frbufferX";
static const StString ST_SETTING_FRBUFF_Y       = "frbufferY";
static const StString ST_SETTING_USER_FRMB_SIZE = "useUserFrmbSize";

// translation resources
enum {
    // devices' names
    STTR_ANAGLYPH_NAME     = 1000,
    STTR_ANAGLYPH_DESC     = 1001,

    // glasses
    STTR_ANAGLYPH_GLASSES  = 1010,
    STTR_ANAGLYPH_REDCYAN  = 1011,
    STTR_ANAGLYPH_YELLOW   = 1012,
    STTR_ANAGLYPH_GREEN    = 1013,

    // parameters
    STTR_ANAGLYPH_VSYNC            = 1100,
    STTR_ANAGLYPH_SHOW_FPS         = 1101,
    STTR_ANAGLYPH_REDCYAN_MENU     = 1102,
    STTR_ANAGLYPH_REDCYAN_SIMPLE   = 1120,
    STTR_ANAGLYPH_REDCYAN_OPTIM    = 1121,
    STTR_ANAGLYPH_REDCYAN_GRAY     = 1122,
    STTR_ANAGLYPH_REDCYAN_DARK     = 1123,
    STTR_ANAGLYPH_AMBERBLUE_MENU   = 1103,
    STTR_ANAGLYPH_AMBERBLUE_SIMPLE = 1130,
    STTR_ANAGLYPH_AMBERBLUE_DUBIOS = 1131,

    // about info
    STTR_PLUGIN_TITLE       = 2000,
    STTR_VERSION_STRING     = 2001,
    STTR_PLUGIN_DESCRIPTION = 2002,
};

void StOutAnaglyph::setShader(const int theGlasses,
                              const int theOptionRedCyan,
                              const int theOptionAmberBlue) {
    glasses         = theGlasses;
    optionRedCyan   = theOptionRedCyan;
    optionAmberBlue = theOptionAmberBlue;
    switch(theGlasses) {
        case GLASSES_TYPE_REDCYAN: {
            switch(optionRedCyan) {
                case REDCYAN_MODE_OPTIM: stereoProgram = &stOptimAnaglyph;  break;
                case REDCYAN_MODE_GRAY:  stereoProgram = &stGrayAnaglyph;   break;
                case REDCYAN_MODE_DARK:  stereoProgram = &stTrueAnaglyph;   break;
                case REDCYAN_MODE_SIMPLE:
                default: stereoProgram = &stSimpleAnaglyph; break;
            }
            break;
        }
        case GLASSES_TYPE_YELLOW: {
            switch(optionAmberBlue) {
                case AMBERBLUE_MODE_DUBOIS: stereoProgram = &stYellowDubiosAnaglyph;  break;
                case AMBERBLUE_MODE_SIMPLE:
                default: stereoProgram = &stYellowAnaglyph; break;
            }
            break;
        }
        case GLASSES_TYPE_GREEN: stereoProgram = &stGreenAnaglyph; break;
    }
    if(optionsStruct != NULL) {
        StSDSwitch_t* optionGlasses  = ((StSDSwitch_t* )optionsStruct->options[DEVICE_OPTION_GLASSES]);
        optionGlasses->value = glasses;
        StSDSwitch_t* optionRCFilter = ((StSDSwitch_t* )optionsStruct->options[DEVICE_OPTION_REDCYAN]);
        optionRCFilter->value = optionRedCyan;
        StSDSwitch_t* optionABFilter = ((StSDSwitch_t* )optionsStruct->options[DEVICE_OPTION_YELLOW]);
        optionABFilter->value = optionAmberBlue;
    }
}

void StOutAnaglyph::optionsStructAlloc() {
    StTranslations stLangMap(ST_OUT_PLUGIN_NAME);

    // create device options structure
    optionsStruct = (StSDOptionsList_t* )StWindow::memAlloc(sizeof(StSDOptionsList_t)); stMemSet(optionsStruct, 0, sizeof(StSDOptionsList_t));
    optionsStruct->curRendererPath = StWindow::memAllocNCopy(pluginPath);
    optionsStruct->curDeviceId = 0;

    optionsStruct->optionsCount = 5;
    optionsStruct->options = (StSDOption_t** )StWindow::memAlloc(sizeof(StSDOption_t*) * optionsStruct->optionsCount);

    // VSync option
    optionsStruct->options[DEVICE_OPTION_VSYNC] = (StSDOption_t* )StWindow::memAlloc(sizeof(StSDOnOff_t));
    optionsStruct->options[DEVICE_OPTION_VSYNC]->optionType = ST_DEVICE_OPTION_ON_OFF;
    ((StSDOnOff_t* )optionsStruct->options[DEVICE_OPTION_VSYNC])->value = isVSync;
    optionsStruct->options[DEVICE_OPTION_VSYNC]->title = StWindow::memAllocNCopy(stLangMap.changeValueId(STTR_ANAGLYPH_VSYNC, "VSync"));

    // Show FPS option
    optionsStruct->options[DEVICE_OPTION_SHOWFPS] = (StSDOption_t* )StWindow::memAlloc(sizeof(StSDOnOff_t));
    optionsStruct->options[DEVICE_OPTION_SHOWFPS]->optionType = ST_DEVICE_OPTION_ON_OFF;
    ((StSDOnOff_t* )optionsStruct->options[DEVICE_OPTION_SHOWFPS])->value = isShowFPS;
    optionsStruct->options[DEVICE_OPTION_SHOWFPS]->title = StWindow::memAllocNCopy(stLangMap.changeValueId(STTR_ANAGLYPH_SHOW_FPS, "Show FPS"));

    // Glasses switch option
    optionsStruct->options[DEVICE_OPTION_GLASSES] = (StSDOption_t* )StWindow::memAlloc(sizeof(StSDSwitch_t));
    optionsStruct->options[DEVICE_OPTION_GLASSES]->title = StWindow::memAllocNCopy(stLangMap.changeValueId(STTR_ANAGLYPH_GLASSES, "Glasses type"));
    optionsStruct->options[DEVICE_OPTION_GLASSES]->optionType = ST_DEVICE_OPTION_SWITCH;
    StSDSwitch_t* switchOptionGlasses = ((StSDSwitch_t* )optionsStruct->options[DEVICE_OPTION_GLASSES]);
    switchOptionGlasses->value = glasses;
    switchOptionGlasses->valuesCount = 3;
    switchOptionGlasses->valuesTitles = (stUtf8_t** )StWindow::memAlloc(switchOptionGlasses->valuesCount * sizeof(stUtf8_t*));
    switchOptionGlasses->valuesTitles[GLASSES_TYPE_REDCYAN] = StWindow::memAllocNCopy(stLangMap.changeValueId(STTR_ANAGLYPH_REDCYAN, "Red-cyan"));
    switchOptionGlasses->valuesTitles[GLASSES_TYPE_YELLOW]  = StWindow::memAllocNCopy(stLangMap.changeValueId(STTR_ANAGLYPH_YELLOW,  "Yellow-Blue"));
    switchOptionGlasses->valuesTitles[GLASSES_TYPE_GREEN]   = StWindow::memAllocNCopy(stLangMap.changeValueId(STTR_ANAGLYPH_GREEN,   "Green-Magenta"));

    // Red-cyan filter switch option
    optionsStruct->options[DEVICE_OPTION_REDCYAN] = (StSDOption_t* )StWindow::memAlloc(sizeof(StSDSwitch_t));
    optionsStruct->options[DEVICE_OPTION_REDCYAN]->title = StWindow::memAllocNCopy(stLangMap.changeValueId(STTR_ANAGLYPH_REDCYAN_MENU, "Red-Cyan filter"));
    optionsStruct->options[DEVICE_OPTION_REDCYAN]->optionType = ST_DEVICE_OPTION_SWITCH;
    StSDSwitch_t* switchOptionRCFilter = ((StSDSwitch_t* )optionsStruct->options[DEVICE_OPTION_REDCYAN]);
    switchOptionRCFilter->value = optionRedCyan;
    switchOptionRCFilter->valuesCount = 4;
    switchOptionRCFilter->valuesTitles = (stUtf8_t** )StWindow::memAlloc(switchOptionRCFilter->valuesCount * sizeof(stUtf8_t*));
    switchOptionRCFilter->valuesTitles[REDCYAN_MODE_SIMPLE] = StWindow::memAllocNCopy(stLangMap.changeValueId(STTR_ANAGLYPH_REDCYAN_SIMPLE, "Simple"));
    switchOptionRCFilter->valuesTitles[REDCYAN_MODE_OPTIM]  = StWindow::memAllocNCopy(stLangMap.changeValueId(STTR_ANAGLYPH_REDCYAN_OPTIM,  "Optimized"));
    switchOptionRCFilter->valuesTitles[REDCYAN_MODE_GRAY]   = StWindow::memAllocNCopy(stLangMap.changeValueId(STTR_ANAGLYPH_REDCYAN_GRAY,   "Grayed"));
    switchOptionRCFilter->valuesTitles[REDCYAN_MODE_DARK]   = StWindow::memAllocNCopy(stLangMap.changeValueId(STTR_ANAGLYPH_REDCYAN_DARK,   "Dark"));

    // Amber-Blue filter switch option
    optionsStruct->options[DEVICE_OPTION_YELLOW] = (StSDOption_t* )StWindow::memAlloc(sizeof(StSDSwitch_t));
    optionsStruct->options[DEVICE_OPTION_YELLOW]->title = StWindow::memAllocNCopy(stLangMap.changeValueId(STTR_ANAGLYPH_AMBERBLUE_MENU, "Yellow filter"));
    optionsStruct->options[DEVICE_OPTION_YELLOW]->optionType = ST_DEVICE_OPTION_SWITCH;
    StSDSwitch_t* switchOptionABFilter = ((StSDSwitch_t* )optionsStruct->options[DEVICE_OPTION_YELLOW]);
    switchOptionABFilter->value = optionAmberBlue;
    switchOptionABFilter->valuesCount = 2;
    switchOptionABFilter->valuesTitles = (stUtf8_t** )StWindow::memAlloc(switchOptionRCFilter->valuesCount * sizeof(stUtf8_t*));
    switchOptionABFilter->valuesTitles[AMBERBLUE_MODE_SIMPLE] = StWindow::memAllocNCopy(stLangMap.changeValueId(STTR_ANAGLYPH_AMBERBLUE_SIMPLE, "Simple"));
    switchOptionABFilter->valuesTitles[AMBERBLUE_MODE_DUBOIS] = StWindow::memAllocNCopy(stLangMap.changeValueId(STTR_ANAGLYPH_AMBERBLUE_DUBIOS, "Dubios"));
}

StOutAnaglyph::StOutAnaglyph()
: stCore(NULL),
  stSettings(NULL),
  pluginPath(),
  frBuffer(NULL),
  stereoProgram(NULL),
  stSimpleAnaglyph("Anaglyph Simple"),
  stGrayAnaglyph("Anaglyph Gray"),
  stTrueAnaglyph("Anaglyph True"),
  stOptimAnaglyph("Anaglyph Optimized"),
  stYellowAnaglyph("Anaglyph Yellow"),
  stYellowDubiosAnaglyph("Anaglyph Yellow Dubios"),
  stGreenAnaglyph("Anaglyph Green"),
  glasses(GLASSES_TYPE_REDCYAN),
  optionRedCyan(REDCYAN_MODE_SIMPLE),
  optionAmberBlue(AMBERBLUE_MODE_SIMPLE),
  optionsStruct(NULL),
  stFPSControl(),
  toSavePlacement(true),
  isVSync(true),
  isUserFrmSize(false),
  isShowFPS(false) {
    //
    stereoProgram = &stSimpleAnaglyph;
}

StOutAnaglyph::~StOutAnaglyph() {
    stSimpleAnaglyph.release();
    stGrayAnaglyph.release();
    stTrueAnaglyph.release();
    stOptimAnaglyph.release();
    stYellowAnaglyph.release();
    stYellowDubiosAnaglyph.release();
    stGreenAnaglyph.release();
    if(stCore != NULL && stSettings != NULL) {
        if(frBuffer != NULL) {
            stSettings->saveInt32(ST_SETTING_FRBUFF_X, frBuffer->getSizeX());
            stSettings->saveInt32(ST_SETTING_FRBUFF_Y, frBuffer->getSizeY());
            delete frBuffer;
        }
        stMemFree(optionsStruct, StWindow::memFree);

        // read windowed placement
        getStWindow()->hide(ST_WIN_MASTER);
        if(toSavePlacement) {
            getStWindow()->setFullScreen(false);
            StRect<int32_t> savedRect = getStWindow()->getPlacement();
            stSettings->saveInt32Rect(ST_SETTING_WINDOWPOS, savedRect);
        }
        stSettings->saveBool(ST_SETTING_VSYNC, isVSync);
        stSettings->saveBool(ST_SETTING_USER_FRMB_SIZE, isUserFrmSize);
        stSettings->saveInt32(ST_SETTING_GLASSES,   glasses);
        stSettings->saveInt32(ST_SETTING_REDCYAN,   optionRedCyan);
        stSettings->saveInt32(ST_SETTING_AMBERBLUE, optionAmberBlue);
    }
    delete stSettings;
    delete stCore;
    StSettings::FREE();
    StCore::FREE();
}

bool StOutAnaglyph::init(const StString& inRendererPath, const int& , const StNativeWin_t* nativeParent) {
    toSavePlacement = nativeParent == NULL;
    pluginPath = inRendererPath;
    if(!StVersionInfo::checkTimeBomb("sView - Anaglyph Output plugin")) {
        return false;
    }
    ST_DEBUG_LOG_AT("INIT Anaglyph output plugin");
    // Firstly INIT core library!
    if(StCore::INIT() != STERROR_LIBNOERROR) {
        stError(ST_OUT_PLUGIN_NAME + " Plugin, Core library not available!");
        return false;
    }

    // INIT settings library
    if(StSettings::INIT() != STERROR_LIBNOERROR) {
        stError(ST_OUT_PLUGIN_NAME + " Plugin, Settings plugin not available!");
        return false;
    }
    stSettings = new StSettings(ST_OUT_PLUGIN_NAME);
    stCore = new StCore();

    // load window position
    StRect<int32_t> loadedRect(256, 768, 256, 1024);
    stSettings->loadInt32Rect(ST_SETTING_WINDOWPOS, loadedRect);
    StMonitor stMonitor = StCore::getMonitorFromPoint(loadedRect.center());
    if(!stMonitor.getVRect().isPointIn(loadedRect.center())) {
        ST_DEBUG_LOG("Warning, stored window position is out of the monitor(" + stMonitor.getId() + ")!" + loadedRect.toString());
        int w = loadedRect.width();
        int h = loadedRect.height();
        loadedRect.left()   = stMonitor.getVRect().left() + 256;
        loadedRect.right()  = loadedRect.left() + w;
        loadedRect.top()    = stMonitor.getVRect().top() + 256;
        loadedRect.bottom() = loadedRect.top() + h;
    }
    getStWindow()->setPlacement(loadedRect);

    stSettings->loadBool(ST_SETTING_VSYNC, isVSync);

    // load framebuffer settings
    GLsizei frbufferSizeX = 1680;
    GLsizei frbufferSizeY = 1050;
    stSettings->loadBool(ST_SETTING_USER_FRMB_SIZE, isUserFrmSize);
    if(isUserFrmSize) {
        stSettings->loadInt32(ST_SETTING_FRBUFF_X, frbufferSizeX);
        stSettings->loadInt32(ST_SETTING_FRBUFF_Y, frbufferSizeY);
    } else {
        frbufferSizeX = stMonitor.getVRect().width();
        frbufferSizeY = stMonitor.getVRect().height();
    }

    // load glasses settings
    stSettings->loadInt32(ST_SETTING_GLASSES,   glasses);
    stSettings->loadInt32(ST_SETTING_REDCYAN,   optionRedCyan);
    stSettings->loadInt32(ST_SETTING_AMBERBLUE, optionAmberBlue);
    setShader(glasses, optionRedCyan, optionAmberBlue);

    // allocate and setup the structure pointer
    optionsStructAlloc();
    getStWindow()->setValue(ST_WIN_DATAKEYS_RENDERER, (size_t )optionsStruct);

    // create our window!
    getStWindow()->setTitle("sView - Anaglyph Renderer plugin");
    StWinAttributes_t attribs = stDefaultWinAttributes();
    getStWindow()->stglCreate(&attribs, nativeParent);

    // INIT GLEW library (help library for OpenGL extensions)
    if(!StGLEW::init()) {
        stError(ST_OUT_PLUGIN_NAME + " Plugin, OpenGL context is broken!\n(OpenGL library internal error?)");
        return false;
    } else if(!GLEW_VERSION_2_0) {
        stError(ST_OUT_PLUGIN_NAME + " Plugin, OpenGL2.0+ not available!");
        return false;
    }
    if(!stglSetVSync(isVSync)) {
        // enable/disable VSync by config
        ST_DEBUG_LOG(ST_OUT_PLUGIN_NAME + " Plugin, VSync extension not available!");
    }

    // INIT framebuffers
    if(!StGLFrameBuffer::isRectangularFBSupported()) {
        if(!isUserFrmSize) {
            StGLFrameBuffer::convertToPowerOfTwo(frbufferSizeX, frbufferSizeY);
        }
        ST_DEBUG_LOG(ST_OUT_PLUGIN_NAME + " Plugin, old videocard detected (GLSL 1.1)!");
    }
    frBuffer = new StGLStereoFrameBuffer();
    if(!frBuffer->init(frbufferSizeX, frbufferSizeY)) {
        stError(ST_OUT_PLUGIN_NAME + " Plugin, Failed to init Frame Buffer");
        return false;
    }

    // INIT shaders
    StString aShadersError(ST_OUT_PLUGIN_NAME + " Plugin, Failed to init Shaders");
    const StString aShadersRoot = StProcess::getStCoreFolder() + "shaders" + SYS_FS_SPLITTER
                                   + ST_OUT_PLUGIN_NAME + SYS_FS_SPLITTER;
    StGLVertexShader stVertexShader("Anaglyph"); // common vertex shader
    if(!stVertexShader.initFile(aShadersRoot + VSHADER)) {
        stError(aShadersError);
        return false;
    }

    StGLFragmentShader stFSimple(stSimpleAnaglyph.getTitle());
    if(!stFSimple.initFile(aShadersRoot + FSHADER_SIMPLE)) {
        stError(aShadersError);
        return false;
    }
    stSimpleAnaglyph.create().attachShader(stVertexShader).attachShader(stFSimple).link();

    StGLFragmentShader stFGray(stGrayAnaglyph.getTitle());
    if(!stFGray.initFile(aShadersRoot + FSHADER_GRAY)) {
        stError(aShadersError);
        return false;
    }
    stGrayAnaglyph.create().attachShader(stVertexShader).attachShader(stFGray).link();

    StGLFragmentShader stFTrue(stTrueAnaglyph.getTitle());
    if(!stFTrue.initFile(aShadersRoot + FSHADER_TRUE)) {
        stError(aShadersError);
        return false;
    }
    stTrueAnaglyph.create().attachShader(stVertexShader).attachShader(stFTrue).link();

    StGLFragmentShader stFOptim(stOptimAnaglyph.getTitle());
    if(!stFOptim.initFile(aShadersRoot + FSHADER_OPTIM)) {
        stError(aShadersError);
        return false;
    }
    stOptimAnaglyph.create().attachShader(stVertexShader).attachShader(stFOptim).link();

    StGLFragmentShader stFYellow(stYellowAnaglyph.getTitle());
    if(!stFYellow.initFile(aShadersRoot + FSHADER_YELLOW)) {
        stError(aShadersError);
        return false;
    }
    stYellowAnaglyph.create().attachShader(stVertexShader).attachShader(stFYellow).link();

    StGLFragmentShader stFYellowDubios(stYellowDubiosAnaglyph.getTitle());
    if(!stFYellowDubios.initFile(aShadersRoot + FSHADER_YELLOWD)) {
        stError(aShadersError);
        return false;
    }
    stYellowDubiosAnaglyph.create().attachShader(stVertexShader).attachShader(stFYellowDubios).link();

    StGLFragmentShader stFGreen(stGreenAnaglyph.getTitle());
    if(!stFGreen.initFile(aShadersRoot + FSHADER_GREEN)) {
        stError(aShadersError);
        return false;
    }
    stGreenAnaglyph.create().attachShader(stVertexShader).attachShader(stFGreen).link();

    return    stSimpleAnaglyph.isValid()
           || stGrayAnaglyph.isValid()
           || stTrueAnaglyph.isValid()
           || stOptimAnaglyph.isValid()
           || stYellowAnaglyph.isValid()
           || stYellowDubiosAnaglyph.isValid()
           || stGreenAnaglyph.isValid();
}

void StOutAnaglyph::callback(StMessage_t* stMessages) {
    stCore->callback(stMessages);
    for(size_t i = 0; stMessages[i].uin != StMessageList::MSG_NULL; ++i) {
        switch(stMessages[i].uin) {
            case StMessageList::MSG_RESIZE: {
                getStWindow()->stglMakeCurrent(ST_WIN_MASTER);
                stglResize(getStWindow()->getPlacement());
                break;
            }
            case StMessageList::MSG_KEYS: {
                bool* keysMap = ((bool* )stMessages[i].data);
                if(keysMap[ST_VK_F1]) {
                    setShader(GLASSES_TYPE_REDCYAN, REDCYAN_MODE_SIMPLE, optionAmberBlue); keysMap[ST_VK_F1] = false;

                    // send 'update' message to StDrawer
                    StMessage_t msg; msg.uin = StMessageList::MSG_DEVICE_OPTION;
                    StSDSwitch_t* option = ((StSDSwitch_t* )optionsStruct->options[DEVICE_OPTION_REDCYAN]);
                    msg.data = (void* )option->valuesTitles[option->value];
                    getStWindow()->appendMessage(msg);
                } else if(keysMap[ST_VK_F2]) {
                    setShader(GLASSES_TYPE_REDCYAN, REDCYAN_MODE_OPTIM, optionAmberBlue);  keysMap[ST_VK_F2] = false;

                    // send 'update' message to StDrawer
                    StMessage_t msg; msg.uin = StMessageList::MSG_DEVICE_OPTION;
                    StSDSwitch_t* option = ((StSDSwitch_t* )optionsStruct->options[DEVICE_OPTION_REDCYAN]);
                    msg.data = (void* )option->valuesTitles[option->value];
                    getStWindow()->appendMessage(msg);
                } else if(keysMap[ST_VK_F3]) {
                    setShader(GLASSES_TYPE_REDCYAN, REDCYAN_MODE_GRAY, optionAmberBlue);   keysMap[ST_VK_F3] = false;

                    // send 'update' message to StDrawer
                    StMessage_t msg; msg.uin = StMessageList::MSG_DEVICE_OPTION;
                    StSDSwitch_t* option = ((StSDSwitch_t* )optionsStruct->options[DEVICE_OPTION_REDCYAN]);
                    msg.data = (void* )option->valuesTitles[option->value];
                    getStWindow()->appendMessage(msg);
                } else if(keysMap[ST_VK_F4]) {
                    setShader(GLASSES_TYPE_REDCYAN, REDCYAN_MODE_DARK, optionAmberBlue);   keysMap[ST_VK_F4] = false;

                    // send 'update' message to StDrawer
                    StMessage_t msg; msg.uin = StMessageList::MSG_DEVICE_OPTION;
                    StSDSwitch_t* option = ((StSDSwitch_t* )optionsStruct->options[DEVICE_OPTION_REDCYAN]);
                    msg.data = (void* )option->valuesTitles[option->value];
                    getStWindow()->appendMessage(msg);
                } else if(keysMap[ST_VK_F5]) {
                    setShader(GLASSES_TYPE_YELLOW, optionRedCyan, optionAmberBlue);        keysMap[ST_VK_F5] = false;

                    // send 'update' message to StDrawer
                    StMessage_t msg; msg.uin = StMessageList::MSG_DEVICE_OPTION;
                    StSDSwitch_t* option = ((StSDSwitch_t* )optionsStruct->options[DEVICE_OPTION_REDCYAN]);
                    msg.data = (void* )option->valuesTitles[option->value];
                    getStWindow()->appendMessage(msg);
                } else if(keysMap[ST_VK_F6]) {
                    setShader(GLASSES_TYPE_GREEN, optionRedCyan, optionAmberBlue);         keysMap[ST_VK_F6] = false;

                    // send 'update' message to StDrawer
                    StMessage_t msg; msg.uin = StMessageList::MSG_DEVICE_OPTION;
                    StSDSwitch_t* option = ((StSDSwitch_t* )optionsStruct->options[DEVICE_OPTION_REDCYAN]);
                    msg.data = (void* )option->valuesTitles[option->value];
                    getStWindow()->appendMessage(msg);
                }
                if(keysMap[ST_VK_F12]) {
                    isShowFPS = !isShowFPS;
                    keysMap[ST_VK_F12] = false;

                    // send 'update' message to StDrawer
                    StMessage_t msg; msg.uin = StMessageList::MSG_DEVICE_OPTION;
                    StSDSwitch_t* option = ((StSDSwitch_t* )optionsStruct->options[DEVICE_OPTION_SHOWFPS]);
                    option->value = isShowFPS; msg.data = (void* )option;
                    getStWindow()->appendMessage(msg);
                }
                break;
            }
            case StMessageList::MSG_DEVICE_OPTION: {
                bool newVSync = ((StSDOnOff_t* )optionsStruct->options[DEVICE_OPTION_VSYNC])->value;
                if(newVSync != isVSync) {
                    isVSync = newVSync;
                    getStWindow()->stglMakeCurrent(ST_WIN_MASTER);
                    stglSetVSync(isVSync);
                }

                isShowFPS = ((StSDOnOff_t* )optionsStruct->options[DEVICE_OPTION_SHOWFPS])->value;
                setShader ((int )((StSDSwitch_t* )optionsStruct->options[DEVICE_OPTION_GLASSES])->value,
                           (int )((StSDSwitch_t* )optionsStruct->options[DEVICE_OPTION_REDCYAN])->value,
                           (int )((StSDSwitch_t* )optionsStruct->options[DEVICE_OPTION_YELLOW ])->value);
                break;
            }
        }
    }
}

void StOutAnaglyph::stglDraw(unsigned int ) {
    stFPSControl.setTargetFPS(getStWindow()->stglGetTargetFps());
    if(isShowFPS && stFPSControl.isUpdated()) {
        getStWindow()->setTitle(StString("Anaglyph Rendering FPS= ") + stFPSControl.getAverage());
    }

    if(!getStWindow()->isStereoOutput()) {
        getStWindow()->stglMakeCurrent(ST_WIN_MASTER);
        stCore->stglDraw(ST_DRAW_LEFT);

        stFPSControl.sleepToTarget(); // decrease FPS to target by thread sleeps
        getStWindow()->stglSwap(ST_WIN_MASTER);
        ++stFPSControl;
        return;
    }
    getStWindow()->stglMakeCurrent(ST_WIN_MASTER);

    // update FBO rendering dimensions
    StRectI_t aWinRect = getStWindow()->getPlacement();
    frBuffer->setVPDimensions(aWinRect.width(), aWinRect.height());

    // draw into virtual frame buffers (textures)
    glPushAttrib(GL_VIEWPORT_BIT); // we save previous (real window) viewport state
        frBuffer->setupViewPort();    // we set TEXTURE sizes here
        frBuffer->bindBufferLeft();
            stCore->stglDraw(ST_DRAW_LEFT);
        frBuffer->bindBufferRight();
            stCore->stglDraw(ST_DRAW_RIGHT);
        frBuffer->unbindBufferRight();
    glPopAttrib();                 // return real WINDOW viewport

    // now draw to real screen buffer
    // clear the screen and the depth buffer
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glDisable(GL_DEPTH_TEST);
    glDisable(GL_BLEND);
    frBuffer->bindMultiTexture();
    frBuffer->drawQuad(stereoProgram);
    frBuffer->unbindMultiTexture();

    stFPSControl.sleepToTarget(); // decrease FPS to target by thread sleeps
    getStWindow()->stglSwap(ST_WIN_MASTER);
    ++stFPSControl;
}

// SDK version was used
ST_EXPORT void getSDKVersion(StVersion* ver) {
    *ver = StVersionInfo::getSDKVersion();
}

// plugin version
ST_EXPORT void getPluginVersion(StVersion* ver) {
    *ver = StVersionInfo::getSDKVersion();
}

// plugin description
static StString initAboutInfo() {
    StTranslations aLangMap(ST_OUT_PLUGIN_NAME);
    StString& aTitle = aLangMap.changeValueId(STTR_PLUGIN_TITLE,
        "sView - Anaglyph Output module");
    StString& aVerString = aLangMap.changeValueId(STTR_VERSION_STRING,
        "version");
    StString& aDescr = aLangMap.changeValueId(STTR_PLUGIN_DESCRIPTION,
        "(C) 2007-2012 Kirill Gavrilov <kirill@sview.ru>\nOfficial site: www.sview.ru\n\nThis library distributed under LGPL3.0");
    return aTitle + '\n' + aVerString + ": " + StVersionInfo::getSDKVersionString() + "\n \n" + aDescr;
}

ST_EXPORT const stUtf8_t* getAboutInfo() {
    static const StString ABOUT_PLUGIN(initAboutInfo());
    return ABOUT_PLUGIN.toCString();
}

static StStereoDeviceInfo_t* prepareDevicesInfo() {
    StTranslations aLangMap(ST_OUT_PLUGIN_NAME);
    static StStereoDeviceInfo ST_DEVICE_INFO("Anaglyph",
        aLangMap.changeValueId(STTR_ANAGLYPH_NAME, "Anaglyph glasses"),
        aLangMap.changeValueId(STTR_ANAGLYPH_DESC, "Simple glasses with color-filters. Commonly used"),
        ST_DEVICE_SUPPORT_LOW); // anaglyph could be run on every display...
    static StStereoDeviceInfo_t ST_DEVICE_INFO_STRUCT = ST_DEVICE_INFO.getStruct();
    return &ST_DEVICE_INFO_STRUCT;
}

ST_EXPORT const StStereoDeviceInfo_t* getSupportedDevicesInfo() {
    static const StStereoDeviceInfo_t* DEVICES_INFO = prepareDevicesInfo();
    return DEVICES_INFO;
}

ST_EXPORT StRendererInterface* StRenderer_new() {
    return new StOutAnaglyph(); }
ST_EXPORT void StRenderer_del(StRendererInterface* inst) {
    delete (StOutAnaglyph* )inst; }
ST_EXPORT StWindowInterface* StRenderer_getStWindow(StRendererInterface* inst) {
    // This is VERY important return libImpl pointer here!
    return ((StOutAnaglyph* )inst)->getStWindow()->getLibImpl(); }
ST_EXPORT stBool_t StRenderer_init(StRendererInterface* inst, const stUtf8_t* rendererPath, const int& deviceId, const StNativeWin_t* nativeParent) {
    return ((StOutAnaglyph* )inst)->init(StString(rendererPath), deviceId, nativeParent); }
ST_EXPORT stBool_t StRenderer_open(StRendererInterface* inst, const StOpenInfo_t* stOpenInfo) {
    return ((StOutAnaglyph* )inst)->open(StOpenInfo(stOpenInfo)); }
ST_EXPORT void StRenderer_callback(StRendererInterface* inst, StMessage_t* stMessages) {
    ((StOutAnaglyph* )inst)->callback(stMessages); }
ST_EXPORT void StRenderer_stglDraw(StRendererInterface* inst, unsigned int views) {
    ((StOutAnaglyph* )inst)->stglDraw(views); }
