/**
 * Copyright © 2007-2010 Kirill Gavrilov <kirill@sview.ru>
 *
 * StOutDual library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * StOutDual library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library.
 * If not, see <http://www.gnu.org/licenses/>.
 */

#include "StOutDual.h"

#include <StGLEW.h>
#include <StGL/StGL.h>
#include <StGL/StGLEnums.h>
#include <StGL/StGLProgram.h>
#include <StGL/StGLFrameBuffer.h>
#include <StCore/StWindow.h>
#include <StSettings/StSettings.h>
#include <StSettings/StTranslations.h>

namespace {
    static bool HAS_LOGGER_ID = StLogger::IdentifyModule("StOutDual");
};

static const StString ST_OUT_PLUGIN_NAME        = "StOutDual";

static const StString ST_SETTING_DEVICE_ID      = "deviceId";
static const StString ST_SETTING_WINDOWPOS      = "windowPos";
static const StString ST_SETTING_SLAVE_ID       = "slaveId";
static const StString ST_SETTING_VSYNC          = "vsync";
static const StString ST_SETTING_FRBUFF_X       = "frbufferX";
static const StString ST_SETTING_FRBUFF_Y       = "frbufferY";
static const StString ST_SETTING_USER_FRMB_SIZE = "useUserFrmbSize";

// translation resources
enum {
    STTR_DUAL_NAME   = 1000,
    STTR_DUAL_DESC   = 1001,
    STTR_MIRROR_NAME = 1002,
    STTR_MIRROR_DESC = 1003,

    // parameters
    STTR_PARAMETER_VSYNC    = 1100,
    STTR_PARAMETER_SHOW_FPS = 1101,

    // about info
    STTR_PLUGIN_TITLE       = 2000,
    STTR_VERSION_STRING     = 2001,
    STTR_PLUGIN_DESCRIPTION = 2002,
};


/**
 * Just dummy GLSL program.
 */
class ST_LOCAL StProgramMM : public StGLProgram {

        private:

    StGLVarLocation atrVVertexLoc;
    StGLVarLocation atrVTexCoordLoc;

        public:

    StProgramMM()
    : StGLProgram("StProgramMM"),
      atrVVertexLoc(),
      atrVTexCoordLoc() {
        //
    }

    virtual ~StProgramMM() {
        //
    }

    StGLVarLocation getVVertexLoc() const {
        return atrVVertexLoc;
    }

    StGLVarLocation getVTexCoordLoc() const {
        return atrVTexCoordLoc;
    }

    virtual bool init() {
        const char VERTEX_SHADER[] =
           "attribute vec4 vVertex; \
            attribute vec2 vTexCoord; \
            varying vec2 fTexCoord; \
            void main(void) { \
                fTexCoord = vTexCoord; \
                gl_Position = vVertex; \
            }";

        const char FRAGMENT_SHADER[] =
           "uniform sampler2D texR, texL; \
            varying vec2 fTexCoord; \
            void main(void) { \
                gl_FragColor = texture2D(texR, fTexCoord); \
            }";

        StGLVertexShader stVertexShader(StGLProgram::getTitle());
        stVertexShader.init(VERTEX_SHADER);

        StGLFragmentShader stFragmentShader(StGLProgram::getTitle());
        stFragmentShader.init(FRAGMENT_SHADER);
        if(!StGLProgram::create().attachShader(stVertexShader).attachShader(stFragmentShader).link()) {
            return false;
        }

        atrVVertexLoc = StGLProgram::getAttribLocation("vVertex");
        atrVTexCoordLoc = StGLProgram::getAttribLocation("vTexCoord");
        return atrVVertexLoc.isValid() && atrVTexCoordLoc.isValid();
    }

};

void StOutDual::replaceDualAttribute(const int& from, const int& to) {
    StWinAttributes_t attribs = stDefaultWinAttributes();
    getStWindow()->getAttributes(&attribs);
    StWinAttributes_t origAttribs = attribs;
    switch(from) {
        case DUALMODE_XMIRROW:
            attribs.isSlaveXMirrow = false;
            break;
        case DUALMODE_YMIRROW:
            attribs.isSlaveYMirrow = false;
    }
    switch(to) {
        case DUALMODE_XMIRROW:
            attribs.isSlaveXMirrow = true;
            break;
        case DUALMODE_YMIRROW:
            attribs.isSlaveYMirrow = true;
            break;
    }
    if(!areSame(&origAttribs, &attribs)) {
        getStWindow()->setAttributes(&attribs);
    }
    this->deviceId = to;
    if(optionsStruct != NULL) {
        optionsStruct->curDeviceId = this->deviceId;
    }
}

void StOutDual::optionsStructAlloc() {
    StTranslations stLangMap(ST_OUT_PLUGIN_NAME);

    // create device options structure
    optionsStruct = (StSDOptionsList_t* )StWindow::memAlloc(sizeof(StSDOptionsList_t)); stMemSet(optionsStruct, 0, sizeof(StSDOptionsList_t));
    optionsStruct->curRendererPath = StWindow::memAllocNCopy(pluginPath);
    optionsStruct->curDeviceId = deviceId;

    optionsStruct->optionsCount = 2;
    optionsStruct->options = (StSDOption_t** )StWindow::memAlloc(sizeof(StSDOption_t*) * optionsStruct->optionsCount);

    // VSync option
    optionsStruct->options[DEVICE_OPTION_VSYNC] = (StSDOption_t* )StWindow::memAlloc(sizeof(StSDOnOff_t));
    optionsStruct->options[DEVICE_OPTION_VSYNC]->optionType = ST_DEVICE_OPTION_ON_OFF;
    ((StSDOnOff_t* )optionsStruct->options[DEVICE_OPTION_VSYNC])->value = isVSync;
    optionsStruct->options[DEVICE_OPTION_VSYNC]->title = StWindow::memAllocNCopy(stLangMap.changeValueId(STTR_PARAMETER_VSYNC, "VSync"));

    // Show FPS option
    optionsStruct->options[DEVICE_OPTION_SHOWFPS] = (StSDOption_t* )StWindow::memAlloc(sizeof(StSDOnOff_t));
    optionsStruct->options[DEVICE_OPTION_SHOWFPS]->optionType = ST_DEVICE_OPTION_ON_OFF;
    ((StSDOnOff_t* )optionsStruct->options[DEVICE_OPTION_SHOWFPS])->value = isShowFPS;
    optionsStruct->options[DEVICE_OPTION_SHOWFPS]->title = StWindow::memAllocNCopy(stLangMap.changeValueId(STTR_PARAMETER_SHOW_FPS, "Show FPS"));
}

StOutDual::StOutDual()
: stCore(NULL),
  stSettings(NULL),
  pluginPath(),
  frBuffer(NULL),
  stProgram(new StProgramMM()),
  stFPSControl(),
  stVertFlatBuf(),
  stVertXMirBuf(),
  stVertYMirBuf(),
  stTexCoordBuf(),
  optionsStruct(NULL),
  deviceId(StRendererInfo::DEVICE_AUTO),
  slaveMovId(1),
  toSavePlacement(true),
  isVSync(true),
  isUserFrmSize(false),
  isShowFPS(false) {
    //
}

StOutDual::~StOutDual() {
    delete stProgram;
    if(stCore != NULL && stSettings != NULL) {
        if(frBuffer != NULL) {
            stSettings->saveInt32(ST_SETTING_FRBUFF_X, frBuffer->getSizeX());
            stSettings->saveInt32(ST_SETTING_FRBUFF_Y, frBuffer->getSizeY());
            delete frBuffer;
        }
        stMemFree(optionsStruct, StWindow::memFree);

        // read windowed placement
        getStWindow()->hide(ST_WIN_MASTER);
        getStWindow()->hide(ST_WIN_SLAVE);
        if(toSavePlacement) {
            getStWindow()->setFullScreen(false);
            stSettings->saveInt32Rect(ST_SETTING_WINDOWPOS, getStWindow()->getPlacement());
        }
        stSettings->saveInt32(ST_SETTING_SLAVE_ID, slaveMovId);
        stSettings->saveBool(ST_SETTING_VSYNC, isVSync);
        stSettings->saveBool(ST_SETTING_USER_FRMB_SIZE, isUserFrmSize);
        stSettings->saveInt32(ST_SETTING_DEVICE_ID, deviceId);
    }
    delete stSettings;
    delete stCore;
    StSettings::FREE();
    StCore::FREE();
}

bool StOutDual::init(const StString& inRendererPath, const int& inDeviceId, const StNativeWin_t* nativeParent) {
    toSavePlacement = nativeParent == NULL;
    deviceId = inDeviceId;
    pluginPath = inRendererPath;
    if(!StVersionInfo::checkTimeBomb("sView - Dual Output plugin")) {
        return false;
    }
    ST_DEBUG_LOG_AT("INIT Dual output plugin");
    // Firstly INIT core library!
    if(StCore::INIT() != STERROR_LIBNOERROR) {
        stError(ST_OUT_PLUGIN_NAME + " Plugin, Core library not available!");
        return false;
    }

    // INIT settings library
    if(StSettings::INIT() != STERROR_LIBNOERROR) {
        stError(ST_OUT_PLUGIN_NAME + " Plugin, Settings plugin not available!");
        return false;
    }
    stSettings = new StSettings(ST_OUT_PLUGIN_NAME);
    stCore = new StCore();

    // load window position
    StRect<int32_t> loadedRect(256, 768, 256, 1024);
    stSettings->loadInt32Rect(ST_SETTING_WINDOWPOS, loadedRect);
    StMonitor stMonitor = StCore::getMonitorFromPoint(loadedRect.center());
    if(!stMonitor.getVRect().isPointIn(loadedRect.center())) {
        ST_DEBUG_LOG("Warning, stored window position is out of the monitor(" + stMonitor.getId() + ")!" + loadedRect.toString());
        int w = loadedRect.width();
        int h = loadedRect.height();
        loadedRect.left() = stMonitor.getVRect().left() + 256;
        loadedRect.right() = loadedRect.left() + w;
        loadedRect.top() = stMonitor.getVRect().top() + 256;
        loadedRect.bottom() = loadedRect.top() + h;
    }
    getStWindow()->setPlacement(loadedRect);

    stSettings->loadBool(ST_SETTING_VSYNC, isVSync);

    // load framebuffer settings
    GLsizei frbufferSizeX = 1680;
    GLsizei frbufferSizeY = 1050;
    stSettings->loadBool(ST_SETTING_USER_FRMB_SIZE, isUserFrmSize);
    if(isUserFrmSize) {
        stSettings->loadInt32(ST_SETTING_FRBUFF_X, frbufferSizeX);
        stSettings->loadInt32(ST_SETTING_FRBUFF_Y, frbufferSizeY);
    } else {
        frbufferSizeX = stMonitor.getVRect().width();
        frbufferSizeY = stMonitor.getVRect().height();
    }

    // load device settings
    if(deviceId == StRendererInfo::DEVICE_AUTO) {
        stSettings->loadInt32(ST_SETTING_DEVICE_ID, deviceId);
        if(deviceId == StRendererInfo::DEVICE_AUTO) {
            deviceId = DUALMODE_SIMPLE;
        }
    }

    // allocate and setup the structure pointer
    optionsStructAlloc();
    getStWindow()->setValue(ST_WIN_DATAKEYS_RENDERER, (size_t )optionsStruct);

    stSettings->loadInt32(ST_SETTING_SLAVE_ID, slaveMovId);

    // create our window!
    getStWindow()->setTitle("sView - Dual Renderer plugin");
    StWinAttributes_t attribs = stDefaultWinAttributes();
    attribs.isSlave = true;
    attribs.slaveMonId = int8_t(slaveMovId);
    getStWindow()->stglCreate(&attribs, nativeParent);
    replaceDualAttribute(DUALMODE_SIMPLE, deviceId);

    // INIT GLEW library (help library for OpenGL extensions)
    if(!StGLEW::init()) {
        stError(ST_OUT_PLUGIN_NAME + " Plugin, OpenGL context is broken!\n(OpenGL library internal error?)");
        return false;
    } else if(!GLEW_VERSION_2_0) {
        stError(ST_OUT_PLUGIN_NAME + " Plugin, OpenGL2.0+ not available!");
        return false;
    }

    getStWindow()->stglMakeCurrent(ST_WIN_MASTER);
    if(!stglSetVSync(isVSync)) {
        // enable/disable VSync by config
        ST_DEBUG_LOG(ST_OUT_PLUGIN_NAME + " Plugin, VSync extension not available!");
    }
    // TODO (Kirill Gavrilov#5) vsync
    ///getStWindow()->stglMakeCurrent(ST_WIN_SLAVE);
    ///stglSetVSync(isVSync);

    // INIT framebuffers
    if(!StGLFrameBuffer::isRectangularFBSupported()) {
        if(!isUserFrmSize) {
            StGLFrameBuffer::convertToPowerOfTwo(frbufferSizeX, frbufferSizeY);
        }
        ST_DEBUG_LOG(ST_OUT_PLUGIN_NAME + " Plugin, old videocard detected (GLSL 1.1)!");
    }
    frBuffer = new StGLFrameBuffer();
    if(!frBuffer->init(frbufferSizeX, frbufferSizeY)) {
        stError(ST_OUT_PLUGIN_NAME + " Plugin, Failed to init Frame Buffer");
        return false;
    }
    if(!stProgram->init()) {
        stError(ST_OUT_PLUGIN_NAME + " Plugin, Failed to init Shader");
        return false;
    }
    // create vertices buffers to draw simple textured quad
    const GLfloat QUAD_VERTICES[4 * 4] = {
         1.0f, -1.0f, 0.0f, 1.0f, // top-right
         1.0f,  1.0f, 0.0f, 1.0f, // bottom-right
        -1.0f, -1.0f, 0.0f, 1.0f, // top-left
        -1.0f,  1.0f, 0.0f, 1.0f  // bottom-left
    };
    const GLfloat QUAD_VERTICES_XMIR[4 * 4] = {
        -1.0f, -1.0f, 0.0f, 1.0f, // top-right
        -1.0f,  1.0f, 0.0f, 1.0f, // bottom-right
         1.0f, -1.0f, 0.0f, 1.0f, // top-left
         1.0f,  1.0f, 0.0f, 1.0f  // bottom-left
    };
    const GLfloat QUAD_VERTICES_YMIR[4 * 4] = {
         1.0f,  1.0f, 0.0f, 1.0f, // top-right
         1.0f, -1.0f, 0.0f, 1.0f, // bottom-right
        -1.0f,  1.0f, 0.0f, 1.0f, // top-left
        -1.0f, -1.0f, 0.0f, 1.0f  // bottom-left
    };

    const GLfloat QUAD_TEXCOORD[2 * 4] = {
        1.0f, 0.0f,
        1.0f, 1.0f,
        0.0f, 0.0f,
        0.0f, 1.0f
    };

    stVertFlatBuf.init(4, 4, QUAD_VERTICES);
    stVertXMirBuf.init(4, 4, QUAD_VERTICES_XMIR);
    stVertYMirBuf.init(4, 4, QUAD_VERTICES_YMIR);
    stTexCoordBuf.init(2, 4, QUAD_TEXCOORD);

    return true;
}

void StOutDual::callback(StMessage_t* stMessages) {
    stCore->callback(stMessages);
    for(size_t i = 0; stMessages[i].uin != StMessageList::MSG_NULL; ++i) {
        switch(stMessages[i].uin) {
            case StMessageList::MSG_RESIZE: {
                StRectI_t rect = getStWindow()->getPlacement();
                getStWindow()->stglMakeCurrent(ST_WIN_MASTER);
                stglResize(rect);
                getStWindow()->stglMakeCurrent(ST_WIN_SLAVE);
                stglResize(rect);
                break;
            }
            case StMessageList::MSG_KEYS: {
                bool* keysMap = ((bool* )stMessages[i].data);
                if(keysMap[ST_VK_F1]) {
                    replaceDualAttribute(deviceId, DUALMODE_SIMPLE);
                    keysMap[ST_VK_F1] = false;
                } else if(keysMap[ST_VK_F2]) {
                    replaceDualAttribute(deviceId, DUALMODE_XMIRROW);
                    keysMap[ST_VK_F2] = false;
                } else if(keysMap[ST_VK_F3]) {
                    replaceDualAttribute(deviceId, DUALMODE_YMIRROW);
                    keysMap[ST_VK_F3] = false;
                }
                if(keysMap[ST_VK_F12]) {
                    isShowFPS = !isShowFPS;
                    keysMap[ST_VK_F12] = false;

                    // send 'update' message to StDrawer
                    StMessage_t msg; msg.uin = StMessageList::MSG_DEVICE_OPTION;
                    StSDSwitch_t* option = ((StSDSwitch_t* )optionsStruct->options[DEVICE_OPTION_SHOWFPS]);
                    option->value = isShowFPS; msg.data = (void* )option;
                    getStWindow()->appendMessage(msg);
                }
                break;
            }
            case StMessageList::MSG_DEVICE_INFO: {
                if(optionsStruct->curDeviceId != deviceId) {
                    StString newPluginPath(optionsStruct->curRendererPath);
                    if(newPluginPath == pluginPath) {
                        replaceDualAttribute(deviceId, optionsStruct->curDeviceId);
                        stMessages[i].uin = StMessageList::MSG_NONE;
                    } // else - another plugin
                }
                break;
            }
            case StMessageList::MSG_DEVICE_OPTION: {
                bool newVSync = ((StSDOnOff_t* )optionsStruct->options[DEVICE_OPTION_VSYNC])->value;
                if(newVSync != isVSync) {
                    isVSync = newVSync;
                    getStWindow()->stglMakeCurrent(ST_WIN_MASTER);
                    stglSetVSync(isVSync);
                    // TODO (Kirill Gavrilov#5) vsync
                    ///getStWindow()->stglMakeCurrent(ST_WIN_SLAVE);
                    ///stglSetVSync(isVSync);
                }

                isShowFPS = ((StSDOnOff_t* )optionsStruct->options[DEVICE_OPTION_SHOWFPS])->value;
                break;
            }
        }
    }
}

void StOutDual::stglDraw(unsigned int ) {
    stFPSControl.setTargetFPS(getStWindow()->stglGetTargetFps());
    if(isShowFPS && stFPSControl.isUpdated()) {
        getStWindow()->setTitle(StString("Dual Rendering FPS= ") + stFPSControl.getAverage());
    }

    if(!getStWindow()->isStereoOutput()) {
        getStWindow()->stglMakeCurrent(ST_WIN_MASTER);
        stCore->stglDraw(ST_DRAW_LEFT);

        // TODO (Kirill Gavrilov#4#) we could do this once
        getStWindow()->stglMakeCurrent(ST_WIN_SLAVE);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        stFPSControl.sleepToTarget(); // decrease FPS to target by thread sleeps
        getStWindow()->stglSwap(ST_WIN_MASTER);
        getStWindow()->stglSwap(ST_WIN_SLAVE);
        ++stFPSControl;
        return;
    }
    getStWindow()->stglMakeCurrent(ST_WIN_MASTER);
    if(deviceId == DUALMODE_SIMPLE) {
        stCore->stglDraw(ST_DRAW_LEFT);

        getStWindow()->stglMakeCurrent(ST_WIN_SLAVE);
        stCore->stglDraw(ST_DRAW_RIGHT);
    } else {
        // update FBO rendering dimensions
        StRectI_t aWinRect = getStWindow()->getPlacement();
        GLsizei aVPSizeX = stMin(aWinRect.width(),  frBuffer->getSizeX());
        GLsizei aVPSizeY = stMin(aWinRect.height(), frBuffer->getSizeY());
        if(aVPSizeX != frBuffer->getVPSizeX()
        || aVPSizeY != frBuffer->getVPSizeY()) {
            // reduce viewport to avoid additional aliasing of narrow lines
            GLfloat aDX = GLfloat(aVPSizeX) / GLfloat(frBuffer->getSizeX());
            GLfloat aDY = GLfloat(aVPSizeY) / GLfloat(frBuffer->getSizeY());
            StArray<StGLVec2> aTCoords(4);
            aTCoords[0] = StGLVec2(aDX,  0.0f);
            aTCoords[1] = StGLVec2(aDX,  aDY);
            aTCoords[2] = StGLVec2(0.0f, 0.0f);
            aTCoords[3] = StGLVec2(0.0f, aDY);
            stTexCoordBuf.init(aTCoords);
            frBuffer->setVPSizeX(aVPSizeX);
            frBuffer->setVPSizeY(aVPSizeY);
        }

        // draw Left View into virtual frame buffer
        glPushAttrib(GL_VIEWPORT_BIT); // we save previous (real window) viewport state
            frBuffer->setupViewPort(); // we set TEXTURE sizes here
            frBuffer->bindBuffer();
                stCore->stglDraw(ST_DRAW_LEFT);
            frBuffer->unbindBuffer();
        glPopAttrib();                 // return real WINDOW viewport

        // now draw to real screen buffer
        // clear the screen and the depth buffer
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        frBuffer->bindTexture();
        stProgram->use();
            stVertFlatBuf.bindVertexAttrib(stProgram->getVVertexLoc());
            stTexCoordBuf.bindVertexAttrib(stProgram->getVTexCoordLoc());

            glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);

            stTexCoordBuf.unBindVertexAttrib(stProgram->getVTexCoordLoc());
            stVertFlatBuf.unBindVertexAttrib(stProgram->getVVertexLoc());
        stProgram->unuse();
        frBuffer->unbindTexture();

        // draw Right View into virtual frame buffer
        glPushAttrib(GL_VIEWPORT_BIT); // we save previous (real window) viewport state
            frBuffer->setupViewPort(); // we set TEXTURE sizes here
            frBuffer->bindBuffer();
                stCore->stglDraw(ST_DRAW_RIGHT);
            frBuffer->unbindBuffer();
        glPopAttrib();                 // return real WINDOW viewport

        getStWindow()->stglMakeCurrent(ST_WIN_SLAVE);
        // clear the screen and the depth buffer
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        frBuffer->bindTexture();
        stProgram->use();
            if(deviceId == DUALMODE_XMIRROW) {
                stVertXMirBuf.bindVertexAttrib(stProgram->getVVertexLoc());
            } else if(deviceId == DUALMODE_YMIRROW) {
                stVertYMirBuf.bindVertexAttrib(stProgram->getVVertexLoc());
            } else {
                stVertFlatBuf.bindVertexAttrib(stProgram->getVVertexLoc());
            }
            stTexCoordBuf.bindVertexAttrib(stProgram->getVTexCoordLoc());

            glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);

            stTexCoordBuf.unBindVertexAttrib(stProgram->getVTexCoordLoc());
            if(deviceId == DUALMODE_XMIRROW) {
                stVertXMirBuf.unBindVertexAttrib(stProgram->getVVertexLoc());
            } else if(deviceId == DUALMODE_YMIRROW) {
                stVertYMirBuf.unBindVertexAttrib(stProgram->getVVertexLoc());
            } else {
                stVertFlatBuf.unBindVertexAttrib(stProgram->getVVertexLoc());
            }
        stProgram->unuse();
        frBuffer->unbindTexture();
    }
    stFPSControl.sleepToTarget(); // decrease FPS to target by thread sleeps
    getStWindow()->stglSwap(ST_WIN_MASTER);
    getStWindow()->stglSwap(ST_WIN_SLAVE);
    ++stFPSControl;
    // make sure all GL changes in callback (in StDrawer) will fine
    getStWindow()->stglMakeCurrent(ST_WIN_MASTER);
}

// SDK version was used
ST_EXPORT void getSDKVersion(StVersion* ver) {
    *ver = StVersionInfo::getSDKVersion();
}

// plugin version
ST_EXPORT void getPluginVersion(StVersion* ver) {
    *ver = StVersionInfo::getSDKVersion();
}

// plugin description
static StString initAboutInfo() {
    StTranslations aLangMap(ST_OUT_PLUGIN_NAME);
    StString& aTitle = aLangMap.changeValueId(STTR_PLUGIN_TITLE,
        "sView - Dual Output module");
    StString& aVerString = aLangMap.changeValueId(STTR_VERSION_STRING,
        "version");
    StString& aDescr = aLangMap.changeValueId(STTR_PLUGIN_DESCRIPTION,
        "(C) 2007-2012 Kirill Gavrilov <kirill@sview.ru>\nOfficial site: www.sview.ru\n\nThis library distributed under LGPL3.0");
    return aTitle + '\n' + aVerString + ": " + StVersionInfo::getSDKVersionString() + "\n \n" + aDescr;
}

ST_EXPORT const stUtf8_t* getAboutInfo() {
    static const StString ABOUT_PLUGIN(initAboutInfo());
    return ABOUT_PLUGIN.toCString();
}

static StStereoDeviceInfo_t* prepareDevicesInfo() {
    int supportLevel = ST_DEVICE_SUPPORT_NONE;
    // Firstly INIT core library!
    if(StCore::INIT() != STERROR_LIBNOERROR) {
        ST_DEBUG_LOG(ST_OUT_PLUGIN_NAME + " Plugin, Core library not available!");
    } else {
        StArrayList<StMonitor> stMonitors = StCore::getStMonitors();
        if(stMonitors.size() >= 2) {
            const StMonitor& mon0 = stMonitors[0];
            const StMonitor& mon1 = stMonitors[1];
            if(   mon0.getVRect().width()  == mon1.getVRect().width()
               && mon0.getVRect().height() == mon1.getVRect().height()
               && mon0.getFreq() == mon1.getFreq()) {
                supportLevel = ST_DEVICE_SUPPORT_HIGHT;
            }
        }
        StCore::FREE();
    }

    //
    StTranslations aLangMap(ST_OUT_PLUGIN_NAME);
    static StStereoDeviceInfo ST_DEVICE_INFO_DUAL("StOutDual",
        aLangMap.changeValueId(STTR_DUAL_NAME, "Dual Output"),
        aLangMap.changeValueId(STTR_DUAL_DESC, "Stereo-device with dual input: some HMD, Mirrored Stereo monitors, Dual-Projectors"),
        supportLevel);
    static StStereoDeviceInfo ST_DEVICE_INFO_MIRROR("StOutMirrorX",
        aLangMap.changeValueId(STTR_MIRROR_NAME, "Mirror Output"),
        aLangMap.changeValueId(STTR_MIRROR_DESC, "Hand-make Mirrored Stereo monitors (mirror in X-direction)"),
        supportLevel);
    static StStereoDeviceInfo_t ST_DEVICE_INFO_DUAL_STRUCT   = ST_DEVICE_INFO_DUAL.getStruct();
    static StStereoDeviceInfo_t ST_DEVICE_INFO_MIRROR_STRUCT = ST_DEVICE_INFO_MIRROR.getStruct();
    ST_DEVICE_INFO_DUAL_STRUCT.nextInList = &ST_DEVICE_INFO_MIRROR_STRUCT;
    return &ST_DEVICE_INFO_DUAL_STRUCT;
}

ST_EXPORT const StStereoDeviceInfo_t* getSupportedDevicesInfo() {
    static const StStereoDeviceInfo_t* DEVICES_INFO = prepareDevicesInfo();
    return DEVICES_INFO;
}

ST_EXPORT StRendererInterface* StRenderer_new() {
    return new StOutDual(); }
ST_EXPORT void StRenderer_del(StRendererInterface* inst) {
    delete (StOutDual* )inst; }
ST_EXPORT StWindowInterface* StRenderer_getStWindow(StRendererInterface* inst) {
    // This is VERY important return libImpl pointer here!
    return ((StOutDual* )inst)->getStWindow()->getLibImpl(); }
ST_EXPORT stBool_t StRenderer_init(StRendererInterface* inst, const stUtf8_t* rendererPath, const int& deviceId, const StNativeWin_t* nativeParent) {
    return ((StOutDual* )inst)->init(StString(rendererPath), deviceId, nativeParent); }
ST_EXPORT stBool_t StRenderer_open(StRendererInterface* inst, const StOpenInfo_t* stOpenInfo) {
    return ((StOutDual* )inst)->open(StOpenInfo(stOpenInfo)); }
ST_EXPORT void StRenderer_callback(StRendererInterface* inst, StMessage_t* stMessages) {
    ((StOutDual* )inst)->callback(stMessages); }
ST_EXPORT void StRenderer_stglDraw(StRendererInterface* inst, unsigned int views) {
    ((StOutDual* )inst)->stglDraw(views); }
