/**
 * Copyright © 2009 Kirill Gavrilov <kirill@sview.ru>
 *
 * StOutPageFlip library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * StOutPageFlip library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library.
 * If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __StVuzixSDK_h_
#define __StVuzixSDK_h_

#include <StLibrary.h>  // header for dynamic library loading

namespace {
    static const StString LIB_TRACKER_NAME   = "IWEARDRV";
    static const StString LIB_STEREO_NAME    = "IWRSTDRV";
    static const StString PNP_ID_VUZIX_VR920 = "IWR0002";
};

class ST_LOCAL StVuzixSDK {

        private:

#if (!defined(_WIN32) && !defined(__WIN32__))
    typedef unsigned long DWORD;
    typedef long LONG;
    typedef void* HANDLE;
#endif

#if (defined(__APPLE__))
    typedef int BOOL;
    typedef unsigned char BYTE;
#endif

    typedef struct tag_IWRVERSION {
        unsigned short DLLMajor;
        unsigned short DLLMinor;
        unsigned short DLLSubMinor;
        unsigned short DLLBuildNumber;
        char USBFirmwareMajor;
        char USBFirmwareMinor;
        char TrackerFirmwareMajor;
        char TrackerFirmwareMinor;
        char VideoFirmware;
    } IWRVERSION, *PIWRVERSION;

    struct IWRsdkFunctions {
        // tracker API
        typedef DWORD (*IWROpenTracker_t)(void);
            IWROpenTracker_t IWROpenTracker;
        typedef void (*IWRCloseTracker_t)(void);
            IWRCloseTracker_t IWRCloseTracker;
        typedef void (*IWRZeroSet_t)(void);
            IWRZeroSet_t IWRZeroSet;
        typedef void (*IWRZeroClear_t)(void);
            IWRZeroClear_t IWRZeroClear;
        typedef DWORD (*IWRGetTracking_t)(LONG* , LONG* , LONG* );
            IWRGetTracking_t IWRGetTracking;
        typedef DWORD (*IWRGetVersion_t)(PIWRVERSION );
            IWRGetVersion_t IWRGetVersion;
        typedef void (*IWRSetFilterState_t)(BOOL );
            IWRSetFilterState_t IWRSetFilterState;
        typedef BOOL (*IWRGetFilterState_t)(void);
            IWRGetFilterState_t IWRGetFilterState;
        // stereo API
        typedef HANDLE (*IWRSTEREO_Open_t)(void);
            IWRSTEREO_Open_t IWRSTEREO_Open;
        typedef void (*IWRSTEREO_Close_t)(HANDLE hDev);
            IWRSTEREO_Close_t IWRSTEREO_Close;
        typedef BOOL (*IWRSTEREO_SetStereo_t)(HANDLE hDev, BOOL on);
            IWRSTEREO_SetStereo_t IWRSTEREO_SetStereo;
        typedef BYTE (*IWRSTEREO_WaitForAck_t)(HANDLE hDev, BOOL eye);
            IWRSTEREO_WaitForAck_t IWRSTEREO_WaitForAck;
        typedef BOOL (*IWRSTEREO_SetLR_t)(HANDLE hDev, BOOL eye);
            IWRSTEREO_SetLR_t IWRSTEREO_SetLR;
        typedef void (*IWRSTEREO_SetLREx_t)(HANDLE hDev, BOOL eye, BOOL wait);
            IWRSTEREO_SetLREx_t IWRSTEREO_SetLREx;
        typedef BOOL (*IWRSTEREO_GetVersion_t)(PIWRVERSION);
            IWRSTEREO_GetVersion_t IWRSTEREO_GetVersion;
    } functions;

    enum {
        IWR_MONO_MODE = 0,
        IWR_STEREO_MODE
    };

    enum {
        LEFT_EYE = 0,
        RIGHT_EYE,
        MONO_EYES
    };

    StLibrary stVTrackerLib;
    StLibrary stVStereoLib;

    HANDLE vrStereoHandle;

        public:

    StVuzixSDK() : functions(), stVTrackerLib(), stVStereoLib(), vrStereoHandle(NULL) {
        //
    }

    ~StVuzixSDK() {
        setMonoOut();
        close();
        if(stVTrackerLib.isOpened() && functions.IWRCloseTracker != NULL) {
            functions.IWRCloseTracker();
        }
        stVTrackerLib.close();
        stVStereoLib.close();
    }

    void open() {
        if(functions.IWRSTEREO_Open != NULL && vrStereoHandle == NULL) {
            vrStereoHandle = functions.IWRSTEREO_Open();
        }
    }

    void close() {
        if(functions.IWRSTEREO_Close != NULL && vrStereoHandle != NULL) {
            functions.IWRSTEREO_Close(vrStereoHandle);
            vrStereoHandle = NULL;
        }
    }

    void setStereoOut() {
        if(vrStereoHandle != NULL && functions.IWRSTEREO_SetStereo != NULL) {
            functions.IWRSTEREO_SetStereo(vrStereoHandle, IWR_STEREO_MODE);
        }
    }

    void setMonoOut() {
        if(vrStereoHandle != NULL && functions.IWRSTEREO_SetStereo != NULL) {
            functions.IWRSTEREO_SetStereo(vrStereoHandle, IWR_MONO_MODE);
        }
    }

    void waitAckLeft() {
        if(vrStereoHandle != NULL && functions.IWRSTEREO_WaitForAck != NULL) {
            functions.IWRSTEREO_WaitForAck(vrStereoHandle, LEFT_EYE);
        }
    }

    void waitAckRight() {
        if(vrStereoHandle != NULL && functions.IWRSTEREO_WaitForAck != NULL) {
            functions.IWRSTEREO_WaitForAck(vrStereoHandle, RIGHT_EYE);
        }
    }

    void setLeft() {
        if(vrStereoHandle != NULL && functions.IWRSTEREO_SetLR != NULL) {
            functions.IWRSTEREO_SetLR(vrStereoHandle, LEFT_EYE);
        }
    }

    void setRight() {
        if(vrStereoHandle != NULL && functions.IWRSTEREO_SetLR != NULL) {
            functions.IWRSTEREO_SetLR(vrStereoHandle, RIGHT_EYE);
        }
    }

    // INIT library function - called by host application to use library classes
    int init() {
        if(!stVTrackerLib.load(LIB_TRACKER_NAME)) {
            return STERROR_LIBLOADFAILED;
        }

        // get functions addresses in library
        stVTrackerLib("IWROpenTracker",    functions.IWROpenTracker);
        stVTrackerLib("IWRCloseTracker",   functions.IWRCloseTracker);
        stVTrackerLib("IWRZeroSet",        functions.IWRZeroSet);
        stVTrackerLib("IWRGetTracking",    functions.IWRGetTracking);
        stVTrackerLib("IWRSetFilterState", functions.IWRSetFilterState);
        if(    functions.IWROpenTracker == NULL || functions.IWRCloseTracker == NULL
            || functions.IWRZeroSet == NULL     || functions.IWRGetTracking == NULL
            || functions.IWRSetFilterState == NULL) {
            ///
            stVTrackerLib.close();
            return STERROR_LIBFUNCTIONNOTFOUND;
        }

        if(!stVStereoLib.load(LIB_STEREO_NAME)) {
            return STERROR_LIBLOADFAILED;
        }

        // get functions addresses in library
        stVStereoLib("IWRSTEREO_Open",       functions.IWRSTEREO_Open);
        stVStereoLib("IWRSTEREO_Close",      functions.IWRSTEREO_Close);
        stVStereoLib("IWRSTEREO_SetLR",      functions.IWRSTEREO_SetLR);
        stVStereoLib("IWRSTEREO_SetStereo",  functions.IWRSTEREO_SetStereo);
        stVStereoLib("IWRSTEREO_WaitForAck", functions.IWRSTEREO_WaitForAck);
        if(    functions.IWRSTEREO_Open == NULL || functions.IWRSTEREO_Close == NULL
            || functions.IWRSTEREO_SetLR == NULL || functions.IWRSTEREO_SetStereo == NULL
            || functions.IWRSTEREO_WaitForAck == NULL) {
            ///
            stVStereoLib.close();
            return STERROR_LIBFUNCTIONNOTFOUND;
        }
        functions.IWROpenTracker();
        // initialization success
        return STERROR_LIBNOERROR;
    }

    static bool isConnected() {
        // be sure Core library loaded before!
        StArrayList<StMonitor> stMonitors = StCore::getStMonitors();
        for(size_t m = 0; m < stMonitors.size(); ++m) {
            const StMonitor& mon = stMonitors[m];
            if(mon.getPnPId() == PNP_ID_VUZIX_VR920) {
                return true;
            }
        }
        return false;
    }

};

#endif //__StVuzixSDK_h_
