/**
 * Copyright © 2009-2010 Kirill Gavrilov <kirill@sview.ru>
 *
 * Distributed under the Boost Software License, Version 1.0.
 * See accompanying file license-boost.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt
 */

#ifndef __StGLProgram_h_
#define __StGLProgram_h_

#include <StGL/StGLShader.h>
#include <StGL/StGLVarLocation.h>

/**
 * Class represents GLSL program.
 */
class ST_LOCAL StGLProgram {

        public:

    static const GLuint NO_PROGRAM = 0;

    enum {
        TEXTURE_SAMPLE_0 = 0, // GL_TEXTURE0
        TEXTURE_SAMPLE_1 = 1, // GL_TEXTURE1
        TEXTURE_SAMPLE_2 = 2, // GL_TEXTURE2
    };

        protected:

    StString title; // just program title
    GLuint glProgramId; // OpenGL shader ID

    /**
     * Check linkage state.
     * @return true if program object linked successfully.
     */
    bool isLinked() const {
        GLint isSuccess = GL_FALSE;
        glGetProgramiv(glProgramId, GL_LINK_STATUS, &isSuccess);
        return isSuccess == GL_TRUE;
    }

    /**
     * Returns linkage information, provided by driver.
     * This string is driver's and vendor's specific.
     * Usually it contains linkage errors and warnings.
     */
    StString getLinkageInfo() const {
        GLint infologLength = 0;
        glGetProgramiv(glProgramId, GL_INFO_LOG_LENGTH, &infologLength);
        if(infologLength > 0) {
            GLchar* infoString = new GLchar[infologLength];
            GLsizei charsWritten = 0;
            glGetProgramInfoLog(glProgramId, infologLength, &charsWritten, infoString);
            StString compileInfo(infoString);
            delete[] infoString;
            return compileInfo;
        }
        return StString();
    }

        public:

    StGLProgram(const StString& title)
    : title(title),
      glProgramId(NO_PROGRAM) {
        //
    }

    /**
     * @return true if program object is valid.
     */
    bool isValid() const {
        return glProgramId != NO_PROGRAM;
    }

    /**
     * @return user-specified title for this program.
     */
    virtual const StString& getTitle() const {
        return title;
    }

    /**
     * Virtual method could be overridden by classes contained
     * program text internally.
     */
    virtual bool init() {
        return false;
    }

    virtual ~StGLProgram() {
        release();
    }

    /**
     * Delete program object and invalidate its id.
     * All attached shaders will be automatically detached
     * thus you should just delete shaders to clean up its memory.
     */
    void release() {
        if(isValid()) {
            glDeleteProgram(glProgramId);
            glProgramId = NO_PROGRAM;
        }
    }

    /**
     * Just create an empty program.
     */
    StGLProgram& create() {
        if(isValid()) {
            release();
        }
        glProgramId = glCreateProgram();
        return *this;
    }

    /**
     * Attach more compiled shader objects to this program.
     */
    StGLProgram& attachShader(const StGLShader& stShader) {
        if(isValid() && stShader.isValid()) {
            glAttachShader(glProgramId, stShader.glShaderId);
        }
        return *this;
    }

    /**
     * Detach shader objects from this program.
     */
    StGLProgram& detachShader(const StGLShader& stShader) {
        if(isValid() && stShader.isValid()) {
            glDetachShader(glProgramId, stShader.glShaderId);
        }
        return *this;
    }

    /**
     * Switch one shader object to another.
     */
    StGLProgram& swapShader(const StGLShader& stShaderFrom, const StGLShader& stShaderTo) {
        if(&stShaderFrom != &stShaderTo) {
            detachShader(stShaderFrom).attachShader(stShaderTo).link();
        }
        return *this;
    }

    /**
     * A vertex shader and fragment shader (and geometry shader) must be put together
     * to a program unit before it is possible to link.
     * Notice that default 0 value for any variable will be set after each (re)link!
     * This is good idea to perform searching for variables locations and setting your default
     * values here (using inheritance).
     * @return true if linkage success.
     */
    virtual bool link() {
        if(!isValid()) {
            return false;
        }
        glLinkProgram(glProgramId);

        StString aLinkageInfo = getLinkageInfo();
        ST_DEBUG_LOG("Linking Program Object '" + title + "' "
                     + (isLinked() ? StString("done") : StString("FAILED"))
                     + (!aLinkageInfo.isEmpty() ? (StString(". Log:\n") + aLinkageInfo) : (StString())));

        // if linkage failed - automatically remove the program!
        if(!isLinked()) {
            release();
            return false;
        }
        return true;
    }

    /**
     * @return uniform variable location in the whole shader (like projection/model matrix).
     */
    StGLVarLocation getUniformLocation(const char* varName) const {
        if(!isValid()) {
            return StGLVarLocation();
        }
        StGLVarLocation uniLocation(glGetUniformLocation(glProgramId, varName));
    #ifdef __ST_DEBUG__
        if(!uniLocation.isValid()) {
            ST_DEBUG_LOG("Warning! Uniform variable '" + varName + "' doesn't found in the whole shader!");
        }
    #endif
        return uniLocation;
    }

    /**
     * @return per-vertex attribute location (in the vertex shader, like vertex, normal, color and so on).
     */
    StGLVarLocation getAttribLocation(const char* varName) const {
        if(!isValid()) {
            return StGLVarLocation();
        }
        StGLVarLocation atrLoc(glGetAttribLocation(glProgramId, varName));
    #ifdef __ST_DEBUG__
        if(!atrLoc.isValid()) {
            ST_DEBUG_LOG("Warning! Attribute variable '" + varName + "' doesn't found in the vertex shader!");
        }
    #endif
        return atrLoc;
    }

    /**
     * Associates a generic vertex attribute index with a named attribute variable.
     * Should be called before linkage!
     */
    void bindAttribLocation(const char*     theVarName,
                            StGLVarLocation theLocation) {
        if(!isValid()) {
            return;
        }
        glBindAttribLocation(glProgramId, theLocation, theVarName);
    }

    /**
     * Use this program.
     */
    virtual void use() const {
        if(isValid()) {
            glUseProgram(glProgramId); // use our shader
        }
    }

    /**
     * Unuse this program.
     */
    virtual void unuse() const {
        unuseGlobal();
    }

    static void unuseGlobal() {
        glUseProgram(NO_PROGRAM); // use fixed instructions
    }

};

#endif //__StGLProgram_h_
