/**
 * Copyright © 2009-2011 Kirill Gavrilov <kirill@sview.ru>
 *
 * Distributed under the Boost Software License, Version 1.0.
 * See accompanying file license-boost.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt
 */

#ifndef __StGLShader_h_
#define __StGLShader_h_

#include <gl/glew.h>

#include <StStrings/StLogger.h>
#include <StFile/StRawFile.h>

/**
 * Virtual class represents GLSL shader.
 */
class ST_LOCAL StGLShader {

        public:

    static const GLuint NO_SHADER = 0;

        private:

    friend class StGLProgram;

        protected:

    StString title; // just shader title
    GLuint glShaderId; // OpenGL shader ID

    StGLShader(const StString& title)
    : title(title),
      glShaderId(NO_SHADER) {
        //
    }

    /**
     * Check compilation state.
     * @return true if shader object compiled successfully.
     */
    bool isCompiled() const {
        GLint isSuccess = GL_FALSE;
        glGetShaderiv(glShaderId, GL_COMPILE_STATUS, &isSuccess);
        return isSuccess == GL_TRUE;
    }

    /**
     * Returns compilation information, provided by driver.
     * This string is driver's and vendor's specific.
     * Usually it contains compilation errors and warnings.
     */
    StString getCompileInfo() const {
        GLint infologLength = 0;
        glGetShaderiv(glShaderId, GL_INFO_LOG_LENGTH, &infologLength);
        if(infologLength > 0) {
            GLchar* infoString = new GLchar[infologLength];
            GLsizei charsWritten = 0;
            glGetShaderInfoLog(glShaderId, infologLength, &charsWritten, infoString);
            StString compileInfo(infoString);
            delete[] infoString;
            return compileInfo;
        }
        return StString();
    }

        public:

    /**
     * Implementations should define this method to return:
     * GL_VERTEX_SHADER, GL_FRAGMENT_SHADER or GL_GEOMETRY_SHADER.
     */
    virtual GLenum getType() const = 0;

    virtual StString getTypeString() const {
        switch(getType()) {
            case GL_VERTEX_SHADER:   return StString("Vertex Shader");
            case GL_FRAGMENT_SHADER: return StString("Fragment Shader");
            ///case GL_GEOMETRY_SHADER: return StString("Geometry Shader");
            default: return StString("Unknown Shader");
        }
    }

    /**
     * @return true if shader object is valid.
     */
    bool isValid() const {
        return glShaderId != NO_SHADER;
    }

    /**
     * @return user-specified title for this shader.
     */
    virtual const StString& getTitle() const {
        return title;
    }

    /**
     * Vitual method could be overridden by classes contained
     * shader program text internally.
     */
    virtual bool init() {
        return false;
    }

    virtual ~StGLShader() {
        release();
    }

    /**
     * Delete shader object and invalidate its id.
     * You should detach this shader from all programs before deleting!
     */
    void release() {
        if(isValid()) {
            glDeleteShader(glShaderId);
            glShaderId = NO_SHADER;
        }
    }

    /**
     * Initialize the shader program from text buffer.
     * @param theSrcLines0 (const char* ) - shader program source;
     * @return true on success.
     */
    virtual bool init(const char* theSrcLines0,
                      const char* theSrcLines1 = NULL,
                      const char* theSrcLines2 = NULL) {
        if(theSrcLines0 == NULL) {
            return false;
        }
        GLsizei aCount = (theSrcLines1 == NULL) ? 1 : ((theSrcLines2 == NULL) ? 2 : 3);
        const char* aSrcArray[3] = {theSrcLines0, theSrcLines1, theSrcLines2};

        glShaderId = glCreateShader(getType());
        glShaderSource(glShaderId, aCount, aSrcArray, NULL);

        // compile shaders
        glCompileShader(glShaderId);

        StString compileInfo = getCompileInfo();
    #ifndef __ST_DEBUG_SHADERS__
        if(!isCompiled()) {
    #endif
        ST_DEBUG_LOG("Compiling " + getTypeString()
                     + " '" + title + "' "
                     + (isCompiled() ? StString("done") : StString("FAILED"))
                     + (!compileInfo.isEmpty() ? (StString(". Log:\n") + compileInfo) : (StString())));
    #ifndef __ST_DEBUG_SHADERS__
        }
    #endif

        // check compile success
        if(!isCompiled()) {
            release();
            return false;
        }
        return true;
    }

    /**virtual bool init(const StString& sourceUtfString) {
        std::string sourceString = sourceUtfString.ansiText();
        return init(sourceString.c_str());
    }*/

    /**
     * Initialize the shader from text file.
     * @param theFileName (const StString& ) - file to read;
     * @return true on success.
     */
    bool initFile(const StString& theFileName) {
        StRawFile aTextFile(theFileName);
        if(!aTextFile.readFile()) {
            ST_DEBUG_LOG("Shader file '" + theFileName + "' can not be read");
            return false;
        }
        return init((const char* )aTextFile.getBuffer());
    }

};

/**
 * Class represents GLSL Vertex Shader.
 */
class ST_LOCAL StGLVertexShader : public StGLShader {

        public:

    StGLVertexShader(const StString& title)
    : StGLShader(title) {
        //
    }

    virtual ~StGLVertexShader() {
        //
    }

    virtual GLenum getType() const {
        return GL_VERTEX_SHADER;
    }

};

/**
 * Class represents GLSL Fragment Shader.
 */
class ST_LOCAL StGLFragmentShader : public StGLShader {

        public:

    StGLFragmentShader(const StString& title)
    : StGLShader(title) {
        //
    }

    virtual ~StGLFragmentShader() {
        //
    }

    virtual GLenum getType() const {
        return GL_FRAGMENT_SHADER;
    }

};

#endif //__StGLShader_h_
