/**
 * This is source code for sView.
 *
 * Copyright © Kirill Gavrilov, 2010-2011
 */

///#ifndef __StGLVec_h_
///#define __StGLVec_h_

#include <StGL/StGLVec.h>

// idiotic inclusion template for FreeType
#include <ft2build.h>
#include FT_FREETYPE_H

/**
 * Class wrapper over FT_Library.
 * Provides access to FreeType library.
 */
class StFTLibrary {

        private:

    FT_Library myFTLib;

        public:

    /**
     * This method should always return true.
     * @return true if FT_Library instance is valid.
     */
    bool isValid() const {
        return myFTLib != NULL;
    }

    /**
     * Initialize new FT_Library instance.
     */
    StFTLibrary()
    : myFTLib(NULL) {
        if(FT_Init_FreeType(&myFTLib) != 0) {
            myFTLib = NULL;
        }
    }

    /**
     * Release FT_Library instance.
     */
    ~StFTLibrary() {
        if(isValid()) {
            FT_Done_FreeType(myFTLib);
        }
    }

    /**
     * Access FT_Library instance.
     */
    FT_Library getInstance() const {
        return myFTLib;
    }

};

class StGLFont : public StGLTextureAtlas {

        private:

    StHandle<StFTLibrary> myFTLib;    //!< handle to the FT library object
    FT_Face               myFTFace;   //!< FT face object
    FTCharmap*            myCharMap;  //!< character map object associated with the current face
    StString              myFontPath; //!< font path
    StImagePlane          myGlyphImg; //!< cached glyph plane

    FT_Vector             myKernAdvance; //!< buffer variable

    StDataMap<stUtf32_t, size_t> myGlyphMap;

        private:

    void release() {
        if(myFTFace != NULL) {
            FT_Done_Face(myFTFace);
            myFTFace = NULL;
        }
    }

    bool loadGlyph(const stUtf32_t theChar) {
        myGlyphImg.nullify();
        return theChar != 0
            && FT_Load_Char(myFTFace, theChar, FT_LOAD_TARGET_NORMAL) == 0
            && myFTFace->glyph != NULL;
    }

    bool loadAndRenderGlyph(const stUtf32_t theChar) {
        myGlyphImg.nullify();
        if(theChar == 0
        || FT_Load_Char(myFTFace, theChar, FT_LOAD_RENDER | FT_LOAD_TARGET_NORMAL) != 0
        || myFTFace->glyph == NULL
        || myFTFace->glyph->format != FT_GLYPH_FORMAT_BITMAP) {
            return false;
        }

        FT_Bitmap aBitmap = myFTFace->glyph->bitmap;
        if(aBitmap.pixel_mode != FT_PIXEL_MODE_GRAY
        || aBitmap.buffer == NULL || aBitmap.width <= 0 || aBitmap.rows <= 0) {
            return false;
        }
        if(!myGlyphImg.initWrapper(StImagePlane::ImgGray, aBitmap.buffer,
                                   aBitmap.width, aBitmap.rows, std::abs(aBitmap.pitch))) {
            return myGlyphImg;
        }
        myGlyphImg.setTopDown(aBitmap.pitch > 0);
        return true;
    }

    /**StGLVec2 getAdvance() const {
        return StGLVec2(myFTFace->glyph->advance.x / 64.0f,
                        myFTFace->glyph->advance.y / 64.0f);


        bool reserveMore(const size_t theTilesNb = 1)
        StGLTile& aNewTile = changeTile(const size_t theTileId);

    }*/

        public:

    bool isValid() const {
        return myFTFace != NULL;
    }

    ~StGLFont() {
        release();
    }

    size_t getTileId(const stUtf32_t theUCharCode) const {
        myGlyphMap.getValue(theUCharCode);
    }

    /**
     * @param thePointSize  - the face size in points (1/72 inch);
     * @param theResolution - the resolution of the target device;
     */
    StGLFont(const StString& theFontPath,
             const unsigned int thePointSize,
             const unsigned int theResolution = 72)
    : myFTFace(NULL),
      myFontPath(theFontPath) {
        StString aFontPath = StFileNode::getCompatibleName(myFontPath);
        if(myFTLib.isNull() || !myFTLib->isValid()
        || FT_New_Face(myFTLib->getInstance(), aFontPath.toCString(), 0, &myFTFace) != 0) {
            ST_DEBUG_LOG("Font '" + aFontPath + "' fail to load!");
            myFTFace = NULL;
            return;
        } else if(FT_Select_Charmap(myFTFace, ft_encoding_unicode) != 0) {
            ST_DEBUG_LOG("Font '" + aFontPath + "' doesn't contains Unicode charmap!");
            release();
            return;
        } else if(FT_Set_Char_Size(myFTFace, 0L, thePointSize * 64, theResolution, theResolution) != 0) {
            ST_DEBUG_LOG("Font '" + aFontPath + "' doesn't contains Unicode charmap!");
            release();
            return;
        }

    }

    bool renderGlyph(const size_t    theTileId
                     const stUtf32_t theChar) {
        if(!loadAndRenderGlyph(theChar)) {
            return false;
        }

        StHandle<StTexture>& aTexture = getTexture(theTileId);
        StGLTile& aTile = changeTile(theTileId);
        StVec2<GLsizei> aPos = getTilePx(theTileId);
        if(!aTexture->fill(myGlyphImg, const GLsizei theRowFrom = 0, const GLsizei theRowTo = 0)) {
            return false;
        }

        FT_Bitmap aBitmap = myFTFace->glyph->bitmap;

        glBindTexture(GL_TEXTURE_2D, myTextureID);
        glTexSubImage2D(GL_TEXTURE_2D, 0, aPos.x(), aPos.y(), aBitmap->width, aBitmap->rows, GL_ALPHA, GL_UNSIGNED_BYTE, aBitmap.buffer);

        aTile.px.left()   = GLfloat(myFTFace->glyph->bitmap_left);
        aTile.px.top()    = GLfloat(myFTFace->glyph->bitmap_top);
        aTile.px.right()  = aTile.px.left() + GLfloat(aBitmap->width);
        aTile.px.bottom() = aTile.px.top()  + GLfloat(aBitmap->rows);

        aTile.uv.left()   = (GLfloat(aPos.x()) / GLfloat(aTexture->getSizeX()));
        aTile.uv.top()    = (GLfloat(aPos.y()) / GLfloat(aTexture->getSizeY()));
        aTile.uv.right()  = (GLfloat(aPos.x() + aBitmap->width)  / GLfloat(aTexture->getSizeX()));
        aTile.uv.bottom() = (GLfloat(aPos.y() + aBitmap->rows)   / GLfloat(aTexture->getSizeY()));
    }

    GLfloat getAdvance(const stUtf32_t theUChar1,
                       const stUtf32_t theUChar2) {
        if(!loadGlyph(theUChar1)) {
            return 0.0f;
        }
        if(FT_HAS_KERNING(myFTFace) == 0 || theUChar2 == 0
        || FT_Get_Kerning(myFTFace, theUChar1, theUChar2, FT_KERNING_UNFITTED, &myKernAdvance) != 0) {
            return GLfloat(myFTFace->glyph->advance.x) / 64.0f;
        }
        return GLfloat(myKernAdvance.x + myFTFace->glyph->advance.x) / 64.0f;
    }

    void addAdvance(const stUtf32_t theUChar1,
                    const stUtf32_t theUChar2,
                    StGLVec2&       theVec) {
        if(!loadGlyph(theUChar1)) {
            return;
        }
        if(FT_HAS_KERNING(myFTFace) == 0 || theUChar2 == 0
        || FT_Get_Kerning(myFTFace, theUChar1, theUChar2, FT_KERNING_UNFITTED, &myKernAdvance) != 0) {
            theVec.x() += GLfloat(myFTFace->glyph->advance.x) / 64.0f;
            theVec.y() += GLfloat(myFTFace->glyph->advance.y) / 64.0f;
        } else {
            theVec.x() += GLfloat(myKernAdvance.x + myFTFace->glyph->advance.x) / 64.0f;
            theVec.y() += GLfloat(myKernAdvance.y + myFTFace->glyph->advance.y) / 64.0f;
        }
    }

};

class StGLText {

        private:

    StArrayList<StGLVertexBuffer> myTextVertBuf;
    StArrayList<StGLVertexBuffer> myTextTCrdBuf;
    StHandle<StGLFont>            myFont;
    StString                      myString;
    size_t                        myLength;

        public:

    StGLText(const StString& theString)
    : myString(theString) {
        //
    }

    void init() {
        StGLVec2 aPen(0.0f, 0.0f);

        StArrayList< StArrayList<StGLVec2> > aVertPerTexture;
        StArrayList< StArrayList<StGLVec2> > aTCrdPerTexture;

        for(StUtf8Iter anIter = myString.iterator(); *anIter != 0;) {
            stUtf32_t aCharThis = *anIter;
            stUtf32_t aCharNext = *++anIter;

            const StGLTile& aTile = myFont->getTile(const size_t theTileId);
            ///const StHandle<StGLTexture>& aTexture = getTexture(const size_t theTileId);
            const size_t aTextureNo = myFont->getTextureNo(const size_t theTileId);
            if((myFont == NULL) || (aTile.px.right() <= 0.0f) || (aTile.px.bottom() <= 0.0f)) {
                return myAdvance;
            }
            StGLRect aRect(aPen.y() + aTile.px.top(),  aPen.y() + aTile.px.bottom(),
                           aPen.x() + aTile.px.left(), aPen.x() + aTile.px.right());
            StArrayList<StGLVec2>& aVertices  = aVertPerTexture.changeValue(aTextureNo);
            aVertices.add(StGLVec2(aRect.left(),  aRect.top()));
            aVertices.add(StGLVec2(aRect.left(),  aRect.bottom()));
            aVertices.add(StGLVec2(aRect.right(), aRect.bottom()));
            aVertices.add(StGLVec2(aRect.right(), aRect.top()));

            StArrayList<StGLVec2>& aTexCoords = aTCrdPerTexture.changeValue(aTextureNo);
            aTexCoords.add(StGLVec2(aTile.uv.left(),  aTile.uv.top()));
            aTexCoords.add(StGLVec2(aTile.uv.left(),  aTile.uv.bottom()));
            aTexCoords.add(StGLVec2(aTile.uv.right(), aTile.uv.bottom()));
            aTexCoords.add(StGLVec2(aTile.uv.right(), aTile.uv.top()));

            aPen.x() += myFont->getAdvance(aCharThis, aCharNext);
        }
    }

    GLfloat getAdvance() {
        GLfloat anAdvance = 0.0f;
        for(StUtf8Iter anIter = myString.iterator(); *anIter != 0;) {
            stUtf32_t aCharThis = *anIter;
            stUtf32_t aCharNext = *++anIter;
            anAdvance += myFont->getAdvance(aCharThis, aCharNext);
        }

        /// TODO - allocate arrays
        return anAdvance;
    }

};

///#endif //__StGLVec_h_
