/**
 * Copyright © 2010-2011 Kirill Gavrilov <kirill@sview.ru>
 *
 * Distributed under the Boost Software License, Version 1.0.
 * See accompanying file license-boost.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt
 */

#ifndef __StGLVertexBuffer_h_
#define __StGLVertexBuffer_h_

#include <gl/glew.h>

#include <StStrings/StString.h>
#include <StGL/StGLVarLocation.h>
#include <StGL/StGLVec.h>

/**
 * Enumeration indicates the data stored in VBO.
 */
typedef enum tagStGLVBOType {
    ST_VBO_VERTEX, //!< vertices
    ST_VBO_NORMAL, //!< normals
    ST_VBO_TCOORD, //!< texture coordinates
    ST_VBO_COLORS, //!< colors
    ST_VBO_INDEX,  //!< indices
} StGLVBOType;

class ST_LOCAL StGLVertexBuffer {

        private:

    GLuint bufferId;
    GLsizeiptr elemSize;
    GLsizeiptr elemsCount;
    GLenum dataType;

        public:

    StGLVertexBuffer()
    : bufferId(0),
      elemSize(4),
      elemsCount(0),
      dataType(GL_FLOAT) {}

    virtual ~StGLVertexBuffer() {
        release();
    }

    bool operator==(const StGLVertexBuffer& compare) const {
        return bufferId == compare.bufferId &&
               getTarget() == compare.getTarget();
    }

    bool operator!=(const StGLVertexBuffer& compare) const {
        return bufferId != compare.bufferId ||
               getTarget() != compare.getTarget();
    }

    bool operator>(const StGLVertexBuffer& compare) const {
        return bufferId > compare.bufferId;
    }

    bool operator<(const StGLVertexBuffer& compare) const {
        return bufferId < compare.bufferId;
    }

    bool operator>=(const StGLVertexBuffer& compare) const {
        return bufferId >= compare.bufferId;
    }

    bool operator<=(const StGLVertexBuffer& compare) const {
        return bufferId <= compare.bufferId;
    }

    bool isValid() const {
        return bufferId != 0;
    }

    bool init() {
        if(!isValid()) {
            glGenBuffers(1, &bufferId);
        }
        return isValid();
    }

    void release() {
        if(isValid()) {
            glDeleteBuffers(1, &bufferId);
            bufferId = 0;
        }
    }

    void bind() const {
        glBindBuffer(getTarget(), bufferId);
    }

    void unbind() const {
        glBindBuffer(getTarget(), 0);
    }

    /**
     * VBO should be binded before call.
     * @param elemSize (GLsizeiptr ) - specifies the number of components per generic vertex attribute. Must be 1, 2, 3, or 4;
     * @param elemsCount (GLsizeiptr ) - elements count;
     * @param data (const GLfloat* ) - data pointer.
     */
    void setData(GLsizeiptr elemSize, GLsizeiptr elemsCount, const GLfloat* data) {
        this->elemSize = elemSize;
        this->elemsCount = elemsCount;
        glBufferData(getTarget(), elemsCount * elemSize * sizeof(GLfloat), data, GL_STATIC_DRAW);
        dataType = GL_FLOAT;
    }

    void setData(GLsizeiptr elemSize, GLsizeiptr elemsCount, const GLuint* data) {
        this->elemSize = elemSize;
        this->elemsCount = elemsCount;
        glBufferData(getTarget(), elemsCount * elemSize * sizeof(GLuint), data, GL_STATIC_DRAW);
        dataType = GL_UNSIGNED_INT;
    }

    void setData(GLsizeiptr elemSize, GLsizeiptr elemsCount, const GLubyte* data) {
        this->elemSize = elemSize;
        this->elemsCount = elemsCount;
        glBufferData(getTarget(), elemsCount * elemSize * sizeof(GLubyte), data, GL_STATIC_DRAW);
        dataType = GL_UNSIGNED_BYTE;
    }

    bool init(const StArray<StGLVec2>& theArray) {
        return init(2, GLsizeiptr(theArray.size()), theArray.getFirst().getData());
    }

    bool init(const StArray<StGLVec3>& theArray) {
        return init(3, GLsizeiptr(theArray.size()), theArray.getFirst().getData());
    }

    bool init(const StArray<StGLVec4>& theArray) {
        return init(4, GLsizeiptr(theArray.size()), theArray.getFirst().getData());
    }

    bool init(GLsizeiptr elemSize, GLsizeiptr elemsCount, const GLfloat* data) {
        if(!init()) {
            return false;
        }
        bind();
        setData(elemSize, elemsCount, data);
        unbind();
        return true;
    }

    bool init(GLsizeiptr elemSize, GLsizeiptr elemsCount, const GLuint* data) {
        if(!init()) {
            return false;
        }
        bind();
        setData(elemSize, elemsCount, data);
        unbind();
        return true;
    }

    bool init(GLsizeiptr elemSize, GLsizeiptr elemsCount, const GLubyte* data) {
        if(!init()) {
            return false;
        }
        bind();
        setData(elemSize, elemsCount, data);
        unbind();
        return true;
    }

    /**
     * @return elemSize (GLsizeiptr ) - specifies the number of components per generic vertex attribute. Must be 1, 2, 3, or 4;
     */
    GLsizeiptr getElemSize() const {
        return elemSize;
    }

    GLsizeiptr getElemsCount() const {
        return elemsCount;
    }

    /**
     * @return the data type of each component in the array.
     */
    GLenum getDataType() const {
        return dataType;
    }

    virtual GLenum getTarget() const {
        return GL_ARRAY_BUFFER;
    }

    void bindVertexAttrib(StGLVarLocation attribLoc) const {
        if(isValid() && attribLoc.isValid()) {
            bind();
            glEnableVertexAttribArray(attribLoc);
            glVertexAttribPointer(attribLoc, GLint(getElemSize()), getDataType(), GL_FALSE, 0, NULL);
        }
    }

    void unBindVertexAttrib(StGLVarLocation attribLoc) const {
        if(isValid() && attribLoc.isValid()) {
            glDisableVertexAttribArray(attribLoc);
            unbind();
        }
    }

    StString toString() const {
        return StString("OpenGL buffer #") + bufferId;
    }

};

template<> inline void StArray<GLuint>::sort() {}

class ST_LOCAL StGLIndexBuffer : public StGLVertexBuffer {

        public:

    StGLIndexBuffer()
    : StGLVertexBuffer() {}

    virtual ~StGLIndexBuffer() {
        //
    }

    virtual GLenum getTarget() const {
        return GL_ELEMENT_ARRAY_BUFFER;
    }

    bool init(const StArray<GLuint>& theArray) {
        return StGLVertexBuffer::init(GLsizeiptr(1), GLsizeiptr(theArray.size()), &theArray.getFirst());
    }

};

#endif //__StGLVertexBuffer_h_
