/**
 * Copyright © 2009-2011 Kirill Gavrilov <kirill@sview.ru>
 *
 * Distributed under the Boost Software License, Version 1.0.
 * See accompanying file license-boost.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt
 */

#ifndef __StStereoParams_h_
#define __StStereoParams_h_

#include <StGL/StGL.h>
#include <StGL/StGLVec.h>

#include <StGLStereo/StFormatEnum.h>
#include <StTemplates/StHandle.h>

class ST_LOCAL StStereoParams {

        public:

    typedef enum tagViewMode {
        FLAT_IMAGE,        // flat image
        PANORAMA_SPHERE,   // spherical panorama
        PANORAMA_CYLINDER, // cylindrical panorama
    } ViewMode;

    static StString GET_VIEW_MODE_NAME(ViewMode viewMode) {
        switch(viewMode) {
            case PANORAMA_SPHERE: return "sphere";
            case FLAT_IMAGE:
            default: return "flat";
        }
    }

    static ViewMode GET_VIEW_MODE_FROM_STRING(const StString& viewModeString) {
        if(viewModeString.isStartsWithIgnoreCase("sphere")) {
            return PANORAMA_SPHERE;
        } else {
            return FLAT_IMAGE;
        }
    }

        private:

    // TODO (Kirill Gavrilov#2#) implement velocity + timer behavior for smoothness
    StFormatEnum mySrcFormat;
    GLint        sepHPx; // horizontal separation in pixels
    GLint    sepHZeroPx; // zero-parallax separation in pixels
    GLint        sepVPx; // vertical separation in pixels
    GLint     sepStepPx;
    GLfloat      sepRGr; // angular separation in degrees
    GLfloat   zRotateGr; // rotation angle in degrees
    StGLVec2   position; // relative position
    GLfloat     posStep;
    GLfloat scaleFactor;
    GLfloat   scaleStep;
    GLfloat   panoTheta; // anles for panorama view
    GLfloat     panoPhi; // anles for panorama view
    ViewMode   viewMode;
    bool         swapLR;

        public:

    StStereoParams(ViewMode viewMode = FLAT_IMAGE)
    : mySrcFormat(ST_V_SRC_MONO),
      sepHPx(0),
      sepHZeroPx(0),
      sepVPx(0),
      sepStepPx(2),
      sepRGr(0.0f),
      zRotateGr(0.0f),
      position(),
      posStep(0.01f),
      scaleFactor(1.0f),
      scaleStep(0.02f),
      panoTheta(0.0f),
      panoPhi(0.0f),
      viewMode(viewMode),
      swapLR(false) {
        //
    }

    ~StStereoParams() {
        //
    }

    bool isMono() const {
        return mySrcFormat == ST_V_SRC_MONO;
    }

    StFormatEnum getSrcFormat() const {
        return mySrcFormat;
    }

    void setSrcFormat(StFormatEnum theSrcFormat) {
        mySrcFormat = theSrcFormat;
    }

    ViewMode getViewMode() const {
        return viewMode;
    }

    void setViewMode(ViewMode viewMode) {
        this->viewMode = viewMode;
    }

    void nextViewMode() {
        switch(viewMode) {
            case FLAT_IMAGE: viewMode = PANORAMA_SPHERE; break;
            case PANORAMA_SPHERE:
            case PANORAMA_CYLINDER:
            default: viewMode = FLAT_IMAGE;
        }
    }

    /**
     * Get scaling factor ('zoom').
     */
    GLfloat getScale() const {
        return scaleFactor;
    }

    GLfloat getScaleStep() const {
        return scaleStep;
    }

    /**
     * Get horizontal separation in pixels.
     * This number means delta between views.
     */
    GLint getSepHorizontal() const {
        return sepHZeroPx + sepHPx;
    }

    void setSepHorizontalBase(GLint theSepHPx) {
        sepHZeroPx = theSepHPx;
    }

    /**
     * Get vertical separation in pixels.
     * This number means delta between views.
     */
    GLint getSepVertical() const {
        return sepVPx;
    }

    /**
     * Get angular separation between views in degrees.
     */
    GLfloat getSepRotation() const {
        return sepRGr;
    }

    /**
     * Get flat posionion vector.
     */
    const StGLVec2& getPos() const {
        return position;
    }

    GLfloat getTheta() const {
        return panoTheta;
    }

    GLfloat getPhi() const {
        return panoPhi;
    }

    /**
     * Get common rotation anlge in degrees.
     */
    GLfloat getZRotate() const {
        return zRotateGr;
    }

    void setZRotate(const GLfloat theAngleDegrees) {
        zRotateGr = theAngleDegrees;
    }

    /**
     * @return true if swap Left/Right flag setted.
     */
    bool isSwapLR() const {
        return swapLR;
    }

    /**
     * @param toSwapLR (bool ) - set swap Left/Right flag.
     */
    void setSwapLR(bool toSwapLR) {
        swapLR = toSwapLR;
    }

    /**
     * Reverse swap Left/Right flag.
     */
    void doSwapLR() {
        swapLR = !swapLR;
    }

    void scaleIn(GLfloat steps = 1.0f) {
        scaleFactor *= (1.0f + steps * scaleStep);
    }

    void scaleOut(GLfloat steps = 1.0f) {
        scaleFactor /= (1.0f + steps * scaleStep);
    }

    void incSepHorizontal() {
        sepHPx += sepStepPx;
    }

    void decSepHorizontal() {
        sepHPx -= sepStepPx;
    }

    void incSepVertical() {
        sepVPx += sepStepPx;
    }

    void decSepVertical() {
        sepVPx -= sepStepPx;
    }

    void incSepRotation() {
        sepRGr += 0.1f;
    }

    void decSepRotation() {
        sepRGr -= 0.1f;
    }

    /**
     * Help method used in StGLImageRegion to fit image into rectangle.
     */
    static inline StGLVec2 getRatioScale(const GLfloat rectRatio, const GLfloat imageRatio) {
        if(rectRatio < imageRatio) {
            return StGLVec2(rectRatio, rectRatio / imageRatio);
        } else {
            return StGLVec2(imageRatio, 1.0f);
        }
    }

    StGLVec2 moveFlatDelta(const StGLVec2& moveVec, const GLfloat rectRatio) const {
        return StGLVec2((moveVec.x() * rectRatio) / scaleFactor, moveVec.y() / scaleFactor);
    }

    void moveFlat(const StGLVec2& moveVec, const GLfloat rectRatio) {
        StGLVec2 mvDeltaVec = moveFlatDelta(moveVec, rectRatio);
        position.x() += mvDeltaVec.x();
        position.y() += mvDeltaVec.y();
    }

    void moveSphere(const StGLVec2& moveVec) {
        panoPhi += moveVec.x();
        panoTheta += moveVec.y();
    }

    void moveToRight() {
        switch(viewMode) {
            case PANORAMA_SPHERE:
            case PANORAMA_CYLINDER: panoPhi += 2.0f; break;
            case FLAT_IMAGE:
            default: position.x() += posStep / scaleFactor;
        }
    }

    void moveToLeft() {
        switch(viewMode) {
            case PANORAMA_SPHERE:
            case PANORAMA_CYLINDER: panoPhi -= 2.0f; break;
            case FLAT_IMAGE:
            default: position.x() -= posStep / scaleFactor;
        }
    }

    void moveToDown() {
        switch(viewMode) {
            case PANORAMA_SPHERE:
            case PANORAMA_CYLINDER: panoTheta -= 2.0f; break;
            case FLAT_IMAGE:
            default: position.y() -= posStep / scaleFactor;
        }
    }

    void moveToUp() {
        switch(viewMode) {
            case PANORAMA_SPHERE:
            case PANORAMA_CYLINDER: panoTheta += 2.0f; break;
            case FLAT_IMAGE:
            default: position.y() += posStep / scaleFactor;
        }
    }

    void incZRotate() {
        zRotateGr -= 90.0f;
    }

    void decZRotate() {
        zRotateGr += 90.0f;
    }

    void incZRotateL() {
        zRotateGr -= 0.1f;
    }

    void decZRotateL() {
        zRotateGr += 0.1f;
    }

    void reset() {
        sepHPx = sepVPx = 0;
        sepRGr = zRotateGr = 0.0f;
        position.x() = 0.0f;
        position.y() = 0.0f;
        panoTheta = 0.0f;
        panoPhi = 0.0f;
        scaleFactor = 1.0f;
        swapLR = false;
    }

};

#endif //__StStereoParams_h_
