/**
 * Copyright © 2009, 2011 Kirill Gavrilov <kirill@sview.ru>
 *
 * Distributed under the Boost Software License, Version 1.0.
 * See accompanying file license-boost.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt
 */

#ifndef __StEvent_h_
#define __StEvent_h_

#if (defined(_WIN32) || defined(__WIN32__))
    #include <windows.h> // we used global header instead Winbase.h to prevent namespaces collisions
#else
    #include <pthread.h>
    #include <unistd.h>
    #include <errno.h>
    #include <sys/time.h>
#endif

#include <stTypes.h>

// MacOS headers currently by default define check macro for assertions
#ifdef check
    #undef check
#endif

/**
 * This is simple Event class that helps to talk threads between themselfs.
 * Object is similar to WinAPI Event.
 */
class ST_LOCAL StEvent {

        private:

#if (defined(_WIN32) || defined(__WIN32__))
    HANDLE event;
#else
    pthread_mutex_t mx;
    pthread_cond_t cv;
    bool flag;
#endif

        public:

    StEvent(bool isSignalling = true) {
    #if (defined(_WIN32) || defined(__WIN32__))
        event = CreateEvent(0,            // default security attributes
                            true,         // manual-reset event
                            isSignalling, // initial state is signaled
                            NULL);        // no name
    #else
        pthread_mutex_init(&mx, 0);
        pthread_cond_init(&cv, 0);
        flag = isSignalling;
    #endif
    }

    ~StEvent() {
    #if (defined(_WIN32) || defined(__WIN32__))
        CloseHandle(event);
    #else
        pthread_mutex_destroy(&mx);
        pthread_cond_destroy(&cv);
    #endif
    }

    /**
     * Set event into signalling state.
     */
    void set() {
    #if (defined(_WIN32) || defined(__WIN32__))
        SetEvent(event);
    #else
        pthread_mutex_lock(&mx);
        flag = true;
        pthread_cond_broadcast(&cv);
        pthread_mutex_unlock(&mx);
    #endif
    }

    /**
     * Reset event (unset signalling state)
     */
    void reset() {
    #if (defined(_WIN32) || defined(__WIN32__))
        ResetEvent(event);
    #else
        pthread_mutex_lock(&mx);
        flag = false;
        pthread_mutex_unlock(&mx);
    #endif
    }

    /**
     * Wait for Event (infinity).
     */
    void wait() {
    #if (defined(_WIN32) || defined(__WIN32__))
        WaitForSingleObject(event, INFINITE);
    #else
        pthread_mutex_lock(&mx);
        if(!flag) {
            pthread_cond_wait(&cv, &mx);
        }
        pthread_mutex_unlock(&mx);
    #endif
    }

    /**
     * Wait for signal requested time.
     * @param timeMilliseconds (const size_t& ) - wait limit in millisecods;
     * @return true if get event.
     */
    bool wait(const size_t& timeMilliseconds) {
    #if (defined(_WIN32) || defined(__WIN32__))
        return (WaitForSingleObject(event, (DWORD )timeMilliseconds) != WAIT_TIMEOUT);
    #else
        bool isSignalled = true;
        pthread_mutex_lock(&mx);
        if(!flag) {
            struct timeval now;
            struct timespec timeout;
            gettimeofday(&now, NULL);
            size_t seconds = (timeMilliseconds / 1000);
            size_t nanoseconds = (timeMilliseconds - seconds * 1000) * 1000;
            timeout.tv_sec = now.tv_sec + (time_t )seconds;
            timeout.tv_nsec = now.tv_usec + (long )nanoseconds;
            isSignalled = (pthread_cond_timedwait(&cv, &mx, &timeout) != ETIMEDOUT);
        }
        pthread_mutex_unlock(&mx);
        return isSignalled;
    #endif
    }

    /**
     * Do not wait for signal - just test it state.
     * @return true if get event.
     */
    bool check() {
    #if (defined(_WIN32) || defined(__WIN32__))
        return (WaitForSingleObject(event, (DWORD )0) != WAIT_TIMEOUT);
    #else
        bool isSignalled = true;
        pthread_mutex_lock(&mx);
        if(!flag) {
            struct timeval now;
            struct timespec timeout;
            gettimeofday(&now, NULL);
            timeout.tv_sec = now.tv_sec;
            // TODO (Kirill Gavrilov) wich values must be here and need we this block here?
            timeout.tv_nsec = now.tv_usec + 100;
            isSignalled = (pthread_cond_timedwait(&cv, &mx, &timeout) != ETIMEDOUT);
        }
        pthread_mutex_unlock(&mx);
        return isSignalled;
    #endif
    }

    /**
     * Method perform two steps at-once - reset the event object
     * and returns true if it was in signaling state.
     * @return true if event object was in signaling state.
     */
    bool checkReset() {
        bool wasSignalled = check();
        reset();
        return wasSignalled;
    }

};

#endif //__StEvent_h_
