/**
 * Copyright © 2009-2010 Kirill Gavrilov <kirill@sview.ru>
 *
 * Distributed under the Boost Software License, Version 1.0.
 * See accompanying file license-boost.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt
 */

#ifndef __StFPSControl_h_
#define __StFPSControl_h_

#include "StFPSMeter.h"
#include "StThreads.h"

/**
 * Class extend FPS measurements features with possibility
 * to adjust FPS to target using thread sleeping.
 */
class ST_LOCAL StFPSControl : public StFPSMeter {

        private:

    class StSleeper {

            private:

        size_t currTimer;
        int   slTimes[4];

        int value() const {
            return slTimes[currTimer];
        }

        void next() {
            if(++currTimer > 3) {
                currTimer = 0;
            }
        }

            public:

        StSleeper(const int slTime)
        : currTimer(0) {
            slTimes[0] = slTimes[1] = slTimes[2] = slTimes[3] = slTime;
        }

        ~StSleeper() {}

        bool isZero() const {
            return slTimes[0] <= 1 && slTimes[1] <= 1 && slTimes[2] <= 1 && slTimes[3] <= 1;
        }

        void zero() {
            slTimes[0] = slTimes[1] = slTimes[2] = slTimes[3] = 1;
        }

        void inc() {
            size_t toInc = 0;
            if(slTimes[0] + slTimes[1] <= slTimes[2] + slTimes[3]) {
                toInc = (slTimes[0] <= slTimes[1]) ? 0 : 1;
            } else {
                toInc = (slTimes[2] <= slTimes[3]) ? 2 : 3;
            }
            ++slTimes[toInc];
            //ST_DEBUG_LOG("sleepToTrg++ " + slTimes[0] + " " + slTimes[1] + " " + slTimes[2] + " " + slTimes[3]);
        }

        void dec() {
            if(isZero()) {
                return;
            }
            size_t toDec = 0;
            if(slTimes[0] + slTimes[1] >= slTimes[2] + slTimes[3]) {
                toDec = (slTimes[0] >= slTimes[1]) ? 0 : 1;
            } else {
                toDec = (slTimes[2] >= slTimes[3]) ? 2 : 3;
            }
            --slTimes[toDec];
            //ST_DEBUG_LOG("sleepToTrg-- " + slTimes[0] + " " + slTimes[1] + " " + slTimes[2] + " " + slTimes[3]);
        }

        void sleep() {
            StThread::sleep(value());
            next();
        }

    };

    StSleeper stSleeper;
    double    targetFPS; // target average FPS
    int        decCount;
    bool    isIncreased;

    /**
     * Try to adjust sleep timers to tend to target FPS.
     */
    void adjustFPSToTarget(const double newFPS) {
        const double aFPSPrecision = targetFPS * 0.01; // 1%
        const double diffFPS = newFPS - targetFPS;
        if(diffFPS > aFPSPrecision) {
            stSleeper.inc();
            if(diffFPS >= 8.0) {
                // great diff - reduce time to adjust
                for(size_t iter = 0; iter < 3; ++iter) {
                    stSleeper.inc();
                }
            }
        } else if(diffFPS < -aFPSPrecision) {
            if(stSleeper.isZero()) {
                // TODO (Kirill Gavrilov#7) if target FPS higher than maximal
                //                          - no mehanism to reduce CPU utilization
                return;
            }
            stSleeper.dec();
            if(diffFPS <= -8.0) {
                // great diff - reduce time to adjust
                for(size_t iter = 0; iter < 3; ++iter) {
                    stSleeper.dec();
                }
            }
        }
    }

    /**
     * Try to reduce CPU utilization (using sleep timers)
     * with minimal affect to FPS.
     */
    void adjustFPSToMax(const double prevFPS, const double newFPS) {
        static const double ST_FPS_MINIMUM = 30.0; // considered as minimal FPS...
        const double aFPSPrecision = 0.5 * (newFPS + prevFPS) * 0.01; // 1%
        const double diffFPS = newFPS - prevFPS;
        if(isIncreased && (diffFPS < -aFPSPrecision || newFPS < ST_FPS_MINIMUM)) {
            // increased sleep time affect in FPS slowdown
            // - reduce sleep time
            stSleeper.dec();
            isIncreased = false;
            ++decCount;
        } else if(decCount > 0 && (diffFPS > aFPSPrecision)) {
            // decreased sleep time affect in FPS increasing
            // - reduce sleep time more to tend to FPS maximum
            stSleeper.dec();
            ++decCount;
        } else {
            // try to reduce CPU utilization while not affecting FPS
            // - increase sleep time
            stSleeper.inc();
            isIncreased = true;
            if(decCount > 0) {
                --decCount;
            }
        }
    }

        public:

    StFPSControl()
    : StFPSMeter(),
      stSleeper(10),
      targetFPS(-1.0),
      decCount(0),
      isIncreased(false) {
        //
    }

    virtual ~StFPSControl() {
        //
    }

    /**
     * Increment frames counter.
     */
    virtual bool nextFrame() {
        const double prevFPS = getAverage();
        if(StFPSMeter::nextFrame()) {
            const double newFPS = getAverage();
            // compute sleep time to get target FPS
            if(targetFPS > 0.0) {
                //ST_DEBUG_LOG("adjustFPSToTarget " + targetFPS);
                adjustFPSToTarget(newFPS);
            } else if(targetFPS == 0.0) {
                //ST_DEBUG_LOG("adjustFPSToMax " + targetFPS);
                adjustFPSToMax(prevFPS, newFPS);
            }
            return true;
        }
        return false;
    }

    /**
     * @param fps (const double& ) - target fps.
     */
    void setTargetFPS(const double fps) {
        targetFPS = fps;
    }

    /**
     * Sleep the thread to fit the target FPS.
     * If target FPS is 0.0 or system (GPU) can't reach the limit
     * sleep time is a highest value not reduced the FPS.
     * If target FPS set to -1.0 sleep ignored.
     * Notice: on some system (Windows) you should adjust system timer before call!
     */
    void sleepToTarget() {
        if(targetFPS >= 0.0) {
            stSleeper.sleep();
        }
    }

};

#endif // __StFPSControl_h_
