/**
 * Copyright © 2008-2011 Kirill Gavrilov <kirill@sview.ru>
 *
 * Distributed under the Boost Software License, Version 1.0.
 * See accompanying file license-boost.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt
 */

#ifndef __StTimer_h_
#define __StTimer_h_

#include <stTypes.h>

#if (defined(_WIN32) || defined(__WIN32__))
    #include <windows.h>
#else
    #include <sys/time.h>
    #include <stdlib.h> // just for NULL declaration
#endif

/**
 * High Resolution Timer.
 * This timer is able to measure the elapsed time with 1 micro-second accuracy
 * in both Windows, Linux and Unix systems.
 */
class ST_LOCAL StTimer {

        private:

#if (defined(_WIN32) || defined(__WIN32__))
    typedef LARGE_INTEGER stTimeCounter_t;
#else
    typedef timeval       stTimeCounter_t;
#endif

    double timeInMicroSec;                  // cumulative elapsed time
    stTimeCounter_t counterStart;
    stTimeCounter_t counterEnd;
    bool isPaused;                          // pause flag

    void fillCounter(stTimeCounter_t& counter) {
    #if (defined(_WIN32) || defined(__WIN32__))
        QueryPerformanceCounter(&counter);
    #else
        gettimeofday(&counter, NULL);
    #endif
    }

#if (defined(_WIN32) || defined(__WIN32__))
    static double winInvFrequency() {
        // according to MSDN frequency cannot change while the system is running
        LARGE_INTEGER frequency; // ticks per second
        QueryPerformanceFrequency(&frequency);
        return (1000000.0 / double(frequency.QuadPart));
    }
#endif

    double timeFromStart() {
        fillCounter(counterEnd);
    #if (defined(_WIN32) || defined(__WIN32__))
        static const double INV_FREQ = winInvFrequency();
        return double(counterEnd.QuadPart - counterStart.QuadPart) * INV_FREQ;
    #else
        return (double(counterEnd.tv_sec - counterStart.tv_sec) * 1000000.0) + double(counterEnd.tv_usec - counterStart.tv_usec);
    #endif
    }

    /**
     * Reset current timestamp to zero.
     */
    void reset() {
        stMemSet(&counterStart, 0, sizeof(counterStart));
        stMemSet(&counterEnd,   0, sizeof(counterEnd));
        timeInMicroSec = 0.0;
    }

        public:

    /**
     * @param isStart (bool ) - start timer on create flag.
     */
    StTimer(bool isStart = false)
    : timeInMicroSec(0.0),
      isPaused(true) {
        if(isStart) {
            restart(0.0);
        } else {
            reset();
        }
    }

    ~StTimer() {}

    /**
     * Just start the timer again.
     */
    void resume() {
        stMemSet(&counterStart, 0, sizeof(counterStart));
        stMemSet(&counterEnd,   0, sizeof(counterEnd));
        isPaused = false;

        fillCounter(counterStart);
    }

    /**
     * Clean start.
     * @param startTime (double ) - initial time in micro second.
     */
    void restart(double theStartFromMicroSec = 0.0) {
        reset();
        timeInMicroSec = theStartFromMicroSec;
        isPaused = false;

        fillCounter(counterStart);
    }

    bool isOn() const {
        return !isPaused;
    }

    /**
     * Pause the timer (freeze current timestamp).
     */
    void pause() {
        // increment our timer value
        timeInMicroSec += getElapsedTimeFromLastStartInMicroSec();
        // set timer paused flag
        isPaused = true;
    }

    /**
     * Stop the timer (and clean current timestamp).
     */
    void stop() {
        pause();
        reset();
    }

    /**
     * @return seconds (double ) - timer in seconds.
     */
    double getElapsedTime() {
        return this->getElapsedTimeInSec();
    }

    /**
     * @return seconds (double ) - timer in seconds.
     */
    double getElapsedTimeInSec() {
        return this->getElapsedTimeInMicroSec() * 0.000001;
    }

    /**
     * @return milliSeconds (double ) - timer in milli-seconds.
     */
    double getElapsedTimeInMilliSec() {
        return this->getElapsedTimeInMicroSec() * 0.001;
    }

    /**
     * Main function.
     * @return microSeconds (double ) - timer in micro-seconds.
     */
    double getElapsedTimeInMicroSec() {
        return timeInMicroSec + getElapsedTimeFromLastStartInMicroSec();
    }

    /**
     * Main function - return time, elapsed from last start.
     * @return microSeconds (double ) - micro-seconds from start.
     */
    double getElapsedTimeFromLastStartInMicroSec() {
        return isPaused ? 0.0 : timeFromStart();
    }

};

#endif // __StTimer_h_
